
import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.tools.io.*;
import org.bouncycastle.graphics.rt.util.ParametersMap;
import org.bouncycastle.graphics.rt.shaders.ct.*;

/**
 * A matte surface shader with tiling.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;Matte diffuse="0.9" blend="1.0"/&gt;
 *         &lt;Tile image="tile.pix" repeating="true"&gt;
 *      &lt;MatteTiled&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float    "CDATA"&gt;
 *     &lt;!ENTITY %Color3f  "CDATA"&gt;
 *     &lt;!ENTITY %Vector2f "CDATA"&gt;
 *
 *     &lt;!ELEMENT MatteTiled (Tile)&gt;
 *     &lt;!ATTLIST MatteTiled
 *         ambient              (%Color3f; | %Float;) "0, 0, 0"
 *         diffuse              (%Color3f; | %Float;) "1, 1, 1"
 *         rotate               %Float; "0"
 *         scale                %Vector2f; "1, 1"&gt;
 * </pre>
 * </code>
 */
public class MatteTiled
    extends Matte
{
    Tile        tile;
    float       blend = 1;
    float       sScale = 1;
    float       tScale = 1;
    float       cosAng = 1;
    float       sinAng = 0;

    public MatteTiled(
        World   world,
        Tile    tile)
    {
        super(world);

        this.tile = tile;
    }

    private MatteTiled(
        World   world)
    {
        super(world);
    }

    public static Shader getInstance(
        World         world,
        RMLReader       in)
        throws IOException
    {
        MatteTiled   m = new MatteTiled(world);

        m.readObject(in);

        return m;
    }

    protected void processObject(
        Object  obj)
    {
        if (obj instanceof Tile)
        {
            this.tile = (Tile)obj;
        }
        else
        {
            super.processObject(obj);
        }
    }

    /**
     * handle the parameters we want, passing the extras
     * to our super class for processing.
     */
    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        setBlend(params.getFloat("blend", "1"));

        if (params.get("scale") != null)
        {
            StringTokenizer st = new StringTokenizer((String)params.get("scale"), " ,");
            setScale(Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()));
        }

        setRotation(params.getFloat("rotate", "0"));
    }

    /**
     * set the blend factor for the tiling.
     */
    public void setBlend(
        float    blend)
    {
        this.blend = blend;
    }

    /**
     * set a rotation in degrees (around "u") for this tiling.
     */
    public void setRotation(
        float   angle)
    {
        this.cosAng = (float)Math.cos(angle * Math.PI / 180.0);
        this.sinAng = (float)Math.sin(angle * Math.PI / 180.0);
    }

    /**
     * set the (s, t) scaling for this tiling.
     */
    public void setScale(
        float    sScale,
        float    tScale)
    {
        this.sScale = sScale;
        this.tScale = tScale;
    }

    /**
     * the shade routine. If you want more than the below, this is
     * the method to override.
     */
    public void shade(
        Hit        hit,
        Pixel      p)
    {
        p.set(0, 0, 0, 1);

        float   s = hit.S() / sScale;
        float   t = hit.T() / tScale;
        Pixel   pix = tile.getPixel(s * cosAng - t * sinAng, t * cosAng + s * sinAng);

        if (pix != null)
        {
            p.scaledAdd(pix, ambientContribution(world));

            p.scaledAdd(pix, diffuse, diffuseContribution(hit.visibleLights(world), hit));
        }
        else
        {
            p.scaledAdd(hit.color(), ambientContribution(world));

            p.scaledAdd(hit.color(), diffuse, diffuseContribution(hit.visibleLights(world), hit));
        }

        p.scale(alpha);
    }
}
