
import java.util.*;
import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.mappers.*;
import org.bouncycastle.graphics.rt.objects.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a sphere with holes in it around the axis... Essentially xHole gives the
 * radius of a cylinder around the xAxis which is removed from the surface,
 * yHole give the radius of a cylinder around the yAxis, etc...
 * <p>
 * Note: the holes are punched in the spheres cannonical coordinate system.
 * <p>
 * Sample RML:
 * <code>
 * <pre>
 *          &lt;PunchedSphere radius="1.0" center="0, 0, 0" xHole="0.2" yHole="0.2" zHole="0.2"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Float "CDATA"&gt;
 *     &lt;!ENTITY %Point3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT PunchedSphere EMPTY&gt;
 *     &lt;!ATTLIST PunchedSphere
 *         color  %Color3f; "1, 1, 1"
 *         radius %Float;   "1"
 *         center %Point3f; "0, 0, 0"
 *         xHole %Float;    "0"
 *         yHole %Float;    "0"
 *         zHole %Float;    "0"&gt;
 * </pre>
 * </code>
 */
public class PunchedSphere
    extends Bounded
    implements NormalCalculator
{
    static final float TOL = (float)0.0001;

    static BBox    canonicalBBox = new BBox(-1, -1, -1, 1, 1, 1);
    private static TextureMapper    stMapper = SphericalMap.canonicalMap;

    private float   tolerance;
    private float   xHoleRadSqu, yHoleRadSqu, zHoleRadSqu;

    public PunchedSphere(
        Point3d       c,
        double        r,
        double        xHole,
        double        yHole,
        double        zHole)
    {
        init(new Point3f(c), r, xHole, yHole, zHole);
    }

    private PunchedSphere()
    {
    }

    public static PunchedSphere getInstance(
        RMLReader       in)
        throws IOException
    {
        PunchedSphere s = new PunchedSphere();

        s.readObject(in);

        return s;
    }

    protected void processParameters(
        ParametersMap   params)
    {
        init(params.getPoint("center", "0, 0, 0"),
                params.getFloat("radius", "1"),
                params.getFloat("xHole", "0"),
                params.getFloat("yHole", "0"),
                params.getFloat("zHole", "0"));

        super.processParameters(params);
    }

    private void init(
        Point3f        c,
        double        r,
        double        xHole,
        double        yHole,
        double        zHole)
    {
        setBBox(canonicalBBox);

        Transform   trans = new Transform();

        trans.scale(r, r, r);

        trans.translate(c.x, c.y, c.z);

        this.transform(trans);

        this.xHoleRadSqu = (float)((xHole * xHole) / (r * r));
        this.yHoleRadSqu = (float)((yHole * yHole) / (r * r));
        this.zHoleRadSqu = (float)((zHole * zHole) / (r * r));

        tolerance = (float)(TOL * r);
    }

    private float adjustT(
        float       t,
        Point3f     org,
        Vector3f    dir)
    {
        float   x = dir.x * t + org.x;
        float   y = dir.y * t + org.y;
        float   z = dir.z * t + org.z;

        if (xHoleRadSqu != 0 && (y * y + z * z) < xHoleRadSqu)
        {
            return 0;
        }

        if (yHoleRadSqu != 0 && (x * x + z * z) < yHoleRadSqu)
        {
            return 0;
        }

        if (zHoleRadSqu != 0 && (x * x + y * y) < zHoleRadSqu)
        {
            return 0;
        }

        return t;
    }

    public boolean isIntersected(
        Ray    r)
    {
        float       a, b, c, d;
        Point3f     org;
        Vector3f    dir;

        Ray     nr = this.createObjectSpaceRay(r);

        org = nr.org;
        dir = nr.dir;

        a = dir.x * dir.x + dir.y * dir.y 
            + dir.z * dir.z;

        b = -(dir.x * org.x + dir.y * org.y 
            + dir.z * org.z);

        if (b < 0 && a > 0)
        {
            nr.dispose();

            return false;
        }

        c = (org.x * org.x + org.y * org.y 
            + org.z * org.z) - 1;

        d = b * b - a * c;

        if (d < 0)
        {
            nr.dispose();

            return false;
        }
    
        d = (float)Math.sqrt(d);

        float   t1 = ((b - d) / a);
        float   t2 = ((b + d) / a);

        t1 = adjustT(t1, org, dir);
        t2 = adjustT(t2, org, dir);

        if (t1 > tolerance)
        {
            r.setHit(this, t1, this, this, stMapper, this.getShader());

            nr.dispose();

            return true;
        }
        else if (t2 > tolerance)
        {
            r.setHit(this, t2, this, this, stMapper, this.getShader());

            nr.dispose();

            return true;
        }

        return false;
    }

    public boolean isShadowing(
        Ray    r)
    {
        if (!isIntersected(r))
        {
            return false;
        }

        return (r.hit.t < r.tMax);
    }

    /**
     * this one can't be used in CSG (an exercise for the reader!)
     */
    public HitList intersections(
        Ray         r)
    {
        return null;
    }

    public void getN(
        Hit         hit,
        Vector3f    n)
    {
        n.set(hit.oP());
    }
}
