package org.bouncycastle.graphics.rt;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * create a spline interpolated color map, using either on input array
 * of colors, or by parsing the appropriate XML.
 * <p>
 * An example XML for a color map is:
 * <pre>
 *      &lt;ColorMap&gt;
 *          0, 0, 0,
 *          0.1, 0.2, 0.1,
 *          0.3, 0.4, 0.1,
 *          0.5, 0.6, 0.1,
 *          0.7, 0.8, 0.1,
 *          1.0, 1.0, 0.1
 *      &lt;/ColorMap&gt;
 * </pre>
 */
public class ColorMap
{
    /**
     * basis matrix for a cardinal (or Catmul-Rom) spline.
     */
    private static Matrix4f  basis = new Matrix4f(
        -0.5f,  1.5f,    -1.5f,   0.5f,
        1.0f,   -2.5f,   2.0f,    -0.5f,
        -0.5f,  0.0f,    0.5f,    0.0f,
        0.0f,   1.0f,    0.0f,    0.0f
    );

    Matrix4f[]  steps;

    public ColorMap(
        Color3f[]   colors)
    {
        init(colors);
    }

    public ColorMap(
        RMLReader   in)
        throws IOException
    {
        Color3f[]   colors = null;
        int         tok;

        while ((tok = in.readNext()) != RMLReader.EOF)
        {
            String  name = in.toString();

            if (tok == RMLReader.END_TAG && name.equals("ColorMap"))
            {
                break;
            }

            StringTokenizer st = new StringTokenizer(in.toString(), ", \t");
   
            if ((st.countTokens() % 3) == 0)
            {
                colors = new Color3f[st.countTokens() / 3];

                for (int i = 0; i != colors.length; i++)
                {
                    Color3f c = new Color3f();
            
                    c.x = Float.parseFloat(st.nextToken());
                    c.y = Float.parseFloat(st.nextToken());
                    c.z = Float.parseFloat(st.nextToken());

                    colors[i] = c;
                }
            }
            else
            {
                throw new IOException("badly formatted ColorMap tag");
            }
        }

        init(colors);
    }

    /**
     * set up the cofactors matrix for the spline, note we need to repeat
     * the end points as a cardinal spline goes from P1 to Pn-1 for P0 to Pn...
     */
    private void init(
        Color3f[]   colors)
    {
        if (colors == null || colors.length < 4)
        {
            throw new IllegalArgumentException("at least 4 colors are required");
        }

        steps = new Matrix4f[colors.length - 1];

        Matrix4f    m = new Matrix4f();

        m.setRow(0, colors[0].x, colors[0].y, colors[0].z, 1);
        m.setRow(1, colors[0].x, colors[0].y, colors[0].z, 1);
        m.setRow(2, colors[1].x, colors[1].y, colors[1].z, 1);
        m.setRow(3, colors[2].x, colors[2].y, colors[2].z, 1);

        steps[0] = new Matrix4f();

        steps[0].mul(basis, m);

        for (int i = 0; i != colors.length - 3; i++)
        {
            m.setRow(0, colors[i].x, colors[i].y, colors[i].z, 1);
            m.setRow(1, colors[i + 1].x, colors[i + 1].y, colors[i + 1].z, 1);
            m.setRow(2, colors[i + 2].x, colors[i + 2].y, colors[i + 2].z, 1);
            m.setRow(3, colors[i + 3].x, colors[i + 3].y, colors[i + 3].z, 1);

            steps[i + 1] = new Matrix4f();

            steps[i + 1].mul(basis, m);
        }

        int         idx = colors.length - 3;

        m.setRow(0, colors[idx].x, colors[idx].y, colors[idx].z, 1);
        m.setRow(1, colors[idx + 1].x, colors[idx + 1].y, colors[idx + 1].z, 1);
        m.setRow(2, colors[idx + 2].x, colors[idx + 2].y, colors[idx + 2].z, 1);
        m.setRow(3, colors[idx + 2].x, colors[idx + 2].y, colors[idx + 2].z, 1);

        steps[idx + 1] = new Matrix4f();

        steps[idx + 1].mul(basis, m);
    }

    /**
     * get the color for the index into the color map f. f is expected
     * to be in the range 0 to 1.
     */
    public Color3f colorAt(
        double   f)
    {
        int stepNo;

        if (f >= 1)
        {
            f = 1;
            stepNo = steps.length - 1;
        }
        else if (f <= 0)
        {
            f = 0;
            stepNo = 0;
        }
        else
        {
            stepNo = (int)(f * steps.length);
            f = (f - stepNo / (float)steps.length) * steps.length;
        }

        float   t = (float)f;
        float   t2 = t * t;
        float   t3 = t * t2;

        Matrix4f    step = steps[stepNo];

        Color3f c = new Color3f(
            step.m00 * t3 + step.m10 * t2 + step.m20 * t + step.m30,
            step.m01 * t3 + step.m11 * t2 + step.m21 * t + step.m31,
            step.m02 * t3 + step.m12 * t2 + step.m22 * t + step.m32);

        return c;
    }
}
