package org.bouncycastle.graphics.rt;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.util.*;

public class Hit
{
    public Renderable           obj;
    public float                t;
    public float                tTotal;
    public NormalCalculator     normCalc;
    public ColorCalculator      colorCalc;
    public TextureMapper        stMapper;
    public Shader               shader;
    public int                  timeSlice;

    public Ray                  i;
    Point3f                     loc = new Point3f();
    private Point3f             objLoc = new Point3f();
    boolean                     objLocSet = false;

    private float[]             st = new float[2];
    boolean                     stSet = false;

    private Color3f             color = new Color3f();

    private Vector3f            objN = new Vector3f();
    private Vector3f            n = new Vector3f();

    public boolean              subtractedSurface = false;

    private float               nDotI;
    private boolean             inside;
    private RandomTable         rand = new RandomTable();

    boolean                     normalSet = false;

    boolean                     disposed = false;

    Hit                  next = null;

    private static Hit   freeList = null;

    Hit()
    {
    }

    /**
     * return an instance of a Hit object, with passed in
     * values set.
     *
     * @param obj the object hit.
     * @param t the distance traveled to this hit.
     * @param colorCalc the base color calculator for the hit.
     * @param normCalc the normal calculator for the hit.
     * @param stMapper the mapper to get texture coordinates from
     * @param shader the shader to invoke for the hit
     * @return a free (possibly new) Hit object.
     */
    public static Hit getInstance(
        Renderable              obj,
        float                   t,
        ColorCalculator         colorCalc,
        NormalCalculator        normCalc,
        TextureMapper           stMapper,
        Shader                  shader)
    {
        Hit    hit;

        if (freeList != null)
        {
            hit = freeList;
            freeList = freeList.next;
        }
        else
        {
            hit = new Hit();
        }

        hit.t = t;
        hit.tTotal = 0;
        hit.obj = obj;
        hit.colorCalc = colorCalc;
        hit.normCalc = normCalc;
        hit.stMapper = stMapper;
        hit.shader = shader;
        hit.stSet = false;
        hit.objLocSet = false;
        hit.disposed = false;
        hit.normalSet = false;
        hit.subtractedSurface = false;
        hit.next = null;

        return hit;
    }

    public void dispose()
    {
        if (this.disposed)
        {
            throw new IllegalStateException("attempt to free Hit already freed");
        }

        this.disposed = true;
        this.next = freeList;

        freeList = this;
    }

    /**
     * return a visible light list for this hit for the given world
     */
    public LightDataList visibleLights(
        World world)
    {
        return world.visibleLights(this);
    }

    /**
     * return the recursion level of this hit.
     *
     * @return the recursion level.
     */
    public final int level()
    {
        return i.level;
    }

    /**
     * return the current shading normal.
     * 
     * @return the current shading normal.
     */
    public Vector3f N()
    {
        if (!normalSet)
        {
            calculateN();
        }

        return n;
    }

    /**
     * return the object space normal.
     * 
     * @return the object space normal.
     */
    public Vector3f oN()
    {
        if (!normalSet)
        {
            calculateN();
        }

        objN.normalize();

        return objN;
    }

    /**
     * calculate the current shading normal.
     */
    private void calculateN()
    {
        if (normalSet)
        {
            return;
        }

        normCalc.getN(this, this.objN);

        obj.transformNormal(objN, n);

        n.normalize();

        if (subtractedSurface)
        {
            n.scale(-1);
        }

        //
        // we need to reverse the normal if it facing away
        // from us. This can happen if we have hit the inside
        // of an object as a result of a csg subtract, or
        // same other clipping operation, or simply a polygon
        // with a backward facing normal.
        //
        // Note: we always assume a reversal means we are leaving
        // otherwise knowing how to handle refraction becomes
        // too hard.
        //
        if ((nDotI = (float)n.dot(i.dir)) > 0)
        {
            n.scale(-1);
            nDotI = -nDotI;
            inside = true;
        }
        else
        {
            inside = false;
        }

        normalSet = true;
    }

    /**
     * Perturb the current shading normal by adding some random
     * deviation perpendicular to the tip of the current shading
     * normal, scaled by the factor factor.
     * <p>
     * Note: this method changes the stored surface normal.
     *
     * @param factor the scale for the perturbation.
     */ 
    public void perturbN(
        float    factor)
    {
        float       deltaU, deltaV;
        float       u, v;
        int         indx;
        SampleMask  mask;
        Vector3f    tmp = new Vector3f();

        mask = SampleMask.getCircularSampleMask(7);

        indx = (int)(rand.getRandNum()
                    * mask.points.length);

        deltaU = mask.magnitude
                * (1 - 2 * rand.getRandNum());
        deltaV = mask.magnitude
                * (1 - 2 * rand.getRandNum());

        u = factor * (mask.points[indx].x + deltaU);
        v = factor * (mask.points[indx].y + deltaV);

        if (!normalSet)
        {
            calculateN();
        }

        Util.generatePerturbationVector(n, u, v, tmp);

        n.add(tmp);

        n.normalize();

        //
        // note: we don't change the value of inside if this
        // happens, we use only the unperturbed surface normal
        // to determine if we are inside an object or not.
        //
        if ((nDotI = (float)n.dot(i.dir)) > 0)
        {
            n.scale(-1);
            nDotI = -nDotI;
        }
    }

    /**
     * Perturb the current shading normal by adding a deviation
     * perpendicular to the tip of the current shading
     * normal, given by the vector (u, v, 0), after being aligned
     * so that it is transformed to being perpendicular to the
     * axis of the shading normal.
     * <p>
     * Note: this method changes the stored surface normal.
     *
     * @param factor the scale for the perturbation.
     */ 
    public void perturbN(
        float    u,
        float    v)
    {
        float       deltaU, deltaV;
        int         indx;
        SampleMask  mask;
        Vector3f    tmp = new Vector3f();

        if (!normalSet)
        {
            calculateN();
        }

        Util.generatePerturbationVector(n, u, v, tmp);

        n.add(tmp);

        n.normalize();

        //
        // note: we don't change the value of inside if this
        // happens, we use only the unperturbed surface normal
        // to determine if we are inside an object or not.
        //
        if ((nDotI = (float)n.dot(i.dir)) > 0)
        {
            n.scale(-1);
            nDotI = -nDotI;
        }
    }

    /**
     * Perturb the current shading normal by adding some vector
     * to it in object space and then transforming it to our
     * current world space.
     * <p>
     * Note: this method changes the stored surface normal.
     *
     * @param pTurb the vector to be used to perturb the shading
     *    normal.
     * @returns transformed perturbation vector in pTurb.
     */ 
    public void perturbN(
        Vector3f    pTurb)
    {
        if (!normalSet)
        {
            calculateN();
        }

        //
        // we have to make sure the perturbation stays the same
        // length, otherwise in large objects it will be lost
        // totally...
        //
        float        lenSqu, newLenSqu;

        lenSqu = pTurb.dot(pTurb);

        obj.transformNormal(pTurb, pTurb);

        newLenSqu = pTurb.dot(pTurb);

        pTurb.scale((float)Math.sqrt(lenSqu / newLenSqu));

        n.add(pTurb);

        n.normalize();

        //
        // note: we don't change the value of inside if this
        // happens, we use only the unperturbed surface normal
        // to determine if we are inside an object or not.
        //
        if ((nDotI = (float)n.dot(i.dir)) > 0)
        {
            n.scale(-1);
            nDotI = -nDotI;
        }
    }

    /**
     * return the incident ray direction vector.
     */
    public Vector3f I()
    {
        return i.dir;
    }

    /**
     * return the dot product of the shading normal N with the
     * direction of the incident ray I.
     *
     * @return N.I
     */
    public float NDotI()
    {
        if (!normalSet)
        {
            calculateN();
        }

        return nDotI;
    }

    /**
     * return the s texture coordinate.
     */
    public float S()
    {
        if (stSet)
        {
            return st[TextureMapper.S];
        }

        stSet = true;

        stMapper.getST(this.oP(), st);

        return st[TextureMapper.S];
    }

    /**
     * return the t texture coordinate.
     */
    public float T()
    {
        if (stSet)
        {
            return st[TextureMapper.T];
        }

        stSet = true;

        stMapper.getST(this.oP(), st);

        return st[TextureMapper.T];
    }

    /**
     * return the base color for the surface this hit is on.
     */
    public Color3f color()
    {
        colorCalc.getColor(this, this.color);

        return color;
    }

    /**
     * return the hit location in object coordinates
     */
    public Point3f oP()
    {
        if (!objLocSet)
        {
            objLocSet = true;

            this.objLoc.set(this.loc);
            this.obj.toObjectSpace(objLoc);
        }

        return this.objLoc;
    }

    /**
     * return the hit location in world coordinates
     */
    public Point3f P()
    {
        return loc;
    }

    /**
     * return the distance traveled to this hit.
     */
    public float distance()
    {
        return t;
    }

    /**
     * return the rough area covered by this object in screen space.
     */
    public float size()
    {
        return shader.world.calculateScreenSize(
                        this.t + this.tTotal, ((Bounded)obj).bBox());
    }

    /**
     * return the object we hit.
     */
    public Renderable object()
    {
        return obj;
    }

    /**
     * returns whether or not the incident ray is inside some
     * object.
     *
     * @return true, if inside, false otherwise.
     */
    public boolean isInside()
    {
        if (!normalSet)
        {
            calculateN();
        }

        return inside;
    }

    /**
     * set r so that it represents a continuation of the incident one,
     * i.e. has the same direction, but its origin is the point
     * the incident one has just hit.
     *
     * @param r the ray to contain the result.
     */
    public Ray continuationRay()
    {
        Ray    r = Ray.getInstance();

        r.dir.set(i.dir);
        r.org.set(loc);

        r.level = i.level + 1;
        r.done = false;
        r.theHit.obj = null;

        return r;
    }

    /**
     * return the reflected ray for the incident one
     * and the current shading normal
     *
     * @return the reflected ray.
     */
    public Ray reflectedRay()
    {
        if (!normalSet)
        {
            calculateN();
        }

        Ray r = Ray.getInstance();

        r.org.set(loc);
        r.dir.scaleAdd(-2 * nDotI, n, i.dir);

        r.level = i.level + 1;
        r.done = false;
        r.theHit.obj = null;
        r.theHit.tTotal = this.t + this.tTotal;

        return r;
    }

    /**
     * return a refracted ray for the incident one and the shading
     * normal. If the ray is internally reflected we return null.
     * The calculation is done using Snell's law.
     *
     * @param riIn    index of refraction for current material.
     * @param riOut    index of refraction for material we are entering.
     * @return the refracted ray.
     */
    public Ray refractedRay(
        float    riIn,
        float    riOut)
    {
        if (!normalSet)
        {
            calculateN();
        }

        //
        // do initial calculation.
        //
        Ray        r = Ray.getInstance();
        float      fact = riIn / riOut;

        r.org.set(loc);

        r.dir.scaleAdd(-nDotI, n, i.dir);
        r.dir.scale(fact);

        //
        // greater than or equal 1.0 means internal reflection
        //
        float        lenSqu;

        if ((lenSqu = r.dir.dot(r.dir)) < 1)
        {
            fact = (float)Math.sqrt(1.0 - lenSqu);

            r.dir.scaleAdd(-fact, n, r.dir);

            r.level = i.level + 1;
            r.done = false;
            r.theHit.obj = null;

            return r;
        }

        r.dispose();

            // We could return a reflected ray here instead.
            // Tradition says we shouldn't but then again...
        return null;
    }

    /**
     * return reflected/refracted rays calculated using the Fresnel
     * relationships and the fractional components they represent.
     * <p>
     * The fractions array contains Kr (fraction of light reflected)
     * and Kt (fraction of light transmitted) in that order, the
     * rays array will be returned with the corresponding rays.
     * <p>
     * Note: if Kr ot Kt are 0 the corresponding ray may be null. 
     *
     * @param riIn    index of refraction for current material.
     * @param riOut    index of refraction for material we are entering.
     */
/*
    public void fresnelValues(
        float    riIn,
        float    riOut,
        float[]  fractions,
        Ray[]    rays)
    {
        if (!normalSet)
        {
            calculateN();
        }

        throw new RuntimeException("Not implemented");
    }
*/
}
