package org.bouncycastle.graphics.rt;

import java.util.*;

public class HitList
{
    private Hit data;
    private Hit tail;
    private int size = 0;

    private HitList     next = null;
    private boolean     disposed = false;

    static  HitList   listList = null;

    private HitList()
    {
    }

    public static HitList getInstance()
    {
        HitList   l;

        if (listList == null)
        {
            l = new HitList();
        }
        else
        {
            l = listList;
            listList = listList.next;
        }

        l.disposed = false;
        l.data = null;
        l.tail = null;
        l.size = 0;

        return l;
    }

    public void dispose()
    {
        if (this.disposed)
        {
            throw new IllegalStateException("attempt to free HitList already freed");
        }

        this.disposed = true;
        this.next = listList;
        listList = this;

        releaseListOfHits(data);
    }

    public void add(
        Hit  newData)
    {
        if (tail == null)
        {
            data = tail = newData;
        }
        else
        {
            tail.next = newData;
        }

        tail = newData;
        newData.next = null;

        size++;
    }

    public void add(
        Ray                     r,
        Renderable              obj,
        float                   t,
        ColorCalculator         colorCalc,
        NormalCalculator        normCalc,
        TextureMapper           stMapper,
        Shader                  shader)
    {
        Hit hit = Hit.getInstance(obj, t, colorCalc, normCalc, stMapper, shader);

        hit.loc.scaleAdd(t, r.dir, r.org);

        this.add(hit);
    }

    public int size()
    {
        return size;
    }

    private void addListOfHits(
        Hit list)
    {
        Hit nextHit;

        for (Hit h = list; h != null; h = nextHit)
        {
            nextHit = h.next;
            this.add(h);
        }
    }

    private void releaseListOfHits(
        Hit list)
    {
        Hit nextHit;

        for (Hit h = list; h != null; h = nextHit)
        {
            nextHit = h.next;
            h.dispose();
        }
    }

    /**
     * merge the passed in list with the current one.
     * <p>
     * The passed in list is disposed of at the end of the operation
     */
    public void merge(
        HitList list)
    {
        Hit left = data;
        Hit right;

        if (list == null)
        {
            return;
        }

        right = list.data;

        //
        // buld up the final result list.
        //
        data = tail = null;
        size = 0;

        while (left != null && right != null)
        {
            if (left.t < right.t)
            {
                Hit next = left.next;
                
                this.add(left);
                
                left = next;
            }
            else
            {
                Hit next = right.next;
                
                this.add(right);
                
                right = next;
            }
        }

        if (left != null)
        {
            addListOfHits(left);
        }
        else
        {
            addListOfHits(right);
        }
    
        list.data = null;
        list.dispose();
    }

    /**
     * change this list of hits so that it represents the UNION of the
     * surfaces represented by this list and the argument list returning
     * the list if there is a result, null otherwise.
     * <p>
     * The argument list is disposed of at the end of the operation, as
     * the current list if the union operation results in a null set.
     */
    public HitList union(
        HitList    list)
    {
        //
        // if either list has an odd number of entries it means
        // we are already inside the object...
        //
        boolean inLeft = ((this.size & 1) == 1);
        boolean inRight = ((list != null) && ((list.size() & 1) == 1));

        Hit left = data;
        Hit right;

        if (list == null)
        {
            right = null;
        }
        else
        {
            right = list.data;
        }

        //
        // buld up the final result list.
        //
        data = tail = null;
        size = 0;

        while (left != null && right != null)
        {
            if (left.t < right.t)
            {
                Hit    hit = left;

                left = left.next;

                if (inRight)
                {        // already in surface - ignore
                    hit.dispose();
                }
                else
                {        // add to list.
                    this.add(hit);
                }
                inLeft = !inLeft;
            }
            else
            {
                Hit    hit = right;

                right = right.next;

                if (inLeft)
                {        // already in surface ignore
                    hit.dispose();
                }
                else
                {        // add to list
                    this.add(hit);
                }
                inRight = !inRight;
            }
        }

        if (left != null)
        {
            if (inRight)
            {
                releaseListOfHits(left);
            }
            else
            {
                addListOfHits(left);
            }
        }
        else if (right != null)
        {
            if (inLeft)
            {
                releaseListOfHits(right);
            }
            else
            {
                addListOfHits(right);
            }
        }

        if (list != null)
        {
            list.data = null;
            list.dispose();
        }

        if (this.data == null)
        {
            this.dispose();

            return null;
        }

        return this;
    }

    /**
     * change this list of hits so that it represents the INTERSECTION of the
     * surfaces represented by this list and the argument list returning
     * the list if there is a result, null otherwise.
     * <p>
     * The argument list is disposed of at the end of the operation, as
     * the current list if the union operation results in a null set.
     */
    public HitList intersect(
        HitList list)
    {
        //
        // if either list has an odd number of entries it means
        // we are already inside the object...
        //
        boolean inLeft = ((this.size & 1) == 1);
        boolean inRight = ((list != null) && ((list.size() & 1) == 1));

        Hit left = data;
        Hit right;

        if (list == null)
        {
            right = null;
        }
        else
        {
            right = list.data;
        }

        //
        // buld up the final result list.
        //
        data = tail = null;
        size = 0;

        while (left != null && right != null)
        {
            if (left.t < right.t)
            {
                Hit    hit = left;

                left = left.next;

                if (inRight)
                {
                    this.add(hit);
                }
                else
                {        // not in other surface - ignore
                    hit.dispose();
                }

                inLeft = !inLeft;
            }
            else
            {
                Hit    hit = right;

                right = right.next;

                if (inLeft)
                {
                    this.add(hit);
                }
                else
                {        // not in other surface - ignore
                    hit.dispose();
                }

                inRight = !inRight;
            }
        }

        if (left != null)
        {
            if (inRight)
            {
                addListOfHits(left);
            }
            else
            {
                releaseListOfHits(left);
            }
        }
        else if (right != null)
        {
            if (inLeft)
            {
                addListOfHits(right);
            }
            else
            {
                releaseListOfHits(right);
            }
        }

        if (list != null)
        {
            list.data = null;
            list.dispose();
        }

        if (this.data == null)
        {
            this.dispose();
            return null;
        }

        return this;
    }

    /**
     * change this list of hits so that it represents the SUBTRACTION of the
     * surfaces represented by this list and the argument list returning
     * the list if there is a result, null otherwise.
     * <p>
     * The argument list is disposed of at the end of the operation, as
     * the current list if the union operation results in a null set.
     */
    public HitList subtract(
        HitList list)
    {
        //
        // if either list has an odd number of entries it means
        // we are already inside the object...
        //
        boolean inLeft = ((this.size & 1) == 1);
        boolean inRight = ((list != null) && ((list.size() & 1) == 1));

        Hit left = data;
        Hit right;

        if (list == null)
        {
            right = null;
        }
        else
        {
            right = list.data;
        }

        //
        // buld up the final result list.
        //
        data = tail = null;
        size = 0;

        while (left != null && right != null)
        {
            if (left.t < right.t)
            {
                Hit    hit = left;

                left = left.next;

                if (inRight)
                {    
                    hit.dispose();
                }
                else
                {        // not in other surface - include
                    this.add(hit);
                }

                inLeft = !inLeft;
            }
            else
            {
                Hit    hit = right;

                right = right.next;

                if (inLeft)
                {
                    this.add(hit);
                }
                else
                {        // not in other surface - ignore
                    hit.dispose();
                }

                inRight = !inRight;
            }
        }

        if (left != null)
        {
            if (inRight)
            {
                releaseListOfHits(left);
            }
            else
            {
                addListOfHits(left);
            }
        }
        else if (right != null)
        {
            if (inLeft)
            {
                addListOfHits(right);
            }
            else
            {
                releaseListOfHits(right);
            }
        }

        if (list != null)
        {
            list.data = null;
            list.dispose();
        }

        if (this.data == null)
        {
            this.dispose();
            return null;
        }

        return this;
    }

    /**
     * return the first hit in the hit list
     */
    public Hit head()
    {
        return data;
    }

    public HitIterator iterator()
    {
        return HitIterator.getInstance(data);
    }
}
