package org.bouncycastle.graphics.rt;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.util.*;

public class MovingCamera
{
    Vector3f    up = new Vector3f(0, 1, 0);
    Vector3f    eye = new Vector3f(0, 0, 0);
    Vector3f    ref = new Vector3f(0, 0, -1);

    Matrix4f[]    rTrans = new Matrix4f[1];

    float        twist = 0;
    float        focalLength = -1;

    float        lensRadius = 0;
    float        lensFocalDistance = 0;

    boolean        orthographic = false;

    float        screenX = 1;
    float        screenY = 1;

    RandomTable    rand = new RandomTable();

    SampleMask    sampleMask = null;
    SampleMask    lensMask = null;

    boolean        jitter = true;
    boolean        moving = false;

    float        zNear = 0;
    float        zFar = Float.POSITIVE_INFINITY;

    public MovingCamera()
    {
        rTrans[0] = new Matrix4f(
                1, 0, 0, 0,
                0, 1, 0, 0,
                0, 0, 1, 0,
                0, 0, 0, 1);
    }

    private void calculateViewMatrix(
        Matrix4f    trans)
    {
        Vector3f    t, u, s;
        Matrix4f    view, m;
        double        val, vy, vz, sinval, cosval;

        /*
         * apply the twist
         */
        val = twist * Math.PI / 180.0;
        sinval = Math.sin(val);
        cosval = Math.cos(val);

        view = new Matrix4f(
            (float)cosval, (float)-sinval, 0, 0,
            (float)sinval, (float)cosval, 0, 0,
            0, 0, 1, 0,
            0, 0, 0, 1);

        /*
         * calculate the lookat
         */
        t = new Vector3f(eye);
        t.sub(ref);

        u = new Vector3f(up);

        t.normalize();

        u.normalize();

        vz = (float)t.dot(u);

        if (Math.abs(vz) == 1.0)
        {
            throw new IllegalStateException("getViewMatrix: up vector and direction of view are the same.");
        }

        vy = (float)Math.sqrt(1.0 - vz * vz);

        u.scaleAdd((float)-vz, t, u);
        u.scale((float)vy);

        s = new Vector3f();
        s.cross(u, t);

        m = new Matrix4f(
            s.x, u.x, t.x, 0,
            s.y, u.y, t.y, 0,
            s.z, u.z, t.z, 0,
            0, 0, 0, 1);

        view.mul(m, view);

        m.set(eye);

        view.mul(m, view);

        trans.mul(view, trans);
    }

    /**
     * set the "look at" details for the camera. The parameter eye
     * represents the position of the observer, ref represents the
     * position the observer is looking at, and up provides the
     * direction of world up.
     *
     * @param eye the eye position.
     * @param ref the point we are looking at.
     * @param up the direction of world up.
     */
    public void lookAt(
        Point3d        eye,
        Point3d        ref,
        Vector3d    up)
    {
        moving = false;

        this.eye = new Vector3f(eye);
        this.ref = new Vector3f(ref);
        this.up = new Vector3f(up);

        calculateViewMatrix(rTrans[0]);
    }

    /**
     * set the "look at" details for the camera. In this case the
     * camera is considered to be moving, and the eyes parameter
     * gives the eye positions for the frame, with the refs and
     * ups arrays providing the corresponding reference points and up
     * vectors.
     *
     * @param eyes the eye positions.
     * @param refs the corresponding reference positions.
     * @param ups the corresponding up vectors.
     */
    public void lookAt(
        Point3d[]    eyes,
        Point3d[]    refs,
        Vector3d[]    ups)
    {
        if (eyes.length != refs.length && refs.length != ups.length)
        {
            throw new IllegalArgumentException(
            "lookAt: eye, ref, and up arrays of different length.");
        }

        moving = true;

        Matrix4f[]    trans = new Matrix4f[eyes.length];

        for (int i = 0; i != eyes.length; i++)
        {
            this.eye = new Vector3f(eyes[i]);
            this.ref = new Vector3f(refs[i]);
            this.up = new Vector3f(ups[i]);

            trans[i] = new Matrix4f(rTrans[0]);

            calculateViewMatrix(trans[i]);
        }

        rTrans = Util.interpolate(Renderer.maxTimeSamples, trans);
    }

    /**
     * set a twist about the line of sight.
     *
     * @param the twist angle, in degrees.
     */
    public void setTwist(
        double    twist)
    {
        this.twist = (float)twist;

        calculateViewMatrix(rTrans[0]);
    }

    /**
     * In the normal situation with a real camera the diameter of the
     * lense opening means that the camera is only capable of rendering
     * objects sharply over a limited range of distances. The smaller
     * the opening the wider this range, or depth of field.
     * <p>
     * The default for a camera in our situation, is that it's a pin-hole
     * and has an infinite depth of field. Finite depth of field, however,
     * adds additional cues for depth perception, as well as adding an
     * aesthetic touch to the image.
     * <p>
     * To add depth of field to an image we have to specify, a <i>focal
     * distance</i>, a <i>focal length</i>, and the <i>f-stop</i>. The
     * <i>f-stop</i> represents the size of the aperture admitting light
     * into the camera, and the depth of field is determined by the
     * ratio of the <i>focal length</i> to the <i>f-stop</i> and the
     * <i>focal distance</i>.
     * <p>
     * Generaly the length units are millimeters, but in this case the
     * units used are irrelevant, the units for <i>focal length</i> just
     * need to be the same as for the <i>focal distance</i>.
     *
     * @param fStop    the f-stop for the camera.
     * @param focalLength the focal length of the lense.
     * @param focalDistance distance at which objects will appear sharpest.
     */
    public void setDepthOfField(
        double    fStop,
        double    focalLength,
        double    focalDistance)
    {
        if (fStop != 0)
        {
            this.lensRadius = (float)(focalLength / fStop / 2);
        }
        else
        {
            this.lensRadius = 0;
        }

        this.lensFocalDistance = (float)focalDistance;
        this.focalLength = (float)focalLength;

        Matrix4f    trans = rTrans[0];

        trans.setIdentity();

        trans.m00 = 1 / (float)focalLength;
        trans.m11 = 1 / (float)focalLength;
    }

    /**
     * Specifiy a perspective projection for the image, with the
     * near clipping plan at 0 and the far one at infinity.
     * 
     * @param fov the field of view.
     */
    public void perspective(
        double    fov)
    {
        Matrix4f    trans = rTrans[0];

        trans.setIdentity();

        float cot = (float)(1.0 / Math.tan(Math.PI / 360.0 * fov));

        trans.m00 = 1 / cot;
        trans.m11 = 1 / cot;

        this.zNear = 0;
        this.zFar = Float.POSITIVE_INFINITY;
    }

    /**
     * Specify a perspective viewing pyramid in world coordinates by
     * giving a field of view and the locations of the near(zNear) and
     * far(zFar) clipping planes in the z direction.
     *
     * @param fov the field of view.
     * @param zNear the near clipping plan.
     * @param zFar the far clipping plan.
     */
    public void perspective(
        double     fov,
        double    zNear,
        double    zFar)
    {
        if ((zFar - zNear) == 0.0)
        {
            throw new IllegalArgumentException(
              "perspective: near clipping plane same as far one.");
        }

        if (fov == 0.0 || fov == 180.0)
        {
            throw new IllegalArgumentException(
                "perspective: bad field of view passed.");
        }

        Matrix4f    trans = rTrans[0];

        trans.setIdentity();

        float cot = (float)(1.0 / Math.tan(Math.PI / 360.0 * fov));

        trans.m00 = 1 / cot;
        trans.m11 = 1 / cot;

        this.zNear = (float)zNear;
        this.zFar = (float)zFar;
    }

    /**
     * Define a three dimensional viewing box by giving the left,
     * right, bottom and top clipping plane locations and the distances
     * along the line of sight to the near and far clipping planes.
     */
    public void ortho(
        double     left,
        double    right,
        double    bottom,
        double    top,
        double    near,
        double    far)
    {
        if ((right - left) == 0.0)
        {
            throw new IllegalArgumentException(
                "ortho: left clipping plane same as right one.");
        }

        if ((top - bottom) == 0.0)
        {
            throw new IllegalArgumentException(
                "ortho: bottom clipping plane same as top one.");
        }

        if ((near - far) == 0.0)
        {
            throw new IllegalArgumentException(
                "ortho: near clipping plane same as far one.");
        }

        Matrix4f    trans = rTrans[0];

        trans.setIdentity();

        trans.m00 = (float)((right - left) / 2);
        trans.m11 = (float)((top - bottom) / 2);
        trans.m22 = -1;

        orthographic = true;
    }

    public void configureLenseMask(
        int    samples,
        boolean    jitter)
    {
        lensMask = SampleMask.getCircularSampleMask(samples);
        this.jitter = true;
    }

    public Ray generateRay(
        int         sampleNo,
        int         time,
        float       x,
        float       y)
    {
        Ray         r;
        Matrix4f    trans;

        r = Ray.getInstance();

        if (rTrans.length != 1)
        {
            trans = rTrans[time];
        }
        else
        {
            trans = rTrans[0];
        }

        if (orthographic)
        {
            r.dir.set(trans.m02, trans.m12, trans.m22);

            r.org.set(x, y, 0);
            trans.transform(r.org);
        }
        else
        {
            r.dir.set(x, y, -1);
            trans.transform(r.dir);

            r.org.set(trans.m03, trans.m13, trans.m23);
        }

        r.dir.normalize();

        //
        // add in depth of field affects
        //
        if (lensRadius != 0)
        {
            float        u, v;

            if (sampleNo > lensMask.points.length)
            {
                sampleNo %= lensMask.points.length;
            }
                
            u = lensRadius * lensMask.points[sampleNo].x;
            v = lensRadius * lensMask.points[sampleNo].y;

            if (jitter)
            {
                u += lensRadius * lensMask.magnitude
                     * (1.0 - rand.getRandNum() * 2);
                v += lensRadius * lensMask.magnitude
                      * (1.0 - rand.getRandNum() * 2);
            }

            r.org.x += u * trans.m00 + v * trans.m01;
            r.org.y += u * trans.m10 + v * trans.m11;
            r.org.z += u * trans.m20 + v * trans.m21;

            r.dir.scale(lensFocalDistance);

            r.dir.x -= u * trans.m00 + v * trans.m01;
            r.dir.y -= u * trans.m10 + v * trans.m11;
            r.dir.z -= u * trans.m20 + v * trans.m21;

            r.dir.normalize();
        }

        return r;
    }
}
