package org.bouncycastle.graphics.rt;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

public abstract class Renderable
    implements Traceable, ColorCalculator
{
    protected Shader        shader = null;

    protected Color3f       color = new Color3f(1, 1, 1);    // base color

    protected Matrix4f      objToRay = new Matrix4f();

    protected Matrix4f      rayToObj = new Matrix4f();

    protected boolean       selfShadowing = true;

    public Renderable()
    {
        objToRay.setIdentity();
        rayToObj.setIdentity();
    }

    /**
     * initialise a renderable object from a RMLReader.
     */
    protected void readObject(
        RMLReader   in)
        throws IOException
    {
        processParameters(in.parameters());

        if (in.token() != RMLReader.EMPTY_TAG)
        {
            Object      obj;

            while ((obj = in.readObject()) != null)
            {
                processObject(obj);
            }
        }
    }

    protected void processParameters(
        ParametersMap    params)
    {
        setColor(params.getColor("color", "1, 1, 1"));
    }

    protected void processObject(
        Object          obj)
    {
    }

    public void setShader(
        Shader shader)
    {
        this.shader = shader;
    }

    public Shader getShader()
    {
        return shader;
    }

    /**
     * set the base color for this object.
     */
    public void setColor(
        double    r,
        double    g,
        double    b)
    {
        color.set((float)r, (float)g, (float)b);
    }

    /**
     * set the base color for this object.
     */
    public void setColor(
        Color3f    color)
    {
        this.color.set(color);
    }

    /**
     * return the base color for the object.
     */
    public Color3f getColor()
    {
        return color;
    }
    
    /**
     * set whether or not we are self shadowing
     */
    public void setSelfShadowing(
        boolean selfShadowing)
    {
        this.selfShadowing = selfShadowing;
    }

    /**
     * return whether or not we are self shadowing.
     */
    public boolean isSelfShadowing()
    {
        return selfShadowing;
    }

    /**
     * return the base color for the object in the passed in Hit.
     */
    public void getColor(
        Hit     hit,
        Color3f c)
    {
        c.set(color);
    }
    
    /**
     * apply the given 4x4 transform to the object.
     */
    public void transform(
        Matrix4f   trans)
    {
        objToRay.mul(trans, objToRay);
        rayToObj.invert(objToRay);
    }

    /**
     * transform the passed vector form object space to ray space.
     */
    public void transformNormal(
        Vector3f    objN,
        Vector3f    rayN)
    {
        rayN.set(
            objN.x * rayToObj.m00
                + objN.y * rayToObj.m10 + objN.z * rayToObj.m20,
            objN.x * rayToObj.m01
                + objN.y * rayToObj.m11 + objN.z * rayToObj.m21,
            objN.x * rayToObj.m02
                + objN.y * rayToObj.m12 + objN.z * rayToObj.m22);
    }

    /**
     * transform the passed point to object space.
     */
    public void toObjectSpace(
        Point3f    p)
    {
        rayToObj.transform(p);
    }

    /**
     * transform the passed vector to object space.
     */
    public void toObjectSpace(
        Vector3f    v)
    {
        rayToObj.transform(v);
    }

    /**
     * return a new ray which represents an object space
     * version of the ray passed in.
     */
    public Ray createObjectSpaceRay(
        Ray r)
    {
        Ray or = Ray.getInstance();

        or.theHit.t = r.theHit.t;
        or.tMax = r.tMax;

        this.rayToObj.transform(r.dir, or.dir);
        this.rayToObj.transform(r.org, or.org);

        return or;
    }

    public abstract boolean isIntersected(Ray r);

    public abstract boolean isShadowing(Ray r);
}
