package org.bouncycastle.graphics.rt;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.objects.*;

public class Renderer
{
    protected static float FACTOR = (float)255.0;

    protected Camera      camera = new Camera();

    protected World       world = new World();

    protected int         screenW, screenH;
    protected int         midX, midY;
    protected int         startX, startY;
    protected int         endX, endY;
    protected int         currentY;
    protected float       pixelGap;                // gap between pixel centers
    protected int         hitMovingObj;
    protected int         currentTimeSlice;
    protected int         samplesPerPixel = 1;
    protected int         timeSamplesPerPixel;
    protected boolean     jitter = true;          // do we jitter?
    protected boolean     clamp = true;            // do we clamp pixel intensity

    public static final int        RGB = 0;
    public static final int        RGBA = 1;
    public static final int        RGBAO = 2;

    public static int        maxTimeSamples = 16;

    protected RandomTable    rand = new RandomTable();

    protected Background    background = new Background();

    //
    // variables that get used a lot in ray generation. We don't
    // want to keep setting these up all the time.
    //
    protected Vector3f    inc = new Vector3f();
    protected Matrix4f    trans;
    protected float        lensRadius, focalDistance;
    protected SampleMask    sampleMask;

    protected Renderer()
    {
        world.setCamera(camera);
    }

    /**
     * set the object environment to be used by this
     * renderer.
     *
     * @param the world we are trying to render.
     */
    public void setObjects(
        Illuminated      objs)
    {
        world.setObjects(objs);
    }

    /**
     * get "world" for this renderer.
     *
     * @return the environment this renderer is trying to handle.
     */
    public World getWorld()
    {
        return world;
    }

    /**
     * set the background for this renderer.
     *
     * @param background the background to use.
     */
    public void setBackground(
        Background    background)
    {
        this.background = background;
    }

    /**
     * get the background for this renderer.
     *
     * @return the renderer's background node.
     */
    public Background getBackground()
    {
        return background;
    }

    /**
     * set whether, or not, to clamp (clamping is the process
     * of truncating overilluminated pixels back to 1).
     *
     * @param clamp true if we should, false otherwise.
     */
    public void setClamp(
        boolean    clamp)
    {
        this.clamp = clamp;
    }

    /**
     * set the size of the image.
     *
     * @param xSize image size in x (width).
     * @param ySize image size in y (height).
     */
    public void setImageSize(
        int    xSize,
        int    ySize)
    {
        screenH = ySize;
        screenW = xSize;

        startX = -screenW / 2;
        endX = screenW / 2;

        startY = -screenH / 2;
        endY = screenH / 2;

        if (screenH < screenW)
        {
            pixelGap = (float)2.0 / screenH;
        }
        else
        {
            pixelGap = (float)2.0 / screenW;
        }

        currentY = endY;

        world.setScreenSize(screenW, screenH);
    }

    public void setWindow(
        int    x,
        int    y,
        int    xLen,
        int    yLen)
    {
        startX = x - screenW / 2;
        endX = x + xLen - screenW / 2;

        startY = y - screenH / 2;
        endY = y + yLen - screenH / 2;

        currentY = endY;
    }

    public void setSamplesPerPixel(
        int    samplesPerPixel)
    {
        this.samplesPerPixel = samplesPerPixel;
    }

    public void setJittering(
        boolean    jitter)
    {
        this.jitter = jitter;
    }

    public void setCamera(
        Camera    camera)
    {
        this.camera = camera;

        world.setCamera(camera);
    }

    public float getRandNum()
    {
        return rand.getRandNum();
    }

    public int getImageWidth()
    {
        return endX - startX;
    }

    public int getImageHeight()
    {
        return endY - startY;
    }

    public int getImageDepth()
    {
        return 24;
    }

    public void doRender(
        int        format,
        OutputStream    out)
        throws IOException
    {
        while (renderLineByLine(format, out))
        {
            ;
        }
    }

    private float clamp(
        float    val)
    {
        if (val > 1)
        {
            return 1;
        }

        return val;
    }

    public boolean renderLineByLine(
        int        format,
        OutputStream    out)
        throws IOException
    {
        Pixel            pixel = new Pixel();

        if (currentY == startY)
        {
            currentY = endY;
            return false;
        }

        if (currentY == endY)
        {
            camera.configureLenseMask(samplesPerPixel, jitter);
        }

        for (int j = startX; j != endX; j++)
        {
            Ray    r;
            float    fRed, fGreen, fBlue, fAlpha;

            fRed = fBlue = fGreen = fAlpha = 0;

            for (int sampleNo = 0;
                sampleNo != samplesPerPixel; sampleNo++)
            {
                float    k1, k2;

                currentTimeSlice =
                    (int)((((float)sampleNo
                            + rand.getRandNum())
                        * maxTimeSamples) /
                        samplesPerPixel);

                k1 = j;
                k2 = currentY;

                if (sampleMask != null)
                {
                    k1 = k1 + sampleMask.points[sampleNo].x;
                    k2 = k2 + sampleMask.points[sampleNo].y;
                    if (jitter)
                    {
                        k1 += sampleMask.magnitude
                          * (1 - rand.getRandNum() * 2);
                        k2 += sampleMask.magnitude
                          * (1 - rand.getRandNum() * 2);
                    }
                }

                r = camera.generateRay(
                        sampleNo,
                        currentTimeSlice,
                        k1 * pixelGap, k2 * pixelGap);

                r.trace(world, background, pixel);

                fRed += pixel.r;
                fGreen += pixel.g;
                fBlue += pixel.b;
                fAlpha += pixel.a;

                r.dispose();
            }

            fRed /= samplesPerPixel;
            fGreen /= samplesPerPixel;
            fBlue /= samplesPerPixel;
            fAlpha /= samplesPerPixel;

            if (clamp)
            {
                fRed = clamp(fRed);
                fGreen = clamp(fGreen);
                fBlue = clamp(fBlue);
                fAlpha = clamp(fAlpha);
            }

            if (format == RGB)
            {
                out.write((byte)(fRed * FACTOR));
                out.write((byte)(fGreen * FACTOR));
                out.write((byte)(fBlue * FACTOR));
            }
            else
            {
                out.write((byte)(fRed * FACTOR));
                out.write((byte)(fGreen * FACTOR));
                out.write((byte)(fBlue * FACTOR));
                out.write((byte)(fAlpha * FACTOR));
            }
        }

        currentY--;
        
        return true;
    }
}
