package org.bouncycastle.graphics.rt;

import  java.io.*;

import  javax.vecmath.*;

import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * The base class for Shaders.
 */
public abstract class Shader
{
    protected float         alpha = 1;
    protected Matrix4f      trans = new Matrix4f();
    protected World         world;

    public Shader(
        World   world)
    {
        trans.setIdentity();
        this.world = world;
    }

    /**
     * initialise a shader from a RMLReader.
     */
    protected void readObject(
        RMLReader   in)
        throws IOException
    {
        processParameters(in.parameters());

        if (in.token() != RMLReader.EMPTY_TAG)
        {
            Object      obj;

            while ((obj = in.readObject()) != null)
            {
                processObject(obj);
            }
        }
    }

    /**
     * apply the default parameters for a shader,
     * from the passed in parameters table.
     */
    protected void processParameters(
        ParametersMap    params)
    {
    }

    /**
     * apply default handling to the given object
     * as part of shader construction.
     */
    protected void processObject(
        Object          obj)
    {
        if (obj instanceof Transform)
        {
            this.transform((Matrix4f)obj);
        }
    }

    /**
     * set the alpha value (coverage value) for this object.
     */
    public void setAlpha(
        double    alpha)
    {
        this.alpha = (float)alpha;
    }

    /**
     * apply the given 4x4 transform to the shader space. Note with
     * shaders, the inverse of the transform is used.
     * <p>
     * This has the effect that when a shader is bound to a group of
     * objects, the shader space will move with the objects when we
     * apply the same transform as we do to the objects. It also
     * means that to make a texture smaller we scale down rather than
     * up, which works for this author...
     */
    public void transform(
        Matrix4f   trans)
    {
        Matrix4f   t = new Matrix4f(trans);

        t.invert();

        this.trans.mul(t, this.trans);
    }

    /**
     * transform the given vector into shader space
     */
    public void toShaderSpace(
        Vector3f   v)
    {
        this.trans.transform(v);
    }

    /**
     * transform the given point into shader space
     */
    public void toShaderSpace(
        Point3f   p)
    {
        this.trans.transform(p);
    }

    /**
     * utility method to allow a value to be clamped between 0 and 1.
     */
    protected float clamp(
        float    a)
    {
        return ((a) < 0 ? 0 : ((a) < 1 ? (a) : 1));
    }

    public abstract void shade(Hit hit, Pixel p);
}
