package org.bouncycastle.graphics.rt;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.image.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * A tile.
 * <p>
 * By default tiles are repeating. Any image format supported by the JDK
 * can be used together with the ".pix" format generated by the default
 * renderer.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;Tile image="tile.gif" repeating="false"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Boolean  "(true|false)"&gt;
 *     &lt;!ENTITY %String   "CDATA"&gt;
 *
 *     &lt;!ELEMENT Tile EMPTY&gt;
 *     &lt;!ATTLIST Tile
 *         image              %String;  #REQUIRED
 *         repeating          %Boolean; "true"&gt;
 * </pre>
 * </code>
 */
public class Tile
{
    protected static final int    U = 0;
    protected static final int    V = 1;

    private int[]       pix;
    private int         width;
    private int         height;

    private boolean     repeating = true;

    /**
     * default constructor, a tiling with a repeating image.
     *
     * @param im the image to be used for the tiling
     */
    public Tile(
        Image        im)
    {
        this(im, true);
    }

    /**
     * constructor, a tiling with an optionally repeating image.
     *
     * @param im the image to be used for the tiling
     * @param repeating true if we repeat, false if we don't.
     */
    public Tile(
        Image       im,
        boolean     repeating)
    {
        PixelGrabber    pg = new PixelGrabber(im, 0, 0, -1, -1, true);

        try
        {
            if (pg.grabPixels())
            {
                pix = (int[])pg.getPixels();
                width = pg.getWidth();
                height = pg.getHeight();
            }
            else
            {
                throw new RuntimeException("can't grab pixels");
            }
        }
        catch (InterruptedException e)
        {
            throw new RuntimeException(e.toString());
        }
    }

    public static Tile getInstance(
        RMLReader   in)
        throws IOException
    {
        ParametersMap   params = in.parameters();
        String          imageName = (String)params.get("image");
        boolean         repeating = true;
        Image           im;

        if (params.get("repeating") != null)
        {
            repeating = Boolean.valueOf((String)params.get("repeating")).booleanValue();
        }

        if (imageName != null)
        {
            Toolkit kit = Toolkit.getDefaultToolkit();

            if (imageName.endsWith(".pix"))
            {
                BufferedReader  rIn = new BufferedReader(new FileReader(imageName));
                String  type = rIn.readLine();
                String  dimensions = rIn.readLine();
                int[]   pix;
                int     width, height;

                if (type.equals("RGB"))
                {
                    StringTokenizer st = new StringTokenizer(dimensions, " ");

                    width = Integer.parseInt(st.nextToken());
                    height = Integer.parseInt(st.nextToken());

                    int depth = Integer.parseInt(st.nextToken());

                    pix = new int[width * height];

                    for (int i = 0; i != pix.length; i++)
                    {
                        pix[i] = (255 << 24) | (rIn.read() << 16) | (rIn.read() << 8) | (rIn.read());
                    }
                }
                else
                {
                    throw new IOException("Unknown image type");
                }

                im = kit.createImage(new MemoryImageSource(width, height, pix, 0, width));
            }
            else
            {
                im = kit.getImage(imageName);
            }
        }
        else
        {
            throw new IOException("No image name specified in tile.");
        }

        if (in.token() != RMLReader.EMPTY_TAG)
        {
            in.skipToEndTag(in.toString());
        }

        return new Tile(im, repeating);
    }

    public Pixel getPixel(
        float       s,
        float       t)
    {
        int    sI, tI, pixel;

        if (!repeating)
        {
            if ((s > 1.0) || (s < 0.0)
                || (t > 1.0) || (t < 0.0))
            {
                return null;
            }
        }

        if (s < 0.0)
        {
            sI = (int)(width * Math.abs(1.0 + s)) % width;
        }
        else
        {
            sI = (int)(width * s) % width;
        }

        if (t < 0.0)
        {
            tI = (int)(height * Math.abs(1.0 + t)) % height;
        }
        else
        {
            tI = (int)(height * t) % height;
        }

        pixel = pix[width * tI + sI];

        //
        // convert the int to a Pixel...
        //
        Pixel p = new Pixel();

        p.a = ((pixel >> 24) & 0xff) / 255.0f;
        p.r = ((pixel >> 16) & 0xff) / 255.0f;
        p.g = ((pixel >> 8) & 0xff) / 255.0f;
        p.b = ((pixel) & 0xff) / 255.0f;
        
        return p;
    }
}
