package org.bouncycastle.graphics.rt;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * A geometric transform.
 * <p>
 * Example:
 * <code>
 * <pre>
 *     &lt;Transform&gt;
 *        &lt;Scale value="0.1, 0.1, 0.1"/&gt;
 *        &lt;Translate value="-1.5, 1, 0"/&gt;
 *    &lt;/Transform&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float    "CDATA"&gt;
 *     &lt;!ENTITY %Point3f  "CDATA"&gt;
 *     &lt;!ENTITY %Vector3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT Transform (Matrix | Rotate | Scale | Translate)+ &gt;
 *     &lt;!ELEMENT Matrix (#PCDATA)&gt;
 *     &lt;!ELEMENT Rotate EMPTY&gt;
 *     &lt;!ATTLIST Rotate
 *          angle %Float;    "0"
 *          axis  %Vector3f; "0, 0, 1"&gt;
 *     &lt;!ELEMENT Scale EMPTY&gt;
 *     &lt;!ATTLIST Scale
 *          value %Vector3f; "1, 1, 1"&gt;
 *     &lt;!ELEMENT Translate EMPTY&gt;
 *     &lt;!ATTLIST Translate
 *          value %Point3f; "0, 0, 0"&gt;
 * </pre>
 * </code>
 */
public class Transform
    extends Matrix4f
{
    public Transform()
    {
        setIdentity();
    }

    public Transform(
        Transform   trans)
    {
        set(trans);
    }

    public static Transform getInstance(
        RMLReader   in)
        throws IOException
    {
        int         tok;
        Transform   trans = new Transform();

        while ((tok = in.readNext()) != RMLReader.EOF)
        {
            String  name = in.toString();

            if (tok == RMLReader.END_TAG && name.equals("Transform"))
            {
                break;
            }

            if (tok == RMLReader.EMPTY_TAG)
            {
                ParametersMap   params = in.parameters();

                if (name.equals("Scale"))
                {
                    Vector3f    v = params.getVector("value", "1, 1, 1");
                    trans.scale(v.x, v.y, v.z);
                }
                else if (name.equals("Translate"))
                {
                    Point3f    v = params.getPoint("value", "0, 0, 0");

                    trans.translate(v.x, v.y, v.z);
                }
                else if (name.equals("Rotate"))
                {
                    Vector3f    v = params.getVector("axis", "0, 0, 1");
                    float       theta = params.getFloat("angle", "0");

                    trans.rotate(v.x, v.y, v.z, theta);
                }
            }
            else if (tok == RMLReader.BEGIN_TAG)
            {
                if (name.equals("Matrix"))
                {
                    trans.transform(readMatrix(in));
                }
            }
        }

        return trans;
    }

    private static Matrix4f readMatrix(
        RMLReader   in)
        throws IOException
    {
        Matrix4f    m = new Matrix4f();

        if (in.readNext() != RMLReader.EOF)
        {
            StringTokenizer st = new StringTokenizer(in.toString(), ", \t");
    
            if (st.countTokens() == 16)
            {
                for (int i = 0; i != 4; i++)
                {
                    for (int j = 0; j != 4; j++)
                    {
                        m.setElement(i, j,
                                    Float.parseFloat(st.nextToken()));
                    }
                }
            }
            else
            {
                throw new IOException("badly formatted geometry tag");
            }
        }
        else
        {
            throw new IOException("unexpected EOF in geometry tag");
        }

        //
        // swallow the end tag.
        //
        if (in.readNext() == RMLReader.EOF)
        {
            throw new IOException("unexpected EOF in matrix tag");
        }

        return m;
    }

    public void transform(
        Matrix4f    m)
    {
        mul(m, this);
    }

    public void scale(
        double    x,
        double    y,
        double    z)
    {
        Matrix4f    m = new Matrix4f(
                    (float)x, 0, 0, 0,
                    0, (float)y, 0, 0,
                    0, 0, (float)z, 0,
                    0, 0, 0, 1);

        mul(m, this);
    }

    public void translate(
        double    x,
        double    y,
        double    z)
    {
        Matrix4f    m = new Matrix4f(
                    1, 0, 0, (float)x,
                    0, 1, 0, (float)y,
                    0, 0, 1, (float)z,
                    0, 0, 0, 1);

        mul(m, this);
    }

    /**
     * multiply in the necessary translations to rotate our world
     * so the z axis "becomes" the direction vector represented x, y and z.
     */
    public void alignZAxis(
        double        x,
        double        y,
        double        z)
    {
        float        l2, l3, sintheta, sinphi, costheta, cosphi;
        Matrix4f    trans = new Matrix4f();

        l2 = (float)Math.sqrt(x * x + z * z);
        l3 = (float)Math.sqrt(x * x + y * y + z * z);

        trans.setIdentity();

        if (l2 != 0.0)
        {
            sintheta = (float)x / l2;
            costheta = (float)z / l2;

            /*
             * Rotate about Y by theta
             */
            trans.m00 = trans.m22 = costheta;
            trans.m02 = sintheta;
            trans.m20 = -sintheta;
        }

        if (y != 0.0)
        {
            Matrix4f    mat = new Matrix4f();

            sinphi = -(float)y / l3;
            cosphi = l2 / l3;

            /*
             * Rotate about X by phi
             */
            mat.setIdentity();

            mat.m11 = mat.m22 = cosphi;
            mat.m12 = -sinphi;
            mat.m21 = sinphi;

            trans.mul(trans, mat);
        }

        mul(trans, this);
    }

    /**
     * rotate by theta degrees around the axis (x, y, z) 
     */
    public void rotate(
        double    x,
        double    y,
        double    z,
        double    theta)
    {
        AxisAngle4f     a = new AxisAngle4f((float)x, (float)y,
                                            (float)z, (float)(theta * Math.PI / 180));
        Matrix4f        m = new Matrix4f();

        m.setIdentity();
        m.set(a);

        mul(m, this);
    }
}
