package org.bouncycastle.graphics.rt.models;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;
import org.bouncycastle.graphics.rt.objects.*;

/**
 * a loader for .obj files (wavefront).
 */
public class ObjFile
    extends IndexedMesh
{
    private static HashMap    geometries = new HashMap();

    private static class GeometryHdl
    {
        short[][]       polys = null;
        Point3f[]       verts = null;
        short[][]       txtIndexes = null;
        Point3f[]       txtVerts = null;
        short[][]       normIndexes = null;
        Vector3f[]      normals = null;
    }

    public ObjFile(
        String    name,
        boolean   smooth)
        throws IOException
    {
        super(getGeometry(name).verts, getGeometry(name).polys, smooth);
    }

    private ObjFile()
    {
    }

    public static ObjFile getInstance(
        RMLReader    in)
        throws IOException
    {
        ObjFile of = new ObjFile();

        of.readObject(in);

        return of;
    }

    protected void processParameters(
        ParametersMap   params)
    {
        super.processParameters(params);

        try
        {
            GeometryHdl     geo = getGeometry((String)params.get("name"));

            init(geo.verts, geo.polys);

            if (geo.normals != null)
            {
                setVertexNormals(geo.normals, geo.normIndexes);
            }

            if (geo.txtVerts != null)
            {
                setVertexTextureCoordinates(geo.txtVerts, geo.txtIndexes);
            }
        }
        catch (IOException e)
        {
            throw new RuntimeException(e.toString());
        }
    }

    private static GeometryHdl getGeometry(
        String    name)
        throws IOException
    {
        GeometryHdl     handle = (GeometryHdl)geometries.get(name);

        if (handle != null)
        {
            return handle;
        }

        BufferedReader  fIn = new BufferedReader(new FileReader(name));
        String          line;
        ArrayList       verts = new ArrayList();
        ArrayList       normals = new ArrayList();
        ArrayList       normIndexes = new ArrayList();
        ArrayList       txtVerts = new ArrayList();
        ArrayList       txtIndexes = new ArrayList();
        ArrayList       facets = new ArrayList();

        handle = new GeometryHdl();

        //
        // read header
        //
        while ((line = fIn.readLine()) != null)
        {
            StringTokenizer    tok = new StringTokenizer(line);

            if (tok.countTokens() == 0)
            {                // blank line
                continue;
            }

            //
            // parse the line (first character is an
            // ascii code denoting what the line contains
            //
            String        code = tok.nextToken();

            switch (code.charAt(0))
            {
            case 'v':
                if (code.length() == 1)
                {
                    verts.add(new Point3f(
                      Float.valueOf(tok.nextToken()).floatValue(),
                      Float.valueOf(tok.nextToken()).floatValue(),
                      Float.valueOf(tok.nextToken()).floatValue()));
                }
                else if (code.charAt(1) == 'n')     // normals
                {
                    normals.add(new Vector3f(
                      Float.valueOf(tok.nextToken()).floatValue(),
                      Float.valueOf(tok.nextToken()).floatValue(),
                      Float.valueOf(tok.nextToken()).floatValue()));
                }
                else    // texture coordinate
                {
                    txtVerts.add(new Point3f(
                      Float.valueOf(tok.nextToken()).floatValue(),
                      Float.valueOf(tok.nextToken()).floatValue(),
                      Float.valueOf(tok.nextToken()).floatValue()));
                }
                break;
            case 'f':
                short[]     points = new short[tok.countTokens()];
                short[]     texVerts = new short[tok.countTokens()];
                short[]     norms = new short[tok.countTokens()];
                boolean     texVertsFound = false;
                boolean     normalsFound = false;
                
                for (int i = 0; i != points.length; i++)
                {
                    StringTokenizer st = new StringTokenizer(tok.nextToken(), "/", true);
                    String          index = st.nextToken();

                    points[i] = (short)(Integer.parseInt(index) - 1);
                    
                    if (st.countTokens() == 4)
                    {
                        texVertsFound = true;
                        normalsFound = true;
                        st.nextToken();
                        index = st.nextToken();
                        texVerts[i] = (short)(Integer.parseInt(index) - 1);
                        st.nextToken();
                        index = st.nextToken();
                        norms[i] = (short)(Integer.parseInt(index) - 1);
                    }
                    else if (st.countTokens() == 3)
                    {
                        normalsFound = true;
                        st.nextToken();
                        st.nextToken();
                        index = st.nextToken();
                        norms[i] = (short)(Integer.parseInt(index) - 1);
                    }
                    else if (st.countTokens() == 2)
                    {
                        texVertsFound = true;
                        st.nextToken();
                        index = st.nextToken();
                        texVerts[i] = (short)(Integer.parseInt(index) - 1);
                    }
                }

                if (points.length == 4)
                {
                    short[]   pnts = new short[3];

                    pnts[0] = points[0];
                    pnts[1] = points[1];
                    pnts[2] = points[2];

                    facets.add(pnts);

                    pnts = new short[3];

                    pnts[0] = points[2];
                    pnts[1] = points[3];
                    pnts[2] = points[0];

                    facets.add(pnts);

                    if (normalsFound)
                    {
                        short[]   nrms = new short[3];

                        nrms[0] = norms[0];
                        nrms[1] = norms[1];
                        nrms[2] = norms[2];

                        normIndexes.add(nrms);

                        nrms = new short[3];

                        nrms[0] = norms[2];
                        nrms[1] = norms[3];
                        nrms[2] = norms[0];

                        normIndexes.add(nrms);
                    }
                }
                else
                {
                    facets.add(points);
                    if (normalsFound)
                    {
                        normIndexes.add(norms);
                    }
                }
            }
        }

        handle.polys = (short[][])facets.toArray(new short[facets.size()][]);
        handle.verts = (Point3f[])verts.toArray(new Point3f[verts.size()]);

        if (txtVerts.size() != 0)
        {
            handle.txtIndexes = (short[][])txtIndexes.toArray(new short[facets.size()][]);
            handle.txtVerts = (Point3f[])txtVerts.toArray(new Point3f[txtVerts.size()]);
        }
        else
        {
            handle.txtVerts = null;
        }

        if (normals.size() != 0)
        {
            handle.normIndexes = (short[][])normIndexes.toArray(new short[facets.size()][]);
            handle.normals = (Vector3f[])normals.toArray(new Vector3f[normals.size()]);
        }
        else
        {
            handle.normals = null;
        }

        geometries.put(name, handle);

        return handle;
    }
}
