package org.bouncycastle.graphics.rt.models;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.objects.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a loader for OFF files.
 */
public class OffFile
    extends IndexedMesh
{
    private static HashMap  geometries = new HashMap();
    private static HashMap  faceColors = new HashMap();
    private static HashMap  vertexNormals = new HashMap();

    private static class GeometryHdl
    {
        Point3f[]       verts;
        short[][]       polys;
    }

    private static class FaceColorsHdl
    {
        int             nColors = 0;
        Color3f[]       colors = null;
        short[]         colorIndexes = null;
    }

    private static class VertexNormalsHdl
    {
        Vector3f[]      norms;
        short[][]       polys;
    }

    public OffFile(
        String    geometry)
        throws IOException
    {
        super(getGeometry(geometry).verts, getGeometry(geometry).polys);
    }

    public OffFile(
        String      geometry,
        boolean     smooth)
        throws IOException
    {
        super(getGeometry(geometry).verts, getGeometry(geometry).polys, smooth);
    }

    public OffFile(
        String      geometry,
        String      propFile,
        boolean     smooth)
        throws IOException
    {
        super(getGeometry(geometry).verts, getGeometry(geometry).polys, smooth);

        if (propFile.endsWith(".ipc"))
        {
            FaceColorsHdl   cols = getIpcFile(propFile);

            setColors(cols.colors, cols.colorIndexes);
        }
        else if (propFile.endsWith(".vnm"))
        {
            VertexNormalsHdl   norms = getVnmFile(propFile);

            setVertexNormals(norms.norms, norms.polys);
        }
    }

    private OffFile()
    {
    }

    public static IndexedMesh getInstance(
        RMLReader    in)
        throws IOException
    {
        OffFile of = new OffFile();

        of.readObject(in);

        return of;
    }

    protected void processParameters(
        ParametersMap   params)
    {
        super.processParameters(params);

        try
        {
            GeometryHdl     geo = getGeometry((String)params.get("geometry"));

            init(geo.verts, geo.polys);
        }
        catch (IOException e)
        {
            throw new RuntimeException(e.toString());
        }

        if (params.get("faceColors") != null)
        {
            try
            {
                FaceColorsHdl   cols = getIpcFile((String)params.get("faceColors"));

                setColors(cols.colors, cols.colorIndexes);
            }
            catch (IOException e)
            {
                throw new RuntimeException(e.toString());
            }
        }

        if (params.get("vertexNormals") != null)
        {
            try
            {
                VertexNormalsHdl   norms = getVnmFile((String)params.get("vertexNormals"));

                setVertexNormals(norms.norms, norms.polys);
            }
            catch (IOException e)
            {
                throw new RuntimeException(e.toString());
            }
        }
    }

    private static GeometryHdl getGeometry(
        String    name)
        throws IOException
    {
        GeometryHdl     handle = (GeometryHdl)geometries.get(name);

        if (handle != null)
        {
            return handle;
        }

        //
        // read in the previously unknown geometry
        //
        int                 nVerts, nPolys, nEdges;
        BufferedReader      fIn = new BufferedReader(new FileReader(name));
        String              line;
        StringTokenizer     tok;

        handle = new GeometryHdl();

        //
        // read header
        //

        line = fIn.readLine();
        tok = new StringTokenizer(line);
        if (tok.countTokens() != 3)
        {
            throw new IOException(
                "OffFile: bad header in off file.");
        }

        nVerts = Integer.parseInt(tok.nextToken());
        nPolys = Integer.parseInt(tok.nextToken());
        nEdges = Integer.parseInt(tok.nextToken());

        //
        // read in vertices.
        //
        handle.verts = new Point3f[nVerts];

        for (int i = 0; i != nVerts; i++)
        {
            line = fIn.readLine();
            tok = new StringTokenizer(line);
            if (tok.countTokens() != 3)
            {
                throw new IOException(
                    "OffFile: bad vertex line in off file.");
            }

            handle.verts[i] = new Point3f(
                   Float.valueOf(tok.nextToken()).floatValue(),
                   Float.valueOf(tok.nextToken()).floatValue(),
                   Float.valueOf(tok.nextToken()).floatValue());
        }

        //
        // read in the polygons
        //
        handle.polys = new short[nPolys][];

        for (int i = 0; i != nPolys; i++)
        {

            line = fIn.readLine();
            tok = new StringTokenizer(line);

            //
            // read in the vertex list for a polygon
            //
            int         nSides = Integer.parseInt(tok.nextToken());
            short[]     points = new short[nSides];

            for (int j = 0; (j != points.length)
                           && (tok.countTokens() != 0); j++)
            {
                points[j] = (short)(Integer.parseInt(
                            tok.nextToken()) - 1);
            }

            handle.polys[i] = points;
        }

        fIn.close();

        geometries.put(name, handle);

        return handle;
    }

    private FaceColorsHdl getIpcFile(
        String      name)
        throws IOException
    {
        FaceColorsHdl     handle = (FaceColorsHdl)faceColors.get(name);

        if (handle != null)
        {
            return handle;
        }

        //
        // read in the previously unknown indexed color file
        //
        BufferedReader      fIn = new BufferedReader(new FileReader(name));
        String              line;
        StringTokenizer     tok;

        handle = new FaceColorsHdl();

        //
        // read header
        //
        int    polyCount;

        line = fIn.readLine();
        tok = new StringTokenizer(line);
        if (tok.countTokens() != 2)
        {
            throw new IOException(
                "OffFile: bad header in off-ipc file.");
        }

        handle.nColors = Integer.parseInt(tok.nextToken());
        polyCount = Integer.parseInt(tok.nextToken());

        //
        // read in colors.
        //
        handle.colors = new Color3f[handle.nColors];

        for (int i = 0; i != handle.nColors; i++)
        {
            line = fIn.readLine();
            tok = new StringTokenizer(line);
            if (tok.countTokens() != 3)
            {
                throw new IOException(
                    "OffFile: bad color line in off-ipc file.");
            }

            handle.colors[i] = new Color3f(
                   Float.valueOf(tok.nextToken()).floatValue(),
                   Float.valueOf(tok.nextToken()).floatValue(),
                   Float.valueOf(tok.nextToken()).floatValue());
        }

        //
        // read in the polygons
        //
        handle.colorIndexes = new short[polyCount];

        for (int i = 0; i != handle.colorIndexes.length; i++)
        {
            line = fIn.readLine();
            handle.colorIndexes[i] = (short)(Integer.parseInt(line) - 1);
        }

        fIn.close();

        faceColors.put(name, handle);

        return handle;
    }

    /**
     *	read in the vertex normal information for a geometry file
     */
    private VertexNormalsHdl getVnmFile(
    	String	name)
        throws IOException
    {
        VertexNormalsHdl     handle = (VertexNormalsHdl)vertexNormals.get(name);

        if (handle != null)
        {
            return handle;
        }

        //
        // read in the previously unknown indexed color file
        //
        BufferedReader      fIn = new BufferedReader(new FileReader(name));
        String              line;
        StringTokenizer     tok;

        handle = new VertexNormalsHdl();

        //
        // read header
        //
        int    nNorms, nPolys, edgeCount;

        line = fIn.readLine();
        tok = new StringTokenizer(line);
        if (tok.countTokens() != 3)
        {
            throw new IOException(
                "OffFile: bad header in off-ipc file.");
        }

        nNorms = Integer.parseInt(tok.nextToken());
        nPolys = Integer.parseInt(tok.nextToken());
        edgeCount = Integer.parseInt(tok.nextToken());

        handle.norms = new Vector3f[nNorms];

    	for (int i = 0; i != nNorms; i++)
        {
            line = fIn.readLine();
            tok = new StringTokenizer(line);
            if (tok.countTokens() != 3)
            {
                throw new IOException(
                    "OffFile: bad normal line in .vnm file.");
            }

            handle.norms[i] = new Vector3f(
                   Float.valueOf(tok.nextToken()).floatValue(),
                   Float.valueOf(tok.nextToken()).floatValue(),
                   Float.valueOf(tok.nextToken()).floatValue());
        }
    
        handle.polys = new short[nPolys][];

        for (int i = 0; i != nPolys; i++)
        {
            line = fIn.readLine();
            tok = new StringTokenizer(line);

            //
            // read in the vertex list for a polygon
            //
            int         nSides = Integer.parseInt(tok.nextToken());
            short[]     points = new short[nSides];

            for (int j = 0; (j != points.length)
                           && (tok.countTokens() != 0); j++)
            {
                points[j] = (short)(Integer.parseInt(
                            tok.nextToken()) - 1);
            }

            handle.polys[i] = points;
        }
    
        fIn.close();

        vertexNormals.put(name, handle);

        return handle;
    }
}
