package org.bouncycastle.graphics.rt.objects;

import java.io.*;
import java.util.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * An aggregate object.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;Aggregate&gt;
 *          &lt;Transform&gt;
 *              &lt;Scale value="0.1, 0.1, 0.1"/&gt;
 *              &lt;Translate value="-1.5, 1, 0"/&gt;
 *          &lt;/Transform&gt;
 *          &lt;OffFile geometry="x29.geom" faceColors="x29.ipc"/&gt;
 *      &lt;/Aggregate&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Renderable  " Aggregate | Box | Cone | Cylinder | IndexedMesh | Polygon | Sphere | SuperQuadric | Torus | TriangularStrip "&gt;
 *     &lt;!ENTITY %Light       " DirectionalLight | DistantLight | PointLight "&gt;
 *
 *     &lt;!ELEMENT Aggregate (Transform | Shader | %Renderable; | %Light;)+ &gt;
 * </pre>
 * </code>
 */
public class Aggregate
    extends Renderable
    implements Illuminated
{
    Transform   trans = new Transform();

    private class ObjItem
    {
        Traceable    obj;
        ObjItem      next;
    }

    private class LightItem
    {
        Light        obj;
        LightItem    next;
    }

    private class ShaderItem
    {
        Shader       obj;
        ShaderItem   next;
    }

    private static TraceableIteratorImpl   traceItList = null;
    private static LightIteratorImpl       lightItList = null;
    private static ShaderIteratorImpl      shaderItList = null;

    private static class TraceableIteratorImpl
        implements TraceableIterator
    {
        ObjItem                 head;
        boolean                 disposed;
        TraceableIteratorImpl   next;

        private TraceableIteratorImpl()
        {
        }

        static TraceableIteratorImpl getInstance(
            ObjItem   head)
        {
            TraceableIteratorImpl  it;

            if (traceItList == null)
            {
                it = new TraceableIteratorImpl();
            }
            else
            {
                it = traceItList;
                traceItList = traceItList.next;
            }
 
            it.head = head;
            it.disposed = false;

            return it;
        }

        public boolean hasNext()
        {
            return (head != null);
        }

        public Traceable next()
        {
            ObjItem  ld = head;

            head = head.next;

            return ld.obj;
        }

        public void dispose()
        {
            if (disposed)
            {
                throw new IllegalStateException("attempt to dispose of renderable iterator twice");
            }

            disposed = true;

            this.next = traceItList;
            traceItList = this;
        }
    }

    private static class LightIteratorImpl
        implements LightIterator
    {
        LightItem               head;
        boolean                 disposed;
        LightIteratorImpl       next;

        private LightIteratorImpl()
        {
        }

        static LightIteratorImpl getInstance(
            LightItem   head)
        {
            LightIteratorImpl   it;

            if (lightItList == null)
            {
                it = new LightIteratorImpl();
            }
            else
            {
                it = lightItList;
                lightItList = lightItList.next;
            }
 
            it.head = head;
            it.disposed = false;

            return it;
        }

        public boolean hasNext()
        {
            return (head != null);
        }

        public Light next()
        {
            LightItem  ld = head;

            head = head.next;

            return ld.obj;
        }

        public void dispose()
        {
            if (disposed)
            {
                throw new IllegalStateException("attempt to dispose of light iterator twice");
            }

            disposed = true;

            this.next = lightItList;
            lightItList = this;
        }
    }

    private static class ShaderIteratorImpl
        implements ShaderIterator
    {
        ShaderItem              head;
        boolean                 disposed;
        ShaderIteratorImpl      next;

        private ShaderIteratorImpl()
        {
        }

        static ShaderIteratorImpl getInstance(
            ShaderItem   head)
        {
            ShaderIteratorImpl   it;

            if (shaderItList == null)
            {
                it = new ShaderIteratorImpl();
            }
            else
            {
                it = shaderItList;
                shaderItList = shaderItList.next;
            }
 
            it.head = head;
            it.disposed = false;

            return it;
        }

        public boolean hasNext()
        {
            return (head != null);
        }

        public Shader next()
        {
            ShaderItem  ld = head;

            head = head.next;

            return ld.obj;
        }

        public void dispose()
        {
            if (disposed)
            {
                throw new IllegalStateException("attempt to dispose of shader iterator twice");
            }

            disposed = true;

            this.next = shaderItList;
            shaderItList = this;
        }
    }

    private ObjItem         head = null;

    private LightItem       lights = null;

    private ShaderItem      shaders = null;

    private int             count = 0;

    BBox                    bBox = null;

    public Aggregate()
    {
    }

    public static Aggregate getInstance(
        RMLReader   in)
        throws IOException
    {
        Aggregate   a = new Aggregate();

        Object      obj;

        while ((obj = in.readObject()) != null)
        {
            if (obj instanceof Transform)
            {
                a.trans.transform((Transform)obj);
            }
            else if (obj instanceof Light)
            {
                a.add((Light)obj);
            }
            else if (obj instanceof Aggregate)
            {
                a.add((Aggregate)obj);
            }
            else if (obj instanceof Traceable)
            {
                a.add((Traceable)obj);
            }
            else if (obj instanceof Shader)
            {
                a.add((Shader)obj);
            }
        }

        return a;
    }

    public void add(
        Light    obj)
    {
        LightItem    item = new LightItem();

        obj.transform(trans);

        item.obj = obj;
        item.next = lights;
        lights = item;
    }

    public void add(
        Shader    obj)
    {
        ShaderItem    item = new ShaderItem();

        obj.transform(trans);

        item.obj = obj;
        item.next = shaders;
        shaders = item;
    }

    public void add(
        Traceable    obj)
    {
        ObjItem    item = new ObjItem();

        //
        // add in the current transform.
        // set the shader from the current one if none is currently
        // set and we have one.
        //
        if (trans != null && obj instanceof Renderable)
        {
            Renderable  r = (Renderable)obj;

            r.transform(trans);

            if ((shaders != null) && (r.getShader() == null))
            {
                r.setShader(shaders.obj);
            }
        }

        if (obj instanceof Partitionable)
        {
            if (bBox == null)
            {
                bBox = new BBox(((Partitionable)obj).bBox());
            }
            else
            {
                bBox.adjustFor(((Partitionable)obj).bBox());
            }
        }

        item.obj = obj;
        item.next = head;
        head = item;
        count++;
    }

    public void add(
        Aggregate    list)
    {
        TraceableIterator   traceIt = list.traceables();

        while (traceIt.hasNext())
        {
            this.add((Renderable)traceIt.next());
        }

        traceIt.dispose();

        LightIterator       lightIt = list.lights();

        while (lightIt.hasNext())
        {
            this.add(lightIt.next());
        }

        lightIt.dispose();

        ShaderIterator      shaderIt = list.shaders();

        while (shaderIt.hasNext())
        {
            this.add(shaderIt.next());
        }

        shaderIt.dispose();
    }

    public int objectCount()
    {
        return count;
    }

    public BBox bBox()
    {
        return bBox;
    }

    public boolean isIntersected(
        Ray    r)
    {
        for (ObjItem i = head; i != null; i = i.next)
        {
            if (i.obj.isIntersected(r))
            {
                if (r.hit.t < r.theHit.t)
                {
                    r.setTheHit();
                }
            }
        }

        return r.theHit.t != Float.MAX_VALUE;
    }

    public HitList intersections(
        Ray    r)
    {
        HitList list = null;

        for (ObjItem i = head; i != null; i = i.next)
        {
            HitList hits = i.obj.intersections(r);

            if (hits != null)
            {
                if (list == null)
                {
                    list = hits;
                }
                else
                {
                    list.merge(hits);
                }
            }
        }

        return list;
    }

    public boolean isShadowing(
        Ray    r)
    {
        for (ObjItem i = head; i != null; i = i.next)
        {
            if (i.obj.isShadowing(r))
            {
                if (r.hit.t < r.tMax)
                {
                    r.setTheHit();
                    return true;
                }
            }
        }

        return false;
    }

    public LightDataList visibleLights(
        Hit    hit)
    {
        return visibleLights(hit, this);
    }

    public LightDataList visibleLights(
        Hit             hit,
        Illuminated     world)
    {
        LightDataList    ldList = LightDataList.getInstance();

        for (LightItem l = lights; l != null; l = l.next)
        {
            Light        light = l.obj;

            ldList.add(light.getLightData(hit, world));
        }

        return ldList;
    }

    public TraceableIterator traceables()
    {
        return TraceableIteratorImpl.getInstance(head);
    }

    public LightIterator lights()
    {
        return LightIteratorImpl.getInstance(lights);
    }

    public ShaderIterator shaders()
    {
        return ShaderIteratorImpl.getInstance(shaders);
    }
}
