package org.bouncycastle.graphics.rt.objects;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.math.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * Based on, "ray tracing implicit surfaces resulting from the summation
 * of bounded volume polynomial functions", By David Tonnesen, Tech Report
 * no 89003, Rensselear Design Research Center, Rensselaer Polytechnic
 * Institute, Troy, Nw York, 12180.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;Blobby color="1, 0, 0" threshold="0.5"&gt;
 *        &lt;Metaball center="-1.0, 0.0, -4.0" radius="1.0" strength="0.8"/&gt;
 *        &lt;Metaball center="0.0, 0.0, -4.0" radius="1.0" strength="0.8"/&gt;
 *        &lt;Metaball center="1.0, 0.0, -4.0" radius="1.0" strength="0.8"/&gt;
 *      &lt;/Blobby&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Float   "CDATA"&gt;
 *
 *     &lt;!ELEMENT Blobby (Metaball)+ &gt;
 *     &lt;!ATTLIST Blobby
 *         color        %Color3f; "1, 1, 1"
 *         threshold    %Float;   "1"&gt;
 * </pre>
 * </code>
 */
public class Blobby
    extends Bounded
    implements NormalCalculator
{
    double              tolerance = 0.001;

    TextureMapper       stMapper = null;
    BvTree              tree;
    Metaball[]          mBalls;
    ArrayList           mBArray = new ArrayList();

    static class MBHit
    {
        boolean     inside;
        boolean     exit;
        Metaball    mb;
        double      t;
        double[]    coef = new double[5];

        MBHit       next;
        boolean     disposed = false;

        private static MBHit    hitList = null;

        static MBHit getInstance(
            Metaball    mb,
            double      t,
            boolean     inside,
            boolean     exit)
        {
            MBHit     hit = hitList;

            if (hit == null)
            {
                hit = new MBHit();
            }
            else
            {
                hitList = hitList.next;
            }

            hit.mb = mb;
            hit.t = t;
            hit.inside = inside;
            hit.exit = exit;
            hit.disposed = false;
            hit.next = null;

            return hit;
        }

        /**
         * calculate the polynomial for this metaball.
         */
        void calcPoly(
            Ray         r,
            double      rDirSqu)
        {
            double      a, b, c;
        
            //
            // precalculation. stage 2
            //
            a = rDirSqu / mb.radSqu;
            b = 2 * mb.v.dot(r.dir) / mb.radSqu;
            c = mb.v.dot(mb.v) / mb.radSqu - 1.0;           // c - 1.0
        
            //
            // coefficients:
            //
            //    a^2.t^4 + 2a.b.t^3 + (2a.c - 2a + b^2)t^2 +
            // (2b.c - 2b)t + (c^2 - 2c + 1) = 0
            //
        
            coef[4] = mb.s * a * a;
            coef[3] = mb.s * 2 * a * b;
            coef[2] = mb.s * (b * b + 2 * a * c);
            coef[1] = mb.s * 2 * b * c;
            coef[0] = mb.s * c * c;
        }

        void dispose()
        {
            if (this.disposed)
            {
                throw new IllegalStateException("attempt to free metaball hit twice");
            }

            this.disposed = true;

            this.next = hitList;
            hitList = this;
        }
    }

    private class Function
    {
        double[]            coef = new double[5];
        QuarticSturmSolver  solver = new QuarticSturmSolver();
        int                 atStart, atEnd;
        double              xStart, xEnd;
        double              threshold;
        
        Function(
            double  threshold)
        {
            this.threshold = -threshold;
        }

        void init()
        {
            coef[0] = threshold;
            coef[1] = coef[2] = coef[3] = coef[4] = 0;
        }

        void startAt(
            double  x)
        {
            xEnd = x;
        }

        boolean hasNoRoots(
            double  x)
        {
            solver.init(coef);

            xStart = xEnd;
            atStart = solver.changesAtX(xStart);

            atEnd = solver.changesAtX(x);
            xEnd = x;

            return ((atStart - atEnd) == 0);
        }

        boolean findRoot(
            double[]    roots)
        {
            return solver.findRoot(xStart, xEnd, roots);
        }

        void addPoly(
            double[]    f)
        {
            coef[4] += f[4];
            coef[3] += f[3];
            coef[2] += f[2];
            coef[1] += f[1];
            coef[0] += f[0];
        }

        void subPoly(
            double[]    f)
        {
            coef[4] -= f[4];
            coef[3] -= f[3];
            coef[2] -= f[2];
            coef[1] -= f[1];
            coef[0] -= f[0];
        }
    }

    private Function    f = new Function(1);

    public Blobby()
    {
    }

    public static Blobby getInstance(
        RMLReader   in)
        throws IOException
    {
        Blobby  b = new Blobby();

        b.readObject(in);

        b.mBalls = (Metaball[])b.mBArray.toArray(new Metaball[b.mBArray.size()]);
        b.tree = new BvTree(b.mBalls);
        b.mBArray = null;       // free up.

        return b;
    }

    protected void processParameters(
        ParametersMap   params)
    {
        super.processParameters(params);

        this.f = new Function(params.getFloat("threshold", "1"));
    }

    protected void processObject(
        Object  obj)
    {
        if (obj instanceof Metaball)
        {
            ((Metaball)obj).setParent(this);
            mBArray.add(obj);
        }
        else
        {
            super.processObject(obj);
        }
    }

    private MBHit   hits = null;    // our list of hits...

    /**
     * merge the passed in list with the current one.
     */
    void mergeHits(
        MBHit list)
    {
        if (list == null)
        {
            return;
        }

        if (hits == null)
        {
            hits = list;
            return;
        }

        MBHit   left = hits;
        MBHit   right = list;

        MBHit   tail = null;
        
        hits = null;

        if (left.t < right.t)
        {
            hits = tail = left;
            left = left.next;
        }
        else
        {
            hits = tail = right;
            right = right.next;
        }

        while (left != null && right != null)
        {
            if (left.t < right.t)
            {
                tail.next = left;
                left = left.next;
            }
            else
            {
                tail.next = right;
                right = right.next;
            }

            tail = tail.next;
        }

        if (left != null)
        {
            tail.next = left;
        }
        else
        {
            tail.next = right;
        }
    }

    public void getN(
        Hit         hit,
        Vector3f    n)
    {
        double  x, y, z, fact;
        Point3f loc = hit.oP();

        x = y = z = 0;

        for (int i = 0; i != mBalls.length; i++)
        {
            Metaball    mb = mBalls[i];

            if (mb.inside)
            {
                n.sub(loc, mb.center);
                fact = mb.s * ((2 * n.dot(n)) / mb.radSqu - 2) / mb.radSqu * 2;

                x += fact * n.x;

                y += fact * n.y;

                z += fact * n.z;
            }
        }

        n.set((float)x, (float)y, (float)z);
    }

    public boolean isObjectIntersected(
        Ray r)
    {
        Ray         nr = this.createObjectSpaceRay(r);

        hits = null;

        tree.intersections(nr);

	    if (hits != null)
        {
		    //
		    // do some useful precalculation...
		    //
		    double      rDirSqu = nr.dir.dot(nr.dir);

            f.init();

            for (int i = 0; i != mBalls.length; i++)
            {
                mBalls[i].inside = false;
            }

		    //
		    // add the density functions for any blobbys we are inside
		    //
		    boolean     onSurf = false;

            for (MBHit h = hits; h != null; h = h.next)
            {
                Metaball    mb = h.mb;

			    h.calcPoly(nr, rDirSqu);

                if (h.inside)
                {
                    onSurf = true;
                    f.addPoly(h.coef);
                }
            }

            //
            // find our closest hit.
            //
            MBHit   ht = hits;

            if (onSurf)
            {
                f.startAt(tolerance);
            }
            else
            {
                Metaball    mb = ht.mb;

                f.startAt(ht.t);

                f.addPoly(ht.coef);
                ht.mb.inside = true;

                ht = ht.next;
            }

            while (ht != null)
            {
                Metaball    mb = ht.mb;

                if (f.hasNoRoots(ht.t))
                {
                    if (ht.exit)
                    {
                        mb.inside = false;
                        f.subPoly(ht.coef);
                    }
                    else
                    {
                        mb.inside = true;
                        f.addPoly(ht.coef);
                    }
                }
                else
                {
                    double[]  roots = new double[4];

                    f.findRoot(roots);

                    r.setHit(this, (float)roots[0], this, this, stMapper, this.getShader());

                    MBHit   nextH;

                    for (MBHit h = hits; h != null; h = nextH)
                    {
                        nextH = h.next;
                        h.dispose();
                    }

                    return true;
                }
              
                ht = ht.next;
            }

            MBHit   nextH;

            for (MBHit h = hits; h != null; h = nextH)
            {
                nextH = h.next;
                h.dispose();
            }
		}

        nr.dispose();

        return false;
    }

    public boolean isIntersected(
        Ray r)
    {
        if (!isObjectIntersected(r))
        {
            return false;
        }

        return true;
    }

    public HitList intersections(
        Ray r)
    {
        return null;
    }

    public boolean isShadowing(
        Ray r)
    {
        if (!isObjectIntersected(r))
        {
            return false;
        }

        return false; // (r.hit.t < r.tMax);
    }
}
