package org.bouncycastle.graphics.rt.objects;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a box, in canonical form a box with one vertex at the origin, the
 * opposite vertex at the point (1, 1, 1).
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;Box color="1, 1, 1" vertex1="-1, -1, -1" vertex2="1, 1, 1"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Point3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT Box EMPTY&gt;
 *     &lt;!ATTLIST Box
 *         color   %Color3f;  "1, 1, 1"
 *         vertex1 %Point3f; "0, 0, 0"
 *         vertex2 %Point3f; "1, 1, 1"&gt;
 * </pre>
 * </code>
 */
public class Box
    extends SimpleObject
{
    public static final float TOL = (float)0.0001;

    private static class BoxNormalCalculator 
            implements NormalCalculator
    {
        float   x, y, z;

        BoxNormalCalculator(
            float    x,
            float    y,
            float    z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }

        public void getN(
            Hit         hd,
            Vector3f    n)
        {
            n.set(x, y, z);
        }
    }

    public static final BoxNormalCalculator NXFACE
                = new BoxNormalCalculator(-1, 0, 0);
    public static final BoxNormalCalculator NYFACE
                = new BoxNormalCalculator(0, -1, 0);
    public static final BoxNormalCalculator NZFACE
                = new BoxNormalCalculator(0, 0, -1);
    public static final BoxNormalCalculator PXFACE
                = new BoxNormalCalculator(1, 0, 0);
    public static final BoxNormalCalculator PYFACE
                = new BoxNormalCalculator(0, 1, 0);
    public static final BoxNormalCalculator PZFACE
                = new BoxNormalCalculator(0, 0, 1);

    static BBox    canonicalBBox = new BBox(0, 0, 0, 1, 1, 1);

    BBox        bb;

    public Box(
        Point3d        p1,
        Point3d        p2)
    {
        init(p1.x, p1.y, p1.z, p2.x, p2.y, p2.z);
    }

    private Box()
    {
    }

    public static Box getInstance(
        RMLReader       in)
        throws IOException
    {
        Box b = new Box();

        b.readObject(in);

        return b;
    }

    public Box(
        Point3f        p1,
        Point3f        p2)
    {
        init(p1.x, p1.y, p1.z, p2.x, p2.y, p2.z);
    }

    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        Point3f p1 = params.getPoint("vertex1", "0, 0, 0");
        Point3f p2 = params.getPoint("vertex2", "1, 1, 1");

        init(p1.x, p1.y, p1.z, p2.x, p2.y, p2.z);
    }

    private void init(
        double    x1,
        double    y1,
        double    z1,
        double    x2,
        double    y2,
        double    z2)
    {
        double        minX, minY, minZ, maxX, maxY, maxZ;

        setSelfShadowing(false);

        if (x1 > x2)
        {
            maxX = x1;
            minX = x2;
        }
        else
        {
            maxX = x2;
            minX = x1;
        }

        if (y1 > y2)
        {
            maxY = y1;
            minY = y2;
        }
        else
        {
            maxY = y2;
            minY = y1;
        }

        if (z1 > z2)
        {
            maxZ = z1;
            minZ = z2;
        }
        else
        {
            maxZ = z2;
            minZ = z1;
        }

        this.setBBox(canonicalBBox);

        Transform   trans = new Transform();

        trans.scale(maxX - minX, maxY - minY, maxZ - minZ);

        trans.translate(minX, minY, minZ);

        this.transform(trans);

        tolerance = TOL;
    }

    protected boolean isObjectIntersected(
        Ray    r)
    {
        NormalCalculator    ent, ext;
        float            dir, org, t, b1, b2;

        t1 = 0;
        t2 = Float.MAX_VALUE;

        dir = r.dir.x;
        org = r.org.x;
        if (dir != 0)
        {
            if (dir < 0)
            {
                if (org < 0)
                {
                    return false;
                }
                b2 = -org;
                b1 = 1 - org;
                enNorm = PXFACE;
                exNorm = NXFACE;
            }
            else
            {
                if (org > 1)
                    return false;
                b2 = 1 - org;
                b1 = -org;
                exNorm = NXFACE;
                enNorm = PXFACE;
            }
                
            t2 = b2 / dir;
            t1 = b1 / dir;
            if (t1 < 0)
            {
                t1 = 0;
            }
        }
        else if (org < 0 || org > 1)
        {
            return false;
        }

        dir = r.dir.y;
        org = r.org.y;
        if (dir != 0)
        {
            if (dir < 0)
            {
                if (org < 0)
                {
                    return false;
                }
                b2 = -org;
                b1 = 1 - org;
                ent = PYFACE; 
                ext = NYFACE;
            }
            else
            {
                if (org > 1)
                {
                    return false;
                }
                b2 = 1 - org;
                b1 = -org;
                ent = NYFACE; 
                ext = PYFACE;
            }
                
            if ((t = b2 / dir) < t1)
            {
                return false;
            }

            if (t < t2)
            {
                t2 = t;
                exNorm = ext;
            }

            if ((t = b1 / dir) > t2)
            {
                return false;
            }

            if (t > t1)
            {
                t1 = t;
                enNorm = ent;
            }
        }
        else if (org < 0 || org > 1)
        {
            return false;
        }

        dir = r.dir.z;
        org = r.org.z;
        if (dir != 0)
        {
            if (dir < 0)
            {
                if (org < 0)
                {
                    return false;
                }
                b2 = -org;
                b1 = 1 - org;
                ent = PZFACE; 
                ext = NZFACE;
            }
            else
            {
                if (org > 1)
                {
                    return false;
                }
                b2 = 1 - org;
                b1 = -org;
                ent = NZFACE; 
                ext = PZFACE;
            }
                
            if ((t = b2 / dir) < t1)
            {
                return false;
            }

            if (t < t2)
            {
                t2 = t;
                exNorm = ext;
            }

            if ((t = b1 / dir) > t2)
            {
                return false;
            }

            if (t > t1)
            {
                t1 = t;
                enNorm = ent;
            }
        }
        else if (org < 0 || org > 1)
        {
            return false;
        }

        if (t1 < tolerance && t2 < tolerance)
        {
            return false;
        }

        enShade = exShade = this.getShader();

        return true;
    }
}
