package org.bouncycastle.graphics.rt.objects;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;

/**
 * A BvTree.
 * <p>
 * We use this as it guarantees that an object will only be
 * tested once per ray, without the need for mail boxes and the like.
 * It's other advantage is, because it does a 3 way split at each level,
 * it doesn't result in duplicated objects.
 */
public class BvTree
    extends Renderable
    implements Illuminated
{
    static final byte   X           = 0;
    static final byte   Y           = 2;
    static final byte   Z           = 4;
    static final byte   EMPTY       = 5;
    static final byte   SPLITABLE   = 6;
    static final byte   LEAF        = 7;

    static final byte   RIGHT       = 0;
    static final byte   LEFT        = 1;
    static final byte   BOTH        = 2;
    static final byte   MASK        = 3;

    private abstract class TreeNode
    {
        byte    type;
        BBox    bBox;

        TreeNode(
            byte    type)
        {
            this.type = type;
        }

        /**
         * return whether or not we contain a given point.
         */
        boolean contains(
            Point3f p)
        {
            if (type == EMPTY)
            {
                return false;
            }

            return bBox.contains(p);
        }

        /**
         * return whether or not we are missed by a given ray.
         */
        boolean isMissedBy(
            Ray r)
        {
            if (type == EMPTY)
            {
                return true;
            }

            return bBox.isMissedBy(r);
        }

        abstract HitList intersections(Ray r);

        abstract boolean isIntersected(Ray r);

        abstract boolean isShadowing(Ray r);
    }

    private class LeafNode
        extends TreeNode
    {
        BranchNode          parent;
        int                 start;
        int                 end;

        LeafNode(
            BranchNode      parent,
            int             start,
            int             end)
        {
            super(((end - start) > 1) ? SPLITABLE
                            : ((end == start) ? EMPTY : LEAF));

            this.parent = parent;
            this.start = start;
            this.end = end;

            if (this.type != EMPTY)
            {
                this.bBox = new BBox(objects[start].bBox());

                for (int i = start + 1; i != end; i++)
                {
                    this.bBox.adjustFor(objects[i].bBox());
                }
            }
        }

        private int adjustStats(
            int      axis,
            float    min,
            float    max,
            float    split,
            int[]    stats)
        {
            if (min > split)
            {
                stats[RIGHT]++;

                return RIGHT;
            }
            else if (max < split)
            {
                stats[LEFT]++;

                return LEFT;
            }
            else
            {
                stats[BOTH]++;

                return BOTH;
            }
        }

        private BranchNode splitNode(
            byte    axis,
            float   splitValue,
            int     left,
            int     right,
            int     start,
            int     end)
        {
            int     l = start;
            int     r = end - 1;

            while (l < r)
            {
                int    lCode = (partition[l] >> axis) & MASK; 
                int    rCode = (partition[r] >> axis) & MASK; 

                while (rCode == RIGHT && l != r)
                {
                    rCode = (partition[--r] >> axis) & MASK; 
                }

                while (lCode != RIGHT && l != r)
                {
                    lCode = (partition[++l] >> axis) & MASK; 
                }

                Partitionable tmp = objects[l];

                objects[l] = objects[r];
                objects[r] = tmp;

                byte        code = partition[l];

                partition[l] = partition[r];
                partition[r] = code;
            }

            l = start;

            while (l < r)
            {
                int    lCode = (partition[l] >> axis) & MASK; 
                int    rCode = (partition[r] >> axis) & MASK; 

                while (rCode == BOTH && l != r)
                {
                    rCode = (partition[--r] >> axis) & MASK; 
                }

                while (lCode != BOTH && l != r)
                {
                    lCode = (partition[++l] >> axis) & MASK; 
                }

                Partitionable tmp = objects[l];

                objects[l] = objects[r];
                objects[r] = tmp;

                byte        code = partition[l];

                partition[l] = partition[r];
                partition[r] = code;
            }

            BranchNode  parent = new BranchNode(axis, this.bBox);

            parent.splitValue = splitValue;
            parent.left = new LeafNode(parent, start, start + left);
            parent.mid = new LeafNode(parent, start + left, end - right);
            parent.right = new LeafNode(parent, end - right, end);

            return parent;
        }

        /**
         * split this leaf node, yielding a branch node and two
         * more leaf nodes.
         */
        private TreeNode split()
        {
            float   xSplit = (bBox.minX + bBox.maxX) / 2,
                    ySplit = (bBox.minY + bBox.maxY) / 2,
                    zSplit = (bBox.minZ + bBox.maxZ) / 2;

            int[]   xStats = new int[3],
                    yStats = new int[3],
                    zStats = new int[3];

            for (int i = start; i != end; i++)
            {
                BBox    b = objects[i].bBox();

                partition[i] =
                    (byte)((adjustStats(X, b.minX, b.maxX, xSplit, xStats) << X)
                 | (adjustStats(Y, b.minY, b.maxY, ySplit, yStats) << Y)
                 | (adjustStats(Z, b.minZ, b.maxZ, zSplit, zStats) << Z));
            }

            //
            // if a split results in either the left or right
            // branch having zero objects of its own we increase
            // the cost of doing a split along that axis.
            //

            if (xStats[LEFT] == 0 || xStats[RIGHT] == 0)
            {
                xStats[BOTH] *= 2;
            }

            if (yStats[LEFT] == 0 || yStats[RIGHT] == 0)
            {
                yStats[BOTH] *= 2;
            }

            if (zStats[LEFT] == 0 || zStats[RIGHT] == 0)
            {
                zStats[BOTH] *= 2;
            }

            //
            // if the following occurs we can't divide any further
            //
            if (xStats[BOTH] == yStats[BOTH] && yStats[BOTH] == zStats[BOTH])
            {
                this.type = LEAF;
                return this;
            }

            //
            // choose the "most efficient" split and create the
            // new branch node.
            //
            BranchNode    node;

            if (xStats[BOTH] < yStats[BOTH])
            {
                if (xStats[BOTH] < zStats[BOTH])
                {
                    node = splitNode(X, xSplit, xStats[LEFT], xStats[RIGHT], start, end);
                }
                else 
                {
                    node = splitNode(Z, zSplit, zStats[LEFT], zStats[RIGHT], start, end);
                }
            }
            else
            {
                if (yStats[BOTH] < zStats[BOTH])
                {
                    node = splitNode(Y, ySplit, yStats[LEFT], yStats[RIGHT], start, end);
                }
                else 
                {
                    node = splitNode(Z, zSplit, zStats[LEFT], zStats[RIGHT], start, end);
                }
            }

            return node;
        }

        HitList intersections(
            Ray    r)
        {
            if (this.isMissedBy(r))
            {
                return null;
            }

            if (type == SPLITABLE)
            {
                TreeNode node = this.split();

                if (this == parent.left)
                {
                    parent.left = node;
                }
                else if (this == parent.mid)
                {
                    parent.mid = node;
                }
                else
                {
                    parent.right = node;
                }

                return node.intersections(r);
            }
            else
            {
                HitList list = null;

                for (int i = this.start; i != this.end; i++)
                {
                    HitList hits = objects[i].intersections(r);

                    if (hits != null)
                    {
                        if (list == null)
                        {
                            list = hits;
                        }
                        else
                        {
                            list.merge(hits);
                        }
                    }
                }

                return list;
            }
        }

        boolean isIntersected(
            Ray    r)
        {
            if (this.isMissedBy(r))
            {
                return false;
            }

            if (type == SPLITABLE)
            {
                TreeNode node = this.split();

                if (this == parent.left)
                {
                    parent.left = node;
                }
                else if (this == parent.mid)
                {
                    parent.mid = node;
                }
                else
                {
                    parent.right = node;
                }

                return node.isIntersected(r);
            }
            else
            {
                boolean hit = false;

                for (int i = this.start; i != this.end; i++)
                {
                    if (objects[i].isIntersected(r))
                    {
                        hit = true;

                        if (r.hit.t < r.theHit.t)
                        {
                            r.setTheHit();
                        }
                    }
                }

                return hit;
            }
        }

        boolean isShadowing(
            Ray    r)
        {
            if (this.isMissedBy(r))
            {
                return false;
            }

            if (type == SPLITABLE)
            {
                TreeNode node = this.split();

                if (this == parent.left)
                {
                    parent.left = node;
                }
                else if (this == parent.mid)
                {
                    parent.mid = node;
                }
                else
                {
                    parent.right = node;
                }

                return node.isShadowing(r);
            }
            else
            {
                boolean hit = false;

                for (int i = this.start; i != this.end; i++)
                {
                    if (objects[i].isShadowing(r))
                    {
                        hit = true;

                        if (r.hit.t < r.theHit.t)
                        {
                            r.setTheHit();
                        }
                    }
                }

                return hit;
            }
        }
    }

    private class BranchNode
        extends TreeNode
    {
        float       splitValue;
        TreeNode    left, mid, right;

        BranchNode(
            byte    type,
            BBox    bBox)
        {
            super(type);

            this.bBox = bBox;
        }

        HitList intersections(
            Ray    r)
        {
            if (this.isMissedBy(r))
            {
                return null;
            }

            TreeNode    next;
            float       axOrg, axDir;

            switch (type)
            {
            case X:
                axOrg = r.org.x;
                axDir = r.dir.x;
                break;
            case Y:
                axOrg = r.org.y;
                axDir = r.dir.y;
                break;
            case Z:
                axOrg = r.org.z;
                axDir = r.dir.z;
                break;
            default:
                throw new IllegalStateException(
                    "bad type in bv-tree");
            }

            if (axDir < 0)
            {
                if (axOrg <= splitValue)
                {
                    HitList list = left.intersections(r);

                    if (list != null)
                    {
                        list.merge(mid.intersections(r));
                    }
                    else
                    {
                        list = mid.intersections(r);
                    }

                    return list;
                }
                else
                {
                    HitList list = right.intersections(r);

                    if (list != null)
                    {
                        list.merge(mid.intersections(r));
                    }
                    else
                    {
                        list = mid.intersections(r);
                    }

                    if (list != null)
                    {
                        list.merge(left.intersections(r));
                    }
                    else
                    {
                        list = left.intersections(r);
                    }

                    return list;
                }
            }
            else
            {
                if (axOrg <= splitValue)
                {
                    HitList list = left.intersections(r);

                    if (list != null)
                    {
                        list.merge(mid.intersections(r));
                    }
                    else
                    {
                        list = mid.intersections(r);
                    }

                    if (list != null)
                    {
                        list.merge(right.intersections(r));
                    }
                    else
                    {
                        list = right.intersections(r);
                    }

                    return list;
                }
                else
                {
                    HitList list = right.intersections(r);

                    if (list != null)
                    {
                        list.merge(mid.intersections(r));
                    }
                    else
                    {
                        list = mid.intersections(r);
                    }

                    return list;
                }
            }
        }

        boolean isIntersected(
            Ray    r)
        {
            if (this.isMissedBy(r))
            {
                return false;
            }

            TreeNode    next;
            float       axOrg, axDir;

            switch (type)
            {
            case X:
                axOrg = r.org.x;
                axDir = r.dir.x;
                break;
            case Y:
                axOrg = r.org.y;
                axDir = r.dir.y;
                break;
            case Z:
                axOrg = r.org.z;
                axDir = r.dir.z;
                break;
            default:
                throw new IllegalStateException(
                    "bad type in bv-tree");
            }

            if (axDir < 0)
            {
                if (axOrg <= splitValue)
                {
                    boolean hit = left.isIntersected(r);

                    hit |= mid.isIntersected(r);

                    return hit;
                }
                else
                {
                    boolean rightHit = right.isIntersected(r);
                    boolean hit = rightHit;

                    hit |= mid.isIntersected(r);

                    if (!rightHit)
                    {
                        hit |= left.isIntersected(r);
                    }

                    return hit;
                }
            }
            else
            {
                if (axOrg <= splitValue)
                {
                    boolean leftHit = left.isIntersected(r);
                    boolean hit = leftHit;

                    hit |= mid.isIntersected(r);

                    if (!leftHit)
                    {
                        hit |= right.isIntersected(r);
                    }

                    return hit;
                }
                else
                {
                    boolean hit = right.isIntersected(r);

                    hit |= mid.isIntersected(r);

                    return hit;
                }
            }
        }

        boolean isShadowing(
            Ray    r)
        {
            if (this.isMissedBy(r))
            {
                return false;
            }

            TreeNode    next;
            float       axOrg, axDir;

            switch (type)
            {
            case X:
                axOrg = r.org.x;
                axDir = r.dir.x;
                break;
            case Y:
                axOrg = r.org.y;
                axDir = r.dir.y;
                break;
            case Z:
                axOrg = r.org.z;
                axDir = r.dir.z;
                break;
            default:
                throw new IllegalStateException(
                    "bad type in bv-tree");
            }

            if (axDir < 0)
            {
                if (axOrg <= splitValue)
                {
                    boolean hit = left.isShadowing(r);

                    hit |= mid.isShadowing(r);

                    return hit;
                }
                else
                {
                    boolean rightHit = right.isShadowing(r);
                    boolean hit = rightHit;

                    hit |= mid.isShadowing(r);

                    if (!rightHit)
                    {
                        hit |= left.isShadowing(r);
                    }

                    return hit;
                }
            }
            else
            {
                if (axOrg <= splitValue)
                {
                    boolean leftHit = left.isShadowing(r);
                    boolean hit = leftHit;

                    hit |= mid.isShadowing(r);

                    if (!leftHit)
                    {
                        hit |= right.isShadowing(r);
                    }

                    return hit;
                }
                else
                {
                    boolean hit = right.isShadowing(r);

                    hit |= mid.isShadowing(r);

                    return hit;
                }
            }
        }
    }

    TreeNode        root;
    Aggregate       lights;
    Partitionable[] objects;
    byte[]          partition;

    public BvTree()
    {
    }

    public BvTree(
        Aggregate    objs)
    {
        //
        // find out how many bounded objects we have.
        //
        int                 size = 0;
        TraceableIterator   objIt = objs.traceables();

        while (objIt.hasNext())
        {
            if (objIt.next() instanceof Partitionable)
            {
                size++;
            }
        }

        objIt.dispose();

        //
        // now create our array of objects
        //
        int     index = 0;
        
        objects = new Partitionable[size];
        partition = new byte[size];

        objIt = objs.traceables();

        while (objIt.hasNext())
        {
            Traceable   obj = objIt.next();

            if (obj instanceof Partitionable)
            {
                objects[index++] = (Partitionable)obj;
            }
        }

        LightIterator   lightIt = objs.lights();

        lights = new Aggregate();

        while (lightIt.hasNext())
        {
            lights.add(lightIt.next());
        }

        lightIt.dispose();

        LeafNode    node = new LeafNode(null, 0, objects.length);

        if (node.type == SPLITABLE)
        {
            root = node.split();
        }
        else
        {
            root = node;
        }
    }

    public BvTree(
        Partitionable[]    objs)
    {
        this.objects = objs;
        partition = new byte[objs.length];

        LeafNode    node = new LeafNode(null, 0, objects.length);

        if (node.type == SPLITABLE)
        {
            root = node.split();
        }
        else
        {
            root = node;
        }
    }

    public LightDataList visibleLights(
        Hit    hit)
    {
        return lights.visibleLights(hit, this);
    }

    public boolean isIntersected(
        Ray    r)
    {
        return root.isIntersected(r);
    }

    public HitList intersections(
        Ray    r)
    {
        return root.intersections(r);
    }

    public boolean isShadowing(
        Ray    r)
    {
        return root.isShadowing(r);
    }
}
