package org.bouncycastle.graphics.rt.objects;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.mappers.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a cone. Canonical form a cone with the tip at the origin, and its
 * base at (0, 0, 1).
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;Cone color="1, 1, 1" baseRadius="1" top="0, 0, 0" base="0, 1, 0"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Float "CDATA"&gt;
 *     &lt;!ENTITY %Point3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT Cone EMPTY&gt;
 *     &lt;!ATTLIST Cone
 *         color      %Color3f; "1, 1, 1"
 *         top        %Point3f; "0, 0, 0"
 *         topRadius  %Float;   "0"
 *         base       %Point3f; "0, 1, 0"
 *         baseRadius %Float;   "1"&gt;
 * </pre>
 * </code>
 */
public class Cone
    extends SimpleObject
{
    static final float TOL        = (float)0.0001;


    private static class FaceNormalCalculator
            implements NormalCalculator
    {
        float   x, y, z;

        FaceNormalCalculator(
            float    x,
            float    y,
            float    z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }

        public void getN(
            Hit         hit,
            Vector3f    n)
        {
            n.set(x, y, z);
        }
    }

    static final NormalCalculator SIDE
                    = new NormalCalculator()
                    {
                        public void getN(
                            Hit         hit,
                            Vector3f    n)
                        {
                            Point3f     loc = hit.oP();

                            n.set(loc.x, loc.y, -loc.z);
                        }
                    };

    static final FaceNormalCalculator NZFACE
                = new FaceNormalCalculator(0, 0, -1);

    static final FaceNormalCalculator PZFACE
                = new FaceNormalCalculator(0, 0, 1);

    float tipValue = 0;

    static BBox    canonicalBBox = new BBox(-1, -1, 0, 1, 1, 1);

    public Cone(
        Point3d       p1,
        double        r1,
        Point3d       p2,
        double        r2)
    {
        init(new Point3f(p1), r1, new Point3f(p2), r2);
    }

    public Cone(
        Point3f       p1,
        double        r1,
        Point3f       p2,
        double        r2)
    {
        init(p1, r1, p2, r2);
    }

    private Cone()
    {
    }

    public static Cone getInstance(
        RMLReader       in)
        throws IOException
    {
        Cone   c = new Cone();

        c.readObject(in);

        return c;
    }

    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        Point3f p1 = params.getPoint("top", "0, 0, 0");
        Point3f p2 = params.getPoint("base", "0, 1, 0");
        float   r1 = params.getFloat("topRadius", "0");
        float   r2 = params.getFloat("baseRadius", "1");

        init(p1, r1, p2, r2);
    }

    private void init(
        Point3f       p1,
        double        r1,
        Point3f       p2,
        double        r2)
    {
        Point3f        b, a;
        double        rad, apexR, baseR, grad, base;

        this.setSelfShadowing(false);

        if (r1 == r2)
        {
            throw new IllegalArgumentException(
                "r1 == r2 in Cone");
        }

        if (r1 > r2)
        {
            a = p2;
            b = p1;
            apexR = r2;
            baseR = r1;
        }
        else
        {
            a = p1;
            b = p2;
            apexR = r1;
            baseR = r2;
        }

        grad = (a.x - b.x) / (baseR - apexR);
        a.x += grad * apexR;
        grad = (a.y - b.y) / (baseR - apexR);
        a.y += grad * apexR;
        grad = (a.z - b.z) / (baseR - apexR);
        a.z += grad * apexR;

        tipValue = (float)(apexR / baseR);

        rad = baseR;

        canonicalBBox.minZ = tipValue;

        this.setBBox(canonicalBBox);

        Transform   trans = new Transform();

        trans.scale(rad, rad,
            (float)Math.sqrt((b.x - a.x) * (b.x - a.x)
                + (b.y - a.y) * (b.y - a.y)
                + (b.z - a.z) * (b.z - a.z)));

        trans.alignZAxis(b.x - a.x,
                b.y - a.y, b.z - a.z);

        trans.translate(a.x, a.y, a.z);

        this.transform(trans);

        tolerance = TOL;
    }

    protected boolean isObjectIntersected(
        Ray        r)
    {
        float       a, b, c, d, t;
        float       ax1, ax2, x, y, z;
        Point3f     org = r.org;
        Vector3f    dir = r.dir;

        a = dir.x * dir.x + dir.y * dir.y - dir.z * dir.z;
        b = org.x * dir.x + org.y * dir.y - dir.z * org.z;
        c = org.x * org.x + org.y * org.y - org.z * org.z;

        d = b * b - a * c;

        if (d < 0)
        {
            return false;
        }
        
        if (Math.abs(a) == 0.0)
        {
            t1 = -c / b;
            enNorm = SIDE;

            z = org.z + t1 * dir.z;

            if (z <= 1 && z >= tipValue)
            {
                t2 = (1 - org.z) / dir.z;
                exNorm = PZFACE;
            }
            else
            {
                t1 = (tipValue - org.z) / dir.z;
                t2 = (1 - org.z) / dir.z;

                if (t1 < TOL && t2 < TOL)
                    return false;

                x = t2 * dir.x + org.x;
                y = t2 * dir.y + org.y;
                if (x * x + y * y > 1.0)
                    return false;

                enNorm = NZFACE;
                exNorm = PZFACE;
            }
        }
        else
        {
            d = (float)Math.sqrt(d);

            enNorm = exNorm = SIDE;

            t1 = (-b - d) / a;
            t2 = (-b + d) / a;

            ax1 = t1 * dir.z + org.z;
            ax2 = t2 * dir.z + org.z;

            //
            // both hits outside range of z in -1 to 1.0
            //
            if (Math.abs(ax1) >= 1 && Math.abs(ax2) >= 1.0)
                return false;

            //
            // both hits in range of z < 0.0
            //
            if (ax1 <= 0.0 && ax2 <= 0.0)
            {
                return false;
            }

            if (ax1 < tipValue && ax2 < tipValue)
            {
                //
                // we pass through truncated section
                //
                if (ax1 > 0.0 && ax2 > 0.0)
                {
                    return false;
                }

                //
                // we must pass through top and base faces
                // of truncated cone.
                //
                enNorm = PZFACE;
                t1 = (1 - org.z) / dir.z;
                exNorm = NZFACE;
                t2 = (tipValue - org.z) / dir.z;

            }
            else if (ax1 >= 1.0)
            {
                //
                // must hit base of cone.
                //
                enNorm = PZFACE;
                t1 = (1 - org.z) / dir.z;

                            // do we hit top? 
                if (ax2 <= tipValue)
                {
                    //
                    // check that we have really hit base
                    //
                    x = t1 * dir.x + org.x;
                    y = t1 * dir.y + org.y;
                    if (x * x + y * y > 1.0)
                        return false;

                    exNorm = NZFACE;
                    t2 = (tipValue - org.z) / dir.z;
                }
            }
            else if (ax1 <= tipValue)
            {
                //
                // must hit top of cone
                //
                enNorm = NZFACE;
                t1 = (tipValue - org.z) / dir.z;

                x = t1 * dir.x + org.x;
                y = t1 * dir.y + org.y;
                if (x * x + y * y > tipValue * tipValue)
                {

                    if (ax2 >= 1.0)
                        return false;

                    enNorm = PZFACE;
                    t1 = (1 - org.z) / dir.z;

                }
                else if (ax2 >= 1.0)
                {
                    exNorm = PZFACE;
                    t2 = (1 - org.z) / dir.z;
                }
            }
            else
            {
                /*
                 * must have hit in the middle, check that
                 * the ax2 value is in range.
                 */
                if (ax2 >= 1.0)
                {
                    exNorm = PZFACE;
                    t2 = (1 - org.z) / dir.z;
                }
                else if (ax2 <= tipValue)
                {
                    exNorm = NZFACE;
                    t2 = (tipValue - org.z) / dir.z;

                    x = t2 * dir.x + org.x;
                    y = t2 * dir.y + org.y;
                    if (x * x + y * y > tipValue * tipValue)
                    {
                            /* must hit base */
                        exNorm = PZFACE;
                        t2 = (1 - org.z) / dir.z;
                    }
                }
            }
        }

        if (t1 > t2)
        {
            NormalCalculator    face;

            t = t1;
            t1 = t2;
            t2 = t;
            face = enNorm;
            enNorm = exNorm;
            exNorm = face;
        }

        if (t2 < TOL)
        {
            return false;
        }

        enShade = exShade = this.getShader();
        enMap = exMap = ConicalMap.canonicalMap;

        return true;
    }
}
