package org.bouncycastle.graphics.rt.objects;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a csg object, these can be nested and support UNION, INTERSECT, and
 * the SUBTRACT operation (right operand affects left).
 */
public class Csg
    extends Bounded
{
    /*
     * CSGNode node structure and defs
     */
    public static final int UNION        = 1;
    public static final int INTERSECT    = 2;
    public static final int SUBTRACT    = 3;

    int                 operation;
    Renderable          left, right;

    private Csg()
    {
    }

    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        String          op = params.getString("operation", "union");

        operation  = UNION;
        if (op.equals("intersect"))
        {
            operation  = INTERSECT;
        }
        else if (op.equals("subtract"))
        {
            operation  = SUBTRACT;
        }
    }

    public static Csg getInstance(
        RMLReader       in)
        throws IOException
    {
        Csg csg = new Csg();

        csg.readObject(in);

        csg.init(csg.operation, csg.left, csg.right);

        return csg;
    }

    protected void processObject(
        Object          obj)
    {
        if (obj instanceof Transform)
        {
            this.transform((Transform)obj);
        }
        else if (obj instanceof Shader)
        {
            this.setShader(shader);
        }
        else if (obj instanceof Renderable)
        {
            if (obj instanceof Aggregate)
            {
                TraceableIterator l = ((Aggregate)obj).traceables();
    
                obj = (Renderable)l.next();
    
                l.dispose();
            }
    
            if (this.left != null)
            {
                this.right = (Renderable)obj;
                this.right.setShader(this.getShader());
                this.right.transform(this.objToRay);
            }
            else
            {
                this.left = (Renderable)obj;
                this.left.setShader(this.getShader());
                this.left.transform(this.objToRay);
            }
        }
        else
        {
            super.processObject(obj);
        }
    }

    public Csg(
        int           operation,
        Renderable    left,
        Renderable    right)
    {
        init(operation, left, right);
    }

    private void init(
        int        operation,
        Renderable    left,
        Renderable    right)
    {
        this.operation = operation;
        this.left = left;
        this.right = right;

        switch (operation)
        {
        case UNION:
            bBox = ((Partitionable)left).bBox().add(((Partitionable)right).bBox());
            break;
        case INTERSECT:
            bBox = ((Partitionable)left).bBox().intersects(((Partitionable)right).bBox());
            break;
        case SUBTRACT:
            bBox = ((Partitionable)left).bBox();
            right.setSelfShadowing(true);
            break;
        default:
            throw new IllegalArgumentException(
                    "unknown operation in Csg");
        }
    }

    /**
     * apply the given 4x4 transform to the object.
     */
    public void transform(
        Matrix4f   trans)
    {
        super.transform(trans);

        if (left != null)
        {
            left.transform(trans);
        }
    
        if (right != null)
        {
            right.transform(trans);
        }
    }

    public void setSelfShadowing(
        boolean selfShadowing)
    {
        super.setSelfShadowing(selfShadowing);

        if (left != null)
        {
            left.setSelfShadowing(selfShadowing);
        }
    
        if (right != null)
        {
            right.setSelfShadowing(selfShadowing);
        }
    }

    /**
     * set the shader for the csg object (note: this will not override shaders
     * already set in sub branches.
     */
    public void setShader(
        Shader   shader)
    {
        super.setShader(shader);

        if (left.getShader() == null)
        {
            left.setShader(shader);
        }

        if (right.getShader() == null)
        {
            right.setShader(shader);
        }
    }

    public HitList intersections(
        Ray         r)
    {
        if (bBox.isMissing(r.org, r.dir))
        {
            return null;
        }

        //
        // get the intersection lists for the left and right branches
        //
        HitList    hits = null;
        HitList    lList, rList;

        switch (operation)
        {
        case UNION:
            lList = left.intersections(r);
            if (lList != null)
            {
                rList = right.intersections(r);
                hits = lList.union(rList);
            }
            else
            {
                hits = right.intersections(r);
            }
            break;
        case INTERSECT:
            lList = left.intersections(r);
            if (lList != null)
            {
                rList = right.intersections(r);
                hits = lList.intersect(rList);
            }
            else
            {
                hits = null;
            }
            break;
        case SUBTRACT:
            lList = left.intersections(r);
            if (lList != null)
            {
                rList = right.intersections(r);
                if (rList != null)
                {
                    HitIterator it = rList.iterator();

                    while (it.hasNext())
                    {
                        it.next().subtractedSurface = true;
                    }

                    it.dispose();
                }

                hits = lList.subtract(rList);
            }
            else
            {
                hits = null;
            }
            break;
        default:
            throw new IllegalStateException(
                    "unknown operation in Csg");
        }

        return hits;
    }

    public boolean isIntersected(
        Ray    r)
    {
        HitList    hits = this.intersections(r);

        if (hits == null)
        {
            return false;
        }

        Hit hit = hits.head();

        r.setHit(hit);

        hits.dispose();

        return true;
    }

    public boolean isShadowing(
        Ray    r)
    {
        HitList    hits = this.intersections(r);

        if (hits == null)
        {
            return false;
        }

        Hit hit = hits.head();

        float t = hit.distance();

        hits.dispose();

        return t < r.tMax;
    }
}
