package org.bouncycastle.graphics.rt.objects;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.mappers.*;

import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a cylinder. In canonical form a cylinder of unit radius, one end at the
 * origin the other at (0, 0, 1).
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;Cylinder top="0, 1, 0" base="0, 0, 0" radius="1"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Float "CDATA"&gt;
 *     &lt;!ENTITY %Point3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT Cylinder EMPTY&gt;
 *     &lt;!ATTLIST Cylinder
 *         color      %Color3f; "1, 1, 1"
 *         top        %Point3f; "0, 1, 0"
 *         base       %Point3f; "0, 0, 0"
 *         radius     %Float;   "1"&gt;
 * </pre>
 * </code>
 */
public class Cylinder
    extends SimpleObject
{
    public static final float TOL = (float)0.0001;

    private static class FaceNormalCalculator
            implements NormalCalculator
    {
        float   x, y, z;

        FaceNormalCalculator(
            float    x,
            float    y,
            float    z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }

        public void getN(
            Hit         hit,
            Vector3f    n)
        {
            n.set(x, y, z);
        }
    }

    static final NormalCalculator SIDE
                    = new NormalCalculator()
                    {
                        public void getN(
                            Hit         hit,
                            Vector3f    n)
                        {
                            Point3f loc = hit.oP();

                            n.set(loc.x, loc.y, 0);
                        }
                    };

    static final FaceNormalCalculator NZFACE
                = new FaceNormalCalculator(0, 0, -1);

    static final FaceNormalCalculator PZFACE
                = new FaceNormalCalculator(0, 0, 1);

    static BBox    canonicalBBox = new BBox(-1, -1, 0, 1, 1, 1);

    public Cylinder(
        Point3d        t,
        Point3d        b,
        double         r)
    {
        this(new Point3f(t), new Point3f(b), r);
    }

    public Cylinder(
        Point3f        t,
        Point3f        b,
        double         r)
    {
        init(t, b, r);
    }

    private Cylinder()
    {
    }

    public static Cylinder getInstance(
        RMLReader       in)
        throws IOException
    {
        Cylinder   c = new Cylinder();

        c.readObject(in);

        return c;
    }

    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        Point3f p1 = params.getPoint("top", "0, 1, 0");
        Point3f p2 = params.getPoint("base", "0, 0, 0");
        float   r = params.getFloat("radius", "1");

        init(p1, p2, r);
    }

    private void init(
        Point3f        t,
        Point3f        b,
        double        r)
    {
        this.setBBox(canonicalBBox);
        this.setSelfShadowing(false);

        Transform   trans = new Transform();

        trans.scale(r, r, (float)Math.sqrt((t.x - b.x) * (t.x - b.x)
                    + (t.y - b.y) * (t.y - b.y)
                    + (t.z - b.z) * (t.z - b.z)));

        trans.alignZAxis(t.x - b.x, t.y - b.y, t.z - b.z);

        trans.translate(b.x, b.y, b.z);

        this.transform(trans);

        tolerance = TOL;
    }

    protected boolean isObjectIntersected(
        Ray        r)
    {
        float           a, b, c, d, t;
        float           ax1, ax2;
        Point3f         org = r.org;
        Vector3f        dir = r.dir;

        a = dir.x * dir.x + dir.y * dir.y;
        b = -(org.x * dir.x + org.y * dir.y);
        c = org.x * org.x + org.y * org.y - 1;

        d = b * b - a * c;

        if (d < 0 || a == 0)
        {
            if (c > 0 || dir.z == 0)
            {
                return false;
            }

            t1 = -org.z / dir.z;
            t2 = (1 - org.z) / dir.z;

            if (t1 < TOL && t2 < TOL)
            {
                return false;
            }

            if (t1 > t2)
            {
                t = t1;
                t1 = t2;
                t2 = t;
                enNorm = PZFACE;
                exNorm = NZFACE;
            }
            else
            {
                enNorm = NZFACE;
                exNorm = PZFACE;
            }
        }
        else
        {
            enNorm = exNorm = SIDE;

            d = (float)Math.sqrt(d);

            t1 = (b - d) / a;
            t2 = (b + d) / a;

            ax1 = t1 * dir.z + org.z;
            ax2 = t2 * dir.z + org.z;

            if (ax1 < 0 && ax2 < 0)
            {
                return false;
            }

            if (ax1 > 1 && ax2 > 1)
            {
                return false;
            }

            if (ax1 > 1)
            {
                enNorm = PZFACE;
                t1 = (1 - org.z) / dir.z;
            }
            else if (ax1 < 0)
            {
                enNorm = NZFACE;
                t1 = -org.z / dir.z;
            } 

            if (ax2 > 1)
            {
                exNorm = PZFACE;
                t2 = (1 - org.z) / dir.z;
            }
            else if (ax2 < 0)
            {
                exNorm = NZFACE;
                t2 = -org.z / dir.z;
            }
        }

        if (t2 < TOL)
        {
            return false;
        }

        enShade = exShade = this.getShader();
        enMap = exMap = CylindricalMap.canonicalMap;

        return true;
    }
}
