package org.bouncycastle.graphics.rt.objects;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a directional light - the basic spotlight. This defines a light
 * with as a point source with the light being shined on objects that
 * fall within the cone described by the lights outside angle and direction.
 * <p>
 * By default is affects all objects unless the addAffected method is used,
 * in which case it only affects those objects specified.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;DirectionalLight location="0, 0, 0" direction="0, 0, -1" color="0.8, 0.7, 1"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float    "CDATA"&gt;
 *     &lt;!ENTITY %Color3f  "CDATA"&gt;
 *     &lt;!ENTITY %Point3f  "CDATA"&gt;
 *     &lt;!ENTITY %Vector3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT DirectionalLight EMPTY&gt;
 *     &lt;!ATTLIST DirectionalLight
 *         color        %Color3f;  "1, 1, 1"
 *         location     %Vector3f; "0, 0, 1"
 *         direction    %Point3f;  "0, 0, 1"
 *         angle        %Float;    "45"
 *         innerAngle   %Float;    "45"&gt;
 * </pre>
 * </code>
 */
public class DirectionalLight
    implements Light
{
    Point3f     loc;
    Vector3f    dir;
    Color3f     col;
    double      cosEdge;
    double      cosInnerEdge;

    boolean     shadows = true;        // whether or not we produce shadows

    Hashtable   affected = null;    // table of objects we affect

    /**
     * basic constructor.
     *
     * @param loc the location of the light source.
     * @param dir axis of the cone of influence for the light.
     * @param angle the spread (in degrees) of the cone of influence.
     * @param innerAngle angle at which the "penumbra" starts.
     * @param col the color of the incoming light.
     */
    public DirectionalLight(
        Point3f     loc,
        Vector3f    dir,
        double      angle,
        double      innerAngle,
        Color3f     col)
    {
        this.loc = loc;
        this.dir = dir;
        this.col = col;

        this.cosEdge = Math.cos(angle * Math.PI / 180.0);
        this.cosInnerEdge = Math.cos(innerAngle * Math.PI / 180.0);

        //
        // we reverse as we always move towards light
        // when we trace.
        //
        this.dir.scale(-1);
    }

    
    private DirectionalLight()
    {
    }

    /**
     * build a directional light from an RML file.
     */
    public static Light getInstance(
        RMLReader       in)
        throws IOException
    {
        DirectionalLight    d = new DirectionalLight();

        ParametersMap   params = in.parameters();

        d.loc = params.getPoint("location", "0, 0, 0");
        d.dir = params.getVector("direction", "0, 0, 1");
        d.col = params.getColor("color", "1, 1, 1");

        d.cosEdge = Math.cos(params.getFloat("angle", "45") * Math.PI / 180.0);
        if (params.get("innerAngle") != null)
        {
            d.cosInnerEdge = Math.cos(params.getFloat("innerAngle", "45") * Math.PI / 180.0);
        }
        else
        {
            d.cosInnerEdge = d.cosEdge;
        }

        //
        // we reverse as we always move towards light
        // when we trace.
        //
        d.dir.scale(-1);

        if (in.token() != RMLReader.EMPTY_TAG)
        {
            in.skipToEndTag(in.toString());
        }

        return d;
    }

    /**
     * basic constructor - no penumbra.
     *
     * @param loc the location of the light source.
     * @param dir axis of the cone of influence for the light.
     * @param angle the spread (in degrees) of the cone of influence.
     * @param col the color of the incoming light.
     */
    public DirectionalLight(
        Point3f        loc,
        Vector3f    dir,
        double        angle,
        Color3f        col)
    {
        this(loc, dir, angle, angle, col);
    }

    /**
     * set whether or not this light can be occluded.
     */
    public void setShadows(
        boolean    shadows)
    {
        this.shadows = shadows;
    }

    /**
     * transform the location of the light.
     */
    public void transform(
        Transform   trans)
    {
        //
    }

    public void addAffected(
        Renderable    obj)
    {
        if (affected == null)
        {
            affected = new Hashtable();
        }

        affected.put(obj, obj);
    }

    public LightData getLightData(
        Hit           hit,
        Illuminated   world)
    {
        if ((affected != null) && (affected.get(hit.obj) == null))
        {
            return null;
        }

        //
        // regular processing
        //
        Vector3f        n = hit.N();
        Ray             lr = Ray.getInstance();
        float           dist;
        float           cos;
        float           nDotL;

        lr.dir.sub(loc, hit.P());

        dist = (float)Math.sqrt(lr.dir.dot(lr.dir));

        lr.dir.scale(1 / dist);

        if ((cos = dir.dot(lr.dir)) < cosEdge)
        {
            return null;
        }

        if ((nDotL = n.dot(lr.dir)) < 0)
        {
            return null;
        }

        lr.org.set(hit.P());
        lr.orgObj = hit.obj;
        lr.tMax = dist;
        lr.col.set(col);

        if (shadows && world.isShadowing(lr))
        {
            return null;
        }
            
        LightData ld = LightData.getInstance(nDotL, lr.dir, lr.col);

        lr.dispose();

        return ld;
    }
}
