package org.bouncycastle.graphics.rt.objects;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a distant light. Basically this emulates a point light source so
 * far away that, to the world, the rays coming from the light source
 * are parallel - it's kind of like the sun! By default is affects all
 * objects unless the addAffected method is used, in which case it only
 * affects those objects specified.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;DistantLight direction="0, 0, -1" color="0.8, 0.7, 1"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float    "CDATA"&gt;
 *     &lt;!ENTITY %Color3f  "CDATA"&gt;
 *     &lt;!ENTITY %Vector3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT DistantLight EMPTY&gt;
 *     &lt;!ATTLIST DistantLight
 *         color        %Color3f;  "1, 1, 1"
 *         direction    %Vector3f; "0, 0, 1"&gt;
 * </pre>
 * </code>
 */
public class DistantLight
    implements Light
{
    Vector3f    dir;
    Color3f     col;

    boolean     shadows = true;     // whether or not we produce shadows

    Hashtable   affected = null;    // table of objects we affect

    /**
     * basic constructor.
     *
     * @param dir the direction of the incoming light.
     * @param col the color of the incoming light.
     */
    public DistantLight(
        Vector3f    dir,
        Color3f     col)
    {
        this.dir = dir;
        this.col = col;

        //
        // we reverse as we always move towards light
        // when we trace.
        //
        this.dir.scale(-1);
    }

    private DistantLight()
    {
    }

    public static Light getInstance(
        RMLReader       in)
        throws IOException
    {
        DistantLight    d = new DistantLight();
        ParametersMap   params = in.parameters();

        d.dir = params.getVector("direction", "0, 0, 0");
        d.col = params.getColor("color", "1, 1, 1");

        if (in.token() != RMLReader.EMPTY_TAG)
        {
            in.skipToEndTag(in.toString());
        }

        return d;
    }

    /**
     * set whether or not this light can be occluded.
     */
    public void setShadows(
        boolean    shadows)
    {
        this.shadows = shadows;
    }

    /**
     * transform the location of the light.
     */
    public void transform(
        Transform   trans)
    {
        //
    }

    public void addAffected(
        Renderable    obj)
    {
        if (affected == null)
        {
            affected = new Hashtable();
        }

        affected.put(obj, obj);
    }

    public LightData getLightData(
        Hit           hit,
        Illuminated   world)
    {
        if ((affected != null) && (affected.get(hit.obj) == null))
        {
            return null;
        }

        //
        // regular processing
        //
        Vector3f    n = hit.N();
        Ray         lr = Ray.getInstance();
        float       dist;
        float       nDotL;

        lr.dir.set(dir);

        if ((nDotL = n.dot(lr.dir)) < 0)
        {
            return null;
        }

        lr.org.set(hit.P());
        lr.orgObj = hit.obj;
        lr.tMax = Float.MAX_VALUE;
        lr.col.set(col);

        if (shadows && world.isShadowing(lr))
        {
            return null;
        }
            
        LightData ld = LightData.getInstance(nDotL, lr.dir, lr.col);

        lr.dispose();

        return ld;
    }
}
