package org.bouncycastle.graphics.rt.objects;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;

/**
 * An indexed polygon mesh.
 */
public class IndexedMesh
    extends Bounded implements TextureMapper
{
    static final byte    X = 0;
    static final byte    Y = 1;
    static final byte    Z = 2;

    static final float TOLERANCE = 0.0001f;

    float[]         xVerts, yVerts, zVerts;

    MeshFacet[]     facets;
    Vector3f[]      norms;
    Color3f[]       colors;
    short[][]       vertexIndexes;

    Color3f[]       faceColors;
    short[]         faceColorIndexes;

    Vector3f[]      vertexNormals;
    short[][]       vertexNormalIndexes;

    Point3f[]       vertexTxtCoords;
    short[][]       vertexTxtCoordsIndexes;

    Color3f[]       vertexColors;
    short[][]       vertexColorsIndexes;

    boolean         smooth = false;

    BvTree          tree;

    Renderable      parent = this;

    Matrix3f        interpMat = new Matrix3f();

    private abstract class MeshFacet
        implements Partitionable, ColorCalculator, NormalCalculator
    {
        short       index;
        byte        axis;
        Vector3f    n;
        float       cnst;
        short[]     indexes;
        Shader      shader;

        public BBox bBox()
        {
            BBox bBox = new BBox(
                xVerts[indexes[0]], yVerts[indexes[0]],
                zVerts[indexes[0]], xVerts[indexes[0]],
                yVerts[indexes[0]], zVerts[indexes[0]]);

            for (int i = 1; i != indexes.length; i++)
            {
                bBox.adjustFor(xVerts[indexes[i]],
                    yVerts[indexes[i]], zVerts[indexes[i]]);
            }

            if ((bBox.maxX - bBox.minX) < TOLERANCE)
            {
                bBox.maxX += TOLERANCE;
                bBox.minX -= TOLERANCE;
            }

            if ((bBox.maxY - bBox.minY) < TOLERANCE)
            {
                bBox.maxY += TOLERANCE;
                bBox.minY -= TOLERANCE;
            }

            if ((bBox.maxZ - bBox.minZ) < TOLERANCE)
            {
                bBox.maxZ += TOLERANCE;
                bBox.minZ -= TOLERANCE;
            }

            return bBox;
        }

        public boolean isIntersected(
            Ray r)
        {
            return isFacetIntersected(r);
        }

        public HitList intersections(
            Ray r)
        {
            if (isFacetIntersected(r))
            {
                HitList list = HitList.getInstance();

                list.add(
                    Hit.getInstance(parent, r.hit.distance(),
                                this, this, (TextureMapper)parent, r.hit.shader));

                return list;
            }

            return null;
        }

        public boolean isShadowing(Ray r)
        {
            if (!isFacetIntersected(r))
            {
                return false;
            }

            // falloff / absoroption

            if (r.hit.t > TOLERANCE)
            {
                return (r.hit.t < r.tMax);
            }

            return false;
        }

        abstract boolean isFacetIntersected(Ray r);

        public abstract void getColor(Hit hit, Color3f c);

        public abstract void getN(Hit hit, Vector3f n);
    }

    private class MeshTriangle
        extends MeshFacet
    {
        MeshTriangle(
            short       index,
            byte        axis,
            Vector3f    n,
            float       cnst,
            short[]     indexes)
        {
            this.index = index;
            this.axis = axis;
            this.n = n;
            this.cnst = cnst;
            this.indexes = indexes;
        }

        private boolean isPointIn2DPolygon(
            float        pu,
            float        pv,
            float[]      us,
            float[]      vs)
        {
            float   m;
            int     crosses = 0;
    
    		if ((vs[indexes[0]] - pv) * (vs[indexes[1]] - pv) < 0.0)
            {
    			if (us[indexes[0]] < pu)
                {
    				if (us[indexes[1]] < pu)
                    {
    					crosses++;
                    }
    				else
                    {
    					m = (us[indexes[1]] - us[indexes[0]]) / (vs[indexes[1]] - vs[indexes[0]]);
    					if (((us[indexes[1]] - pu) - m * (vs[indexes[1]] - pv)) < 0.0)
                        {
    						crosses++;
                        }
    				}
    			}
                else if (us[indexes[1]] < pu)
                {
    				m = (us[indexes[1]] - us[indexes[0]]) / (vs[indexes[1]] - vs[indexes[0]]);
    				if (((us[indexes[1]] - pu) - m * (vs[indexes[1]] - pv)) < 0.0)
                    {
    					crosses++;
                    }
    			}
    		}
    
    		if ((vs[indexes[1]] - pv) * (vs[indexes[2]] - pv) < 0.0)
            {
    			if (us[indexes[1]] < pu)
                {
    				if (us[indexes[2]] < pu)
                    {
    					crosses++;
                    }
    				else
                    {
    					m = (us[indexes[2]] - us[indexes[1]]) / (vs[indexes[2]] - vs[indexes[1]]);
    					if (((us[indexes[2]] - pu) - m * (vs[indexes[2]] - pv)) < 0.0)
                        {
    						crosses++;
                        }
    				}
    			}
                else if (us[indexes[2]] < pu)
                {
    				m = (us[indexes[2]] - us[indexes[1]]) / (vs[indexes[2]] - vs[indexes[1]]);
    				if (((us[indexes[2]] - pu) - m * (vs[indexes[2]] - pv)) < 0.0)
                    {
    					crosses++;
                    }
    			}
    		}
    
    		if ((vs[indexes[2]] - pv) * (vs[indexes[0]] - pv) < 0.0)
            {
    			if (us[indexes[2]] < pu)
                {
    				if (us[indexes[0]] < pu)
                    {
    					crosses++;
                    }
    				else
                    {
    					m = (us[indexes[0]] - us[indexes[2]]) / (vs[indexes[0]] - vs[indexes[2]]);
    					if (((us[indexes[0]] - pu) - m * (vs[indexes[0]] - pv)) < 0.0)
                        {
    						crosses++;
                        }
    				}
    			}
                else if (us[indexes[0]] < pu)
                {
    				m = (us[indexes[0]] - us[indexes[2]]) / (vs[indexes[0]] - vs[indexes[2]]);
    				if (((us[indexes[0]] - pu) - m * (vs[indexes[0]] - pv)) < 0.0)
                    {
    					crosses++;
                    }
    			}
    		}

            return ((crosses & 1) == 1);
        }

        boolean isFacetIntersected(
            Ray    r)
        {
            float           t, pu, pv;
            boolean         hit = false;
            Point3f         org = r.org;
            Vector3f        dir = r.dir;

            t = n.dot(dir);

            if (t == 0)
            {
                return false;
            }

            t = -((n.x * org.x + n.y * org.y + n.z * org.z) + cnst) / t;

            if (t < TOLERANCE)
            {
                return false;
            }

            switch (axis)
            {
            case X:
                pu = org.y + t * dir.y;
                pv = org.z + t * dir.z;

                hit = isPointIn2DPolygon(pu, pv, yVerts, zVerts);
                break;
            case Y:
                pu = org.x + t * dir.x;
                pv = org.z + t * dir.z;

                hit = isPointIn2DPolygon(pu, pv, xVerts, zVerts);
                break;
            case Z:
                pu = org.x + t * dir.x;
                pv = org.y + t * dir.y;

                hit = isPointIn2DPolygon(pu, pv, xVerts, yVerts);
                break;
            default:
                throw new IllegalStateException(
                            "isFacetIntersected: bad axis.\n");
            }

            if (hit)
            {
                r.setHit(parent, t, this, this, (TextureMapper)parent, shader);
            }

            return hit;
        }

        /**
         * calculates the interpolated normal for the triangle
         * described by the normals norms. In this case we have to
         * use area (barycentric) coordinates. P is the (x, y, z)
         * point we are on.
         */
        private void interp3(
            float       u,
            float       v,
            float[]     us,
            float[]     vs,
            Tuple3f[]   values,
            Tuple3f     n)
        {
            int        i1 = indexes[0];
            int        i2 = indexes[1];
            int        i3 = indexes[2];
            float      a1, a2, a3, area;

            if (axis == Y)
            {
                interpMat.m00 = (us[i3] - us[i1]);
                interpMat.m10 = (us[i1] - us[i2]);
                interpMat.m20 = (us[i2] - us[i3]);

                interpMat.m01 = (vs[i3] - vs[i1]);
                interpMat.m11 = (vs[i1] - vs[i2]);
                interpMat.m21 = (vs[i2] - vs[i3]);
            }
            else
            {
                interpMat.m00 = (us[i1] - us[i3]);
                interpMat.m10 = (us[i2] - us[i1]);
                interpMat.m20 = (us[i3] - us[i2]);

                interpMat.m01 = (vs[i1] - vs[i3]);
                interpMat.m11 = (vs[i2] - vs[i1]);
                interpMat.m21 = (vs[i3] - vs[i2]);
            }

            interpMat.m02 = us[i3] * interpMat.m01 - vs[i3] * interpMat.m00;
            interpMat.m12 = us[i1] * interpMat.m11 - vs[i1] * interpMat.m10;
            interpMat.m22 = us[i2] * interpMat.m21 - vs[i2] * interpMat.m20;

            a1 = (v * interpMat.m00 - u * interpMat.m01 + interpMat.m02);

            a2 = (v * interpMat.m10 - u * interpMat.m11 + interpMat.m12);

            a3 = (v * interpMat.m20 - u * interpMat.m21 + interpMat.m22);

            area = a1 + a2 + a3;

            //
            // calculate the weighted normal
            //
            Tuple3f    n1 = values[i1];
            Tuple3f    n2 = values[i2];
            Tuple3f    n3 = values[i3];

            n.x = (a1 * n2.x + a2 * n3.x + a3 * n1.x) / area;
            n.y = (a1 * n2.y + a2 * n3.y + a3 * n1.y) / area;
            n.z = (a1 * n2.z + a2 * n3.z + a3 * n1.z) / area;
        }

        public void getN(
            Hit         hit,
            Vector3f    oN)
        {
            if (vertexNormals != null)
            {
                Point3f loc = hit.oP();

                switch (axis)
                {
                case X:
                    interp3(loc.y, loc.z, yVerts, zVerts, vertexNormals, oN);
                    return;
                case Y:
                    interp3(loc.x, loc.z, xVerts, zVerts, vertexNormals, oN);
                    return;
                case Z:
                    interp3(loc.x, loc.y, xVerts, yVerts, vertexNormals, oN);
                    return;
                default:
                    throw new IllegalStateException("getN: bad axis.\n");
                }
            }
            else
            {
                oN.set(this.n);
            }
        }

        public void getColor(
            Hit         hit,
            Color3f     c)
        {
            if (vertexColors != null)
            {
                Point3f loc = hit.oP();

                switch (axis)
                {
                case X:
                    interp3(loc.y, loc.z, yVerts, zVerts, vertexColors, c);
                    return;
                case Y:
                    interp3(loc.x, loc.z, xVerts, zVerts, vertexColors, c);
                    return;
                case Z:
                    interp3(loc.x, loc.y, xVerts, yVerts, vertexColors, c);
                    return;
                default:
                    throw new IllegalStateException("getColor: bad axis.\n");
                }
            }
            else if (faceColors != null)
            {
                c.set(faceColors[faceColorIndexes[this.index]]);
            }
            else
            {
                c.set(parent.getColor());
            }
        }
    }

    private class MeshPolygon
        extends MeshFacet
    {
        MeshPolygon(
            short       index,
            byte        axis,
            Vector3f    n,
            float       cnst,
            short[]     indexes)
        {
            this.index = index;
            this.axis = axis;
            this.n = n;
            this.cnst = cnst;
            this.indexes = indexes;
        }

        private boolean isPointIn2DPolygon(
            float       pu,
            float       pv,
            float[]     us,
            float[]     vs)
        {
            int    crosses = 0;

            for (int cp = 0, np = indexes.length - 1;
                            np > -1; cp = np--)
            {
                //
                // If we are both above, or both below the
                // intersection point, go onto the next one.
                //

                if (vs[indexes[cp]] < pv)
                {
                    if (vs[indexes[np]] < pv)
                    {
                        continue;
                    }
                }
                else
                {
                    if (vs[indexes[np]] >= pv)
                    {
                        continue;
                    }
                }

                //
                // Find out if we have crossed in the negative.
                //

                if (us[indexes[cp]] < pu)
                {
                    if (us[indexes[np]] < pu)
                    {        // crossed on left
                        crosses++;
                        continue;
                    }
                }
                else if (us[indexes[np]] >= pu)
                {            // both to right
                    continue;
                }

                // 
                // We calculate the u at the intersection of
                // the half-ray and the edge and test it
                // against pu. 
                //
                float    m;

                m = (us[indexes[np]] - us[indexes[cp]])
                                / (vs[indexes[np]] - vs[indexes[cp]]);
                if (((us[indexes[np]] - pu) - m * (vs[indexes[np]] - pv)) < 0)
                {
                    crosses++;
                }
            }

            return ((crosses & 1) == 1);
        }

        boolean isFacetIntersected(
            Ray    r)
        {
            float           t, pu, pv;
            boolean         hit = false;
            Point3f         org = r.org;
            Vector3f        dir = r.dir;

            t = n.dot(dir);

            if (t == 0)
            {
                return false;
            }

            t = -((n.x * org.x + n.y * org.y + n.z * org.z)
                                + cnst) / t;

            if (t < TOLERANCE)
            {
                return false;
            }

            switch (axis)
            {
            case X:
                pu = org.y + t * dir.y;
                pv = org.z + t * dir.z;

                hit = isPointIn2DPolygon(pu, pv, yVerts, zVerts);
                break;
            case Y:
                pu = org.x + t * dir.x;
                pv = org.z + t * dir.z;

                hit = isPointIn2DPolygon(pu, pv, xVerts, zVerts);
                break;
            case Z:
                pu = org.x + t * dir.x;
                pv = org.y + t * dir.y;

                hit = isPointIn2DPolygon(pu, pv, xVerts, yVerts);
                break;
            default:
                throw new IllegalStateException(
                            "isFacetIntersected: bad axis.\n");
            }

            if (hit)
            {
                r.setHit(parent, t, this, this, (TextureMapper)parent, shader);
            }

            return hit;
        }

        public void getColor(
            Hit         hit,
            Color3f     c)
        {
            if (faceColors != null)
            {
                c.set(
                    faceColors[faceColorIndexes[this.index]]);
            }
            else
            {
                c.set(parent.getColor());
            }
        }

        public void getN(
            Hit         hit,
            Vector3f    oN)
        {
            oN.set(this.n);
        }
    }

    public IndexedMesh(
        Point3f[]   vertices,
        short[][]   indxs)
    {
        init(vertices, indxs, false);
    }

    public IndexedMesh(
        Point3f[]   vertices,
        short[][]   indxs,
        boolean     smooth)
    {
        init(vertices, indxs, smooth);
    }

    protected IndexedMesh()
    {
    }

    protected void processParameters(
        ParametersMap   params)
    {
        super.processParameters(params);

        if (params.get("smooth") != null)
        {
            smooth = Boolean.valueOf((String)params.get("smooth")).booleanValue();
        }
    }

    protected void init(
        Point3f[]   vertices,
        short[][]   indxs,
        boolean     smooth)
    {
        this.smooth = smooth;

        init(vertices, indxs);
    }

    protected void init(
        Point3f[]   vertices,
        short[][]   indxs)
    {
        vertexIndexes = indxs;

        xVerts = new float[vertices.length];
        yVerts = new float[vertices.length];
        zVerts = new float[vertices.length];

        for (int i = 0; i != vertices.length; i++)
        {
            xVerts[i] = vertices[i].x;
            yVerts[i] = vertices[i].y;
            zVerts[i] = vertices[i].z;
        }

        //
        // create the polygons from the mesh, eliminating the duds...
        //
        int             duds = 0;
        MeshFacet[]     treeFacets;

        treeFacets = facets = new MeshFacet[indxs.length];

        if (smooth)
        {
            vertexNormalIndexes = indxs;
            vertexNormals = new Vector3f[vertices.length];

            for (int i = 0; i != vertexNormals.length; i++)
            {
                vertexNormals[i] = new Vector3f();
            }
        }

        for (int i = 0; i != indxs.length; i++)
        {
            short[]    inds = indxs[i];

            facets[i] = makePoly(i, inds);
            if (facets[i] == null)
            {
                duds++;
            }
            else
            {
                if (smooth)
                {
                    for (int j = 0; j != inds.length; j++)
                    {
                        vertexNormals[inds[j]].add(facets[i].n);
                    }
                }
            }
        }

        if (smooth)
        {
            for (int i = 0; i != vertexNormals.length; i++)
            {
                Vector3f    v = vertexNormals[i];

                if ((v.x != 0) || (v.y != 0) || (v.z != 0))
                {
                    v.normalize();
                }
            }
        }

        if (duds != 0)
        {
            int        count = 0;

            treeFacets = new MeshFacet[indxs.length - duds];
            for (int i = 0; i != indxs.length; i++)
            {
                if (facets[i] != null)
                {
                    treeFacets[count++] = facets[i];
                }
            }
        }

        //
        // calculate the bounding box
        //
        bBox = new BBox(vertices[0].x, vertices[0].y, vertices[0].z,
                vertices[0].x, vertices[0].y, vertices[0].z);

        for (int i = 1; i != vertices.length; i++)
        {
            bBox.adjustFor(vertices[i]);
        }

        if ((bBox.maxX - bBox.minX) < TOLERANCE)
        {
            bBox.minX -= TOLERANCE;
            bBox.maxX += TOLERANCE;
        }

        if ((bBox.maxY - bBox.minY) < TOLERANCE)
        {
            bBox.minY -= TOLERANCE;
            bBox.maxY += TOLERANCE;
        }

        if ((bBox.maxZ - bBox.minZ) < TOLERANCE)
        {
            bBox.minZ -= TOLERANCE;
            bBox.maxZ += TOLERANCE;
        }

        tree = new BvTree(treeFacets);
    }

    private MeshFacet makePoly(
        int             index,
        short[]         indexes)
    {
        float       cnst;
        byte        axis;
        Vector3f    v1 = new Vector3f();
        Vector3f    v2 = new Vector3f();
        Vector3f    v = new Vector3f();
        Vector3f    n = new Vector3f();

        //
        // calculate normal
        //
        v1.x = xVerts[indexes[1]] - xVerts[indexes[0]];
        v1.y = yVerts[indexes[1]] - yVerts[indexes[0]];
        v1.z = zVerts[indexes[1]] - zVerts[indexes[0]];

        v2.x = xVerts[indexes[2]] - xVerts[indexes[1]];
        v2.y = yVerts[indexes[2]] - yVerts[indexes[1]];
        v2.z = zVerts[indexes[2]] - zVerts[indexes[1]];

        n.cross(v1, v2);

        if (n.x == 0 && n.y == 0 && n.z == 0)
        {
            return null;        // a dud 
        }

        n.normalize();

        //
        // calculate plane constant
        //
        v.x = xVerts[indexes[0]];
        v.y = yVerts[indexes[0]];
        v.z = zVerts[indexes[0]];

        cnst = -n.dot(v);

        //
        // find major axis.
        //
        if (Math.abs(n.x) > Math.abs(n.y))
        {
            if (Math.abs(n.x) > Math.abs(n.z))
            {
                axis = X;
            }
            else
            {
                axis = Z;
            }
        }
        else
        {
            if (Math.abs(n.y) > Math.abs(n.z))
            {
                axis = Y;
            }
            else
            {
                axis = Z;
            }
        }

        if (indexes.length == 3)
        {
            return new MeshTriangle((short)index, axis, n, cnst, indexes);
        }
        
        return new MeshPolygon((short)index, axis, n, cnst, indexes);
    }

    private boolean isObjectIntersected(
        Ray    r)
    {
        Ray         nr;
        boolean     hit = false;

        nr = this.createObjectSpaceRay(r);

        hit = tree.isIntersected(nr);
        if (hit)
        {
            r.setHit(this, nr.theHit.t,
                nr.theHit.colorCalc, nr.theHit.normCalc, nr.theHit.stMapper, nr.theHit.shader);
        }

        nr.dispose();

        return hit;
    }

    public boolean isShadowing(
        Ray    r)
    {
        if (!isObjectIntersected(r))
        {
            return false;
        }

        if (r.hit.t > TOLERANCE)
        {
            return (r.hit.t < r.tMax);
        }

        return false;
    }

    public HitList intersections(
        Ray         r)
    {
        Ray         nr;
        HitList     list;

        nr = this.createObjectSpaceRay(r);

        list = tree.intersections(nr);

        nr.dispose();

        return list;
    }

    public boolean isIntersected(
        Ray    r)
    {
        return isObjectIntersected(r);
    }

    public boolean getST(
        Tuple3f l,
        float[] uv)
    {
        uv[S] = l.x;
        uv[T] = l.y;

        return true;
    }

    public void setShader(
        Shader    shader)
    {
        super.setShader(shader);

        for (int i = 0; i != facets.length; i++)
        {
            if (facets[i] != null)
            {
                facets[i].shader = shader;
            }
        }
    }

    public void setShader(
        Shader[]    shaders,
        short[]     shaderIndexes)
    {
        for (int i = 0; i != facets.length; i++)
        {
            if (facets[i] != null)
            {
                facets[i].shader = shaders[shaderIndexes[i]];
            }
        }
    }

    protected short[][] vertexIndexes()
    {
        return vertexIndexes;
    }

    public void setColors(
        Color3f[]   colors,
        short[]     colorIndexes)
    {
        this.faceColors = colors;
        this.faceColorIndexes = colorIndexes;
    }

    public void setVertexNormals(
        Vector3f[]   normals,
        short[][]    indxs)
    {
        this.vertexNormals = normals;
        this.vertexNormalIndexes = indxs;
    }

    public void setVertexTextureCoordinates(
        Point3f[]    coords,
        short[][]    indxs)
    {
        this.vertexTxtCoords = coords;
        this.vertexTxtCoordsIndexes = indxs;
    }

    public void setVertexColors(
        Color3f[]    colors,
        short[][]    indxs)
    {
        this.vertexColors = colors;
        this.vertexColorsIndexes = indxs;
    }
}
