package org.bouncycastle.graphics.rt.objects;

import java.util.*;
import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a metaball.
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;Metaball strength="0.8" radius="1" center="0, 0, 0"/&gt;
 * </pre>
 * </code>
 * Note that the strength may be negative.
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float   "CDATA"&gt;
 *     &lt;!ENTITY %Point3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT Metaball EMPTY&gt;
 *     &lt;!ATTLIST Metaball
 *         center   %Point3f; "0, 0, 0"
 *         radius   %Float;   "1"
 *         strength %Float;   "1"&gt;
 * </pre>
 * </code>
 */
public class Metaball
        implements Partitionable
{
    Point3f     center;
    BBox        bBox;
    float       radSqu, s;
    boolean     inside;
    int         inside1, inside2;
    Vector3f    v = new Vector3f();
    float       tolerance = 0.0001f;
    Blobby      parent;
    
    /**
     * basic constructor.
     *
     * @param center the center.
     * @param radius the radius.
     * @param strength the strength (can be negative or positive).
     */
    public Metaball(
        Point3f center,
        double  radius,
        double  strength)
    {
        init(center, radius, strength);
    }
    
    private Metaball()
    {
    }

    public static Metaball getInstance(
        RMLReader   in)
        throws IOException
    {
        Metaball        m = new Metaball();
        ParametersMap   params = in.parameters();

        m.init(params.getPoint("center", "0, 0, 0"),
                params.getFloat("radius", "1"),
                params.getFloat("strength", "1"));

        if (in.token() != RMLReader.EMPTY_TAG)
        {
            in.skipToEndTag(in.toString());
        }

        return m;
    }

    private void init(
        Point3f c,
        double  r,
        double  s)
    {
        this.center = c;
        this.s = (float)s;
        this.radSqu = (float)(r * r);
        this.bBox = new BBox(
                    (float)(c.x - r), (float)(c.y - r), (float)(c.z - r),
                    (float)(c.x + r), (float)(c.y + r), (float)(c.z + r));
    }

    public BBox bBox()
    {
        return bBox;
    }
    
    void setParent(
        Blobby  parent)
    {
        this.parent = parent;
    }

    public boolean isIntersected(
        Ray r)
    {
        throw new RuntimeException("not implemented!");
    }
    
    public HitList intersections(
        Ray r)
    {
        //
        // get interval for ri = 0 to R
        //
        double      a, b, c, d;
    
        v.set(r.org);
        v.sub(center);
    
        a = r.dir.dot(r.dir);
        b = -v.dot(r.dir);
    
        if ((b < 0.0) && (a > 0.0))
        {
            return null;
        }
    
        c = v.dot(v) - radSqu;
    
        if ((d = b * b - a * c) < 0.0)
        {
            return null;
        }
    
        d = Math.sqrt(d);
    
        //
        // add in our hits.
        //
        HitList list = null;
        double  t = (b - d) / a;
    
        if (t > tolerance)
        {
            Blobby.MBHit  hit = Blobby.MBHit.getInstance(this, t, false, false);
    
            hit.next = Blobby.MBHit.getInstance(this, ((b + d) / a), false, true);
            parent.mergeHits(hit);
        }
        else if ((t = (b + d) / a) > tolerance)
        { 
            Blobby.MBHit  hit = Blobby.MBHit.getInstance(this, t, true, true);
    
            parent.mergeHits(hit);
        }
    
        return null;
    }
    
    public boolean isShadowing(
        Ray r)
    {
        throw new RuntimeException("not implemented!");
    }
}
