package org.bouncycastle.graphics.rt.objects;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;
    
/**
 * A parametric bicubic surface.
 * <p>
 * A sample piece of RML for a patch:
 * <pre>
 *      &lt;Patch maxSubdivideLevel="6" basis="Bezier"&gt;
 *          &lt;GeomX&gt;
 *              1.400000, 1.337500, 1.437500, 1.500000,
 *              1.400000, 1.337500, 1.437500, 1.500000,
 *              0.784000, 0.749000, 0.805000, 0.840000,
 *              0.000000, 0.000000, 0.000000, 0.000000
 *          &lt;/GeomX&gt;
 *          &lt;GeomY&gt;
 *              2.250000, 2.381250, 2.381250, 2.250000,
 *              2.250000, 2.381250, 2.381250, 2.250000,
 *              2.250000, 2.381250, 2.381250, 2.250000,
 *              2.250000, 2.381250, 2.381250, 2.250000
 *          &lt;/GeomY&gt;
 *          &lt;GeomZ&gt;
 *              0.000000, 0.000000, 0.000000, 0.000000,
 *              0.784000, 0.749000, 0.805000, 0.840000,
 *              1.400000, 1.337500, 1.437500, 1.500000,
 *              1.400000, 1.337500, 1.437500, 1.500000
 *          &lt;/GeomZ&gt;
 *      &lt;/Patch&gt;
 * </pre>
 */
public class Patch
    extends Bounded
    implements NormalCalculator
{
    public static final String  BEZIER = "Bezier";
    public static final String  CARDINAL = "Cardinal";
    public static final String  B_SPLINE = "B-Spline";

    //
    // the matrices to assist in calculating the convex hull for
    // the first half and second half of a cubic curve.
    //
    private static Matrix4f  half1 = new Matrix4f(
        1 / 8.0f,   0.0f,       0.0f,       0.0f,
        0.0f,       1 / 4.0f,   0.0f,       0.0f,
        0.0f,       0.0f,       1 / 2.0f,   0.0f,
        0.0f,       0.0f,       0.0f,       1.0f
    );
    
    private static Matrix4f  half2 = new Matrix4f(
        1 / 8.0f,   0.0f,       0.0f,       0.0f,
        3 / 8.0f,   1 / 4.0f,   0.0f,       0.0f,
        3 / 8.0f,   1 / 2.0f,   1 / 2.0f,   0.0f,
        1 / 8.0f,   1 / 4.0f,   1 / 2.0f,   1.0f
    );
    
    /*
     * basis matrices...
     */
   private static Matrix4f  bezier = new Matrix4f(
        -1.0f,      3.0f,    -3.0f,   1.0f,
        3.0f,       -6.0f,   3.0f,    0.0f,
        -3.0f,      3.0f,    0.0f,    0.0f,
        1.0f,       0.0f,    0.0f,    0.0f
    );
    
    //
    // you could say this one is only for completeness!
    //
   private static Matrix4f  bezierTranspose = new Matrix4f(
        -1.0f,      3.0f,   -3.0f,      1.0f,
        3.0f,       -6.0f,  3.0f,       0.0f,
        -3.0f,      3.0f,   0.0f,       0.0f,
        1.0f,       0.0f,   0.0f,       0.0f
    );
    
    private static Matrix4f  cardinal = new Matrix4f(      // Or catmull-rom
        -0.5f,  1.5f,    -1.5f,   0.5f,
        1.0f,   -2.5f,   2.0f,    -0.5f,
        -0.5f,  0.0f,    0.5f,    0.0f,
        0.0f,   1.0f,    0.0f,    0.0f
    );
    
    private static Matrix4f  cardinalTranspose = new Matrix4f(
        -0.5f,  1.0f,   -0.5f,  0.0f,
        1.5f,   -2.5f,  0.0f,   1.0f,
        -1.5f,  2.0f,   0.5f,   0.0f,
        0.5f,   -0.5f,  0.0f,   0.0f
    );
    
    private static Matrix4f  bspline = new Matrix4f(
        -1.0f / 6.0f,    3.0f / 6.0f,       -3.0f / 6.0f,     1.0f / 6.0f,
        3.0f / 6.0f,     -6.0f / 6.0f,      3.0f / 6.0f,      0.0f,
        -3.0f / 6.0f,    0.0f,              3.0f / 6.0f,      0.0f,
        1.0f / 6.0f,     4.0f / 6.0f,       1.0f / 6.0f,      0.0f
    );
    
    private static Matrix4f  bsplineTranspose = new Matrix4f(
        -1.0f / 6.0f,   3.0f / 6.0f,    -3.0f / 6.0f,   1.0f / 6.0f, 
        3.0f / 6.0f,    -6.0f / 6.0f,   0.0f,           4.0f / 6.0f,
        -3.0f / 6.0f,   3.0f / 6.0f,    3.0f / 6.0f,    1.0f / 6.0f,
        1.0f / 6.0f,    0.0f,           0.0f,           0.0f
    );
    
    private static final byte   PX = 0;
    private static final byte   PY = 1;
    private static final byte   PZ = 2;

    private class SubNode
    {
        BBox    bb = new BBox();

        /**
         * calculate the bounding box for the set of control
         * points described in geomX, geomY, and geomZ.
         */
        public void setBBox(
            Matrix4f    geomX,
            Matrix4f    geomY,
            Matrix4f    geomZ)
        {
            bb.maxX = bb.minX = geomX.m00;
            bb.maxY = bb.minY = geomY.m00;
            bb.maxZ = bb.minZ = geomZ.m00;
        
            for (int i = 0; i != 4; i++)
            {
                for (int j = 0; j != 4; j++)
                {
                    float   val = geomX.getElement(i, j);

                    if (bb.maxX < val)
                    {
                        bb.maxX = val;
                    }
                    else if (bb.minX > val)
                    {
                        bb.minX = val;
                    }

                    val = geomY.getElement(i, j);

                    if (bb.maxY < val)
                    {
                        bb.maxY = val;
                    }
                    else if (bb.minY > val)
                    {
                        bb.minY = val;
                    }

                    val = geomZ.getElement(i, j);

                    if (bb.maxZ < val)
                    {
                        bb.maxZ = val;
                    }
                    else if (bb.minZ > val)
                    {
                        bb.minZ = val;
                    }
                }
            }
        }
    }

    private class BranchNode
        extends SubNode
    {
        byte    type;
        float   splitval;
        SubNode left;
        SubNode right;
    }

    private class LeafNode
        extends SubNode
    {
        float   maxu, midu, minu;
        float   maxv, midv, minv;
    }

    Vector4f    U = new Vector4f();
    Vector4f    V = new Vector4f();
    Vector4f    T1 = new Vector4f();
    Vector4f    T2 = new Vector4f();

    SubNode     tree;
    Matrix4f    basis;
    Matrix4f    basisTranspose;

    //
    // geometry matrices - object space.
    //
    Matrix4f    oGeomX;
    Matrix4f    oGeomY;
    Matrix4f    oGeomZ;

    //
    // geometry matrices - ray space.
    //
    Matrix4f    rGeomX = new Matrix4f();
    Matrix4f    rGeomY = new Matrix4f();
    Matrix4f    rGeomZ = new Matrix4f();

    //
    // u and v values for last hit
    //
    float       hitU, hitV;

    //
    // tolerance for this patch.
    //
    float       tolerance = 0.01f;

    /**
     * evaluate the dot product of the partial derivative of the patch function
     * in v and the ray at the point on the surface (u, v).
     */
    private float eval_fdu(
        float       u,
        float       v,
        Matrix4f    H)
    {
        int     i, j;
    
        U.set(3 * u * u, 2 * u, 1, 0);
    
        //
        //    U' * M
        //
        basis.transform(U);
    
        //
        //    U' * M * H
        //
        H.transform(U);
    
        //
        //     U' * M * H * transpose(M)
        //
        basisTranspose.transform(U);

        //
        // Now return U' * M * H * transpose(M) * V
        //
        return(v * (v * (v * U.x + U.y) + U.z) + U.w);
    }
    
    /**
     * evaluate the dot product of the partial derivative of the patch function
     * in v and the ray at the point on the surface (u, v).
     */
    private float eval_fdv(
        float       u,
        float       v,
        Matrix4f    H)
    {
        int    i, j;
    
        U.w = 1;
        U.z = u;
        U.y = u * u;
        U.x = U.y * u;
    
        //
        //    U * M
        //
        basis.transform(U);
    
        //
        //    U * M * H
        //
        H.transform(U);
    
        //
        //     U * M * H * transpose(M)
        //
        basisTranspose.transform(U);

        /*
         * Now return U * M * H * transpose(M) * V'
         */
        return(v * (3 * v * U.x + 2 * U.y) + U.z);
    }
    
    /**
     * evaluate the dot product of the patch and the ray at the
     * point on the surface (u, v).
     */
    private float eval_f(
        float       u,
        float       v,
        Matrix4f    H)
    {
        int         i, j;
    
        U.w = 1;
        U.z = u;
        U.y = u * u;
        U.x = U.y * u;
    
        //
        //    U * M
        //
        basis.transform(U);
    
        //
        //    U * M * H
        //
        H.transform(U);
    
        //
        //     U * M * H * transpose(M)
        //
        basisTranspose.transform(U);
    
        //
        // Now return U * M * H * transpose(M) * V
        //
        return(v * (v * (v * U.x + U.y) + U.z) + U.w);
    }
    
    //
    // these are common to the top intersection routine and the
    // final one. They only need to be calculated once per ray.
    //
    private Vector3f        P = new Vector3f();
    private Matrix4f        H1 = new Matrix4f();
    private Matrix4f        H2 = new Matrix4f();
    private Vector3f        P1 = new Vector3f();
    private Vector3f        P2 = new Vector3f();
    private float           D1, D2;

    private static float    REL_ERROR = 0.001f;
    
    /*
     * intersect
     *
     *    calculate the rays intersection with the patch.
     */
    private boolean intersect(
        Ray     r,
        float   u,
        float   v)
    {
        int         i, j;
        float       a, b, d;
        float       E1, E2;
        float       x, y, err = Float.MAX_VALUE, lasterr = Float.MAX_VALUE;
        float       E1du, E1dv, E2du, E2dv;
    
        for (i = 0; i != 50; i++)
        {
            E1 = eval_f(u, v, H1) - D1;
            E2 = eval_f(u, v, H2) - D2;
    
            //
            // check if we have converged
            //
            if ((err = Math.abs(E1) + Math.abs(E2)) < REL_ERROR)
            {
                break;
            }
            
            //
            // if this is true, we're not heading for a solution
            //
            if (i > 5 && err > lasterr)
            {
                break;
            }
    
            E1du = eval_fdu(u, v, H1);
            E1dv = eval_fdv(u, v, H1);
    
            E2du = eval_fdu(u, v, H2);
            E2dv = eval_fdv(u, v, H2);
            
            u += (E2 * E1dv - E1 * E2dv) / (E1du * E2dv - E1dv * E2du);
            v += (E2 * E1du - E1 * E2du) / (E1dv * E2du - E1du * E2dv);
    
            lasterr = err;
        }
    
        if (err < REL_ERROR && err < lasterr)
        {
            if (u < 0.0 || u > 1.0)
            {
                return false;
            }
    
            if (v < 0.0 || v > 1.0)
            {
                return false;
            }
    
            //
            // caclulate the t value
            //
            float   t;

            hitU = u;
            hitV = v;
    
            x = Math.abs(r.dir.x);
            y = Math.abs(r.dir.y);
            if (x > y)
            {
                if (x > Math.abs(r.dir.z))
                {
                    t = (eval_f(u, v, rGeomX) - r.org.x) / r.dir.x;
                }
                else
                {
                    t = (eval_f(u, v, rGeomZ) - r.org.z) / r.dir.z;
                }
            }
            else
            {
                if (y > Math.abs(r.dir.z))
                {
                    t = (eval_f(u, v, rGeomY) - r.org.y) / r.dir.y;
                }
                else
                {
                    t = (eval_f(u, v, rGeomZ) - r.org.z) / r.dir.z;
                }
            }
            
            if (t < tolerance)
            {
                return false;
            }
    
            r.setHit(this, t, this, this, null, this.getShader());
    
            return true;
        }

        return false;
    }
    
    /*
     * findnode
     *
     *    find the subdivision branch node closest to the patch, that
     * the ray intersects. We use the u and v values for this node to give
     * a starting value for the root solver.
     */
    private boolean findnode(
        Ray     r,
        SubNode root)
    {
        boolean     hit = false;

        if (root instanceof LeafNode)
        {
            LeafNode    node = (LeafNode)root;

            return(intersect(r, (node.minu + node.maxu) / 2,
                                        (node.minv + node.maxv) / 2));
        }
        else
        {
            BranchNode  node = (BranchNode)root;
            float       org, dir;

            if (node.type == PX)
            {
                org = r.org.x;
                dir = r.dir.x;
            }
            else if (node.type == PY)
            {
                org = r.org.y;
                dir = r.dir.y;
            }
            else
            {
                org = r.org.z;
                dir = r.dir.z;
            }
    
            if (org <= node.splitval)
            {
                if (!node.left.bb.isMissedBy(r))
                {
                    hit = findnode(r, node.left);
                }
                if (!hit && (dir >= 0.0) && !node.right.bb.isMissedBy(r))
                {
                    return(findnode(r, node.right));
                }
            }
            else
            {
                if (!node.right.bb.isMissedBy(r))
                {
                    hit = findnode(r, node.right);
                }
                if (!hit && (dir <= 0.0) && !node.left.bb.isMissedBy(r))
                {
                    return(findnode(r, node.left));
                }
            }
        }
    
        return(hit);
    }
    
    private boolean isObjectIntersected(
        Ray r)
    {
        /*
         * calculate two planes:
         * P1.(x, y, z) = D1 and P2.(x, y, z) = D2
         *
         *    the line along which this intersect represents the ray.
         */
        P1.set(r.org);
        P1.cross(P1, r.dir);
        P1.normalize();
    
        D1 = P1.x * r.org.x + P1.y * r.org.y + P1.z * r.org.z;
    
        P2.cross(P1, r.dir);
        P2.normalize();
    
        D2 = P2.x * r.org.x + P2.y * r.org.y + P2.z * r.org.z;
    
        /*
         * Calculate the coefficients matrices. These are made up of
         * the dot products of the control points and P1 and P2.
         */
        P.set(rGeomX.m00, rGeomY.m00, rGeomZ.m00);
        H1.m00 = P.dot(P1);
        H2.m00 = P.dot(P2);

        P.set(rGeomX.m01, rGeomY.m01, rGeomZ.m01);
        H1.m01 = P.dot(P1);
        H2.m01 = P.dot(P2);

        P.set(rGeomX.m02, rGeomY.m02, rGeomZ.m02);
        H1.m02 = P.dot(P1);
        H2.m02 = P.dot(P2);

        P.set(rGeomX.m03, rGeomY.m03, rGeomZ.m03);
        H1.m03 = P.dot(P1);
        H2.m03 = P.dot(P2);


        P.set(rGeomX.m10, rGeomY.m10, rGeomZ.m10);
        H1.m10 = P.dot(P1);
        H2.m10 = P.dot(P2);

        P.set(rGeomX.m11, rGeomY.m11, rGeomZ.m11);
        H1.m11 = P.dot(P1);
        H2.m11 = P.dot(P2);

        P.set(rGeomX.m12, rGeomY.m12, rGeomZ.m12);
        H1.m12 = P.dot(P1);
        H2.m12 = P.dot(P2);

        P.set(rGeomX.m13, rGeomY.m13, rGeomZ.m13);
        H1.m13 = P.dot(P1);
        H2.m13 = P.dot(P2);


        P.set(rGeomX.m20, rGeomY.m20, rGeomZ.m20);
        H1.m20 = P.dot(P1);
        H2.m20 = P.dot(P2);

        P.set(rGeomX.m21, rGeomY.m21, rGeomZ.m21);
        H1.m21 = P.dot(P1);
        H2.m21 = P.dot(P2);

        P.set(rGeomX.m22, rGeomY.m22, rGeomZ.m22);
        H1.m22 = P.dot(P1);
        H2.m22 = P.dot(P2);

        P.set(rGeomX.m23, rGeomY.m23, rGeomZ.m23);
        H1.m23 = P.dot(P1);
        H2.m23 = P.dot(P2);


        P.set(rGeomX.m30, rGeomY.m30, rGeomZ.m30);
        H1.m30 = P.dot(P1);
        H2.m30 = P.dot(P2);

        P.set(rGeomX.m31, rGeomY.m31, rGeomZ.m31);
        H1.m31 = P.dot(P1);
        H2.m31 = P.dot(P2);

        P.set(rGeomX.m32, rGeomY.m32, rGeomZ.m32);
        H1.m32 = P.dot(P1);
        H2.m32 = P.dot(P2);

        P.set(rGeomX.m33, rGeomY.m33, rGeomZ.m33);
        H1.m33 = P.dot(P1);
        H2.m33 = P.dot(P2);

        return(findnode(r, tree));
    }
    
    public boolean isIntersected(
        Ray r)
    {
        return isObjectIntersected(r);
    }

    public HitList intersections(
        Ray r)
    {
        return null;
    }

    public boolean isShadowing(
        Ray r)
    {
        if (!isObjectIntersected(r))
        {
            return false;
        }

        return (r.hit.t < r.tMax);
    }

    private Vector3f v1 = new Vector3f();
    private Vector3f v2 = new Vector3f();

    /**
     * returns the normal to the patch
     */
    public void getN(
        Hit         hit,
        Vector3f    n)
    {
        v1.x = eval_fdu(hitU, hitV, oGeomX);
        v1.y = eval_fdu(hitU, hitV, oGeomY);
        v1.z = eval_fdu(hitU, hitV, oGeomZ);
    
        v2.x = eval_fdv(hitU, hitV, oGeomX);
        v2.y = eval_fdv(hitU, hitV, oGeomY);
        v2.z = eval_fdv(hitU, hitV, oGeomZ);
    
        n.cross(v1, v2);
    
        n.normalize();
    }
    
/**
    void
    patchc(o, txt, l, n, pcol, type)
        object    *o;
        texture    *txt;
        vector    *l, *n;
        pixel    *pcol;
        int    type;
    {
        int    w, h, indx;
        tiletxt    *tp;
        tp = txt->u.t;
    
        w = o->obj.pch->u * tp->scalew;
        h = o->obj.pch->v * tp->scaleh;
    
        indx = (w % tp->pixw + (h % tp->pixh) * tp->pixw) * 3;
    
        pcol->r = (unsigned char)tp->map[indx] / 255.0;
        pcol->g = (unsigned char)tp->map[indx + 1] / 255.0;
        pcol->b = (unsigned char)tp->map[indx + 2] / 255.0;
    }
*/
    
    
    /*
     * subdivide
     *
     *    subdivide the patch area in root to calculate new geometry
     * points to give the bounding boxes for the two smaller patches.
     * The sudivision keeps going until the level of a leaf node equals
     * maxsublevel.
     */
    public SubNode subdivide(
        SubNode     root,
        Matrix4f    geomX,
        Matrix4f    geomY,
        Matrix4f    geomZ,
        float       minu,
        float       maxu,
        float       minv,
        float       maxv,
        int         maxsublevel,
        int         level)
    {
        if (level == maxsublevel)
        {
            return root;            // finished
        }
    
        SubNode     left, right;
        Matrix4f    ngeomX = new Matrix4f();
        Matrix4f    ngeomY = new Matrix4f();
        Matrix4f    ngeomZ = new Matrix4f();
        Matrix4f    tmp = new Matrix4f();
        float       xgap, ygap, zgap;
        
        BranchNode  nRoot = new BranchNode();

                               // split in v
        if ((level & 1) != 0)
        {
            LeafNode    leaf;

            leaf = new LeafNode();
    
            tmp.set(basis);
            tmp.invert();
            tmp.mul(tmp, half1);
            tmp.mul(tmp, basis);
    
            ngeomX.mul(tmp, geomX);
            ngeomY.mul(tmp, geomY);
            ngeomZ.mul(tmp, geomZ);
    
            leaf.setBBox(ngeomX, ngeomY, ngeomZ);
    
            leaf.minu = minu;
            leaf.maxu = maxu;
            leaf.minv = minv;
            leaf.maxv = (minv + maxv) / 2;
    
            left = subdivide(leaf,
                ngeomX, ngeomY, ngeomZ,
                minu, maxu,
                minv, (maxv + minv) / 2,
                maxsublevel, level + 1);
    
            leaf = new LeafNode();
    
            tmp.set(basis);
            tmp.invert();
            tmp.mul(tmp, half2);
            tmp.mul(tmp, basis);
    
            ngeomX.mul(tmp, geomX);
            ngeomY.mul(tmp, geomY);
            ngeomZ.mul(tmp, geomZ);
    
            leaf.setBBox(ngeomX, ngeomY, ngeomZ);
    
            leaf.minu = minu;
            leaf.maxu = maxu;
            leaf.minv = (minv + maxv) / 2;
            leaf.maxv = maxv;
    
            right = subdivide(leaf,
                ngeomX, ngeomY, ngeomZ,
                minu, maxu,
                (maxv + minv) / 2, maxv,
                maxsublevel, level + 1);
    
        }
        else                // split in u
        {
            LeafNode    leaf;

            leaf = new LeafNode();
    
            tmp.set(basis);
            tmp.invert();
            tmp.mul(tmp, half1);
            tmp.mul(tmp, basis);
    
            tmp.transpose();
    
            ngeomX.mul(geomX, tmp);
            ngeomY.mul(geomY, tmp);
            ngeomZ.mul(geomZ, tmp);
    
            leaf.setBBox(ngeomX, ngeomY, ngeomZ);
    
            leaf.minu = minu;
            leaf.maxu = (minu + maxu) / 2;
            leaf.minv = minv;
            leaf.maxv = maxv;
    
            left = subdivide(leaf,
                ngeomX, ngeomY, ngeomZ,
                minu, (maxu + minu) / 2,
                minv, maxv,
                maxsublevel, level + 1);
    
    
            leaf = new LeafNode();
    
            tmp.set(basis);
            tmp.invert();
            tmp.mul(tmp, half2);
            tmp.mul(tmp, basis);
    
            tmp.transpose();
    
            ngeomX.mul(geomX, tmp);
            ngeomY.mul(geomY, tmp);
            ngeomZ.mul(geomZ, tmp);
    
            leaf.setBBox(ngeomX, ngeomY, ngeomZ);
    
            leaf.minu = (minu + maxu) / 2;
            leaf.maxu = maxu;
            leaf.minv = minv;
            leaf.maxv = maxv;
    
            right = subdivide(leaf,
                ngeomX, ngeomY, ngeomZ,
                (maxu + minu) / 2, maxu,
                minv, maxv,
                maxsublevel, level + 1);
        }

        //
        // recalculate bounding box for root using the bounding boxes of
        // the left and right nodes.
        //
        nRoot.bb.set(left.bb);
        nRoot.bb.adjustFor(right.bb);
    
        //
        // find a splitting plane.
        //
        xgap = Math.abs(left.bb.minX - right.bb.minX);
        ygap = Math.abs(left.bb.minY - right.bb.minY);
        zgap = Math.abs(left.bb.minZ - right.bb.minZ);
    
        if (xgap > ygap)
        {
            if (xgap > zgap)
            {
                nRoot.type = PX;
            }
            else
            {
                nRoot.type = PZ;
            }
        }
        else
        {
            if (ygap > zgap)
            {
                nRoot.type = PY;
            }
            else
            {
                nRoot.type = PX;
            }
        }
    
        nRoot.left = left;
        nRoot.right = right;

        switch (nRoot.type)
        {
        case PX:
            if (left.bb.minX > right.bb.minX)
            {
                nRoot.right = left;
                nRoot.left = right;
            }
            nRoot.splitval = nRoot.right.bb.minX;
            break;
        case PY:
            if (left.bb.minY > right.bb.minY)
            {
                nRoot.right = left;
                nRoot.left = right;
            }
            nRoot.splitval = nRoot.right.bb.minY;
            break;
        case PZ:
            if (left.bb.minZ > right.bb.minZ)
            {
                nRoot.right = left;
                nRoot.left = right;
            }
            nRoot.splitval = nRoot.right.bb.minZ;
            break;
        default:
            throw new IllegalStateException("subdivide: unknown type in switch.");
        }

        return nRoot;
    }
    
    public void transform(
        Matrix4f    trans)
    {
        Vector3f    v = new Vector3f();

        super.transform(trans);

        for (int i = 0; i != 4; i++)
        {
            for (int j = 0; j != 4; j++)
            {
                v.x = oGeomX.getElement(i, j);
                v.y = oGeomY.getElement(i, j);
                v.z = oGeomZ.getElement(i, j);
                objToRay.transform(v);
                rGeomX.setElement(i, j, v.x);
                rGeomY.setElement(i, j, v.y);
                rGeomZ.setElement(i, j, v.z);
            }
        }
    }
    
    public Patch(
        String      basisName,
        Matrix4f    geomX,
        Matrix4f    geomY,
        Matrix4f    geomZ)
    {
        init(basisName, 6, geomX, geomY, geomZ);
    }

    public Patch(
        String      basisName,
        int         maxSubdivisionLevel,
        Matrix4f    geomX,
        Matrix4f    geomY,
        Matrix4f    geomZ)
    {
        init(basisName, maxSubdivisionLevel, geomX, geomY, geomZ);
    }

    private Patch()
    {
    }

    public static Patch getInstance(
        RMLReader       in)
        throws IOException
    {
        Patch   p = new Patch();

        p.readObject(in);

        return p;
    }

    protected void readObject(
        RMLReader   in)
        throws IOException
    {
        ParametersMap   params = in.parameters();

        int     maxSubdivisionLevel = params.getInt("maxSubdivisionLevel", "6");
        String  basisName = params.getString("basis", "Bezier");

        super.readObject(in);

        init(basisName, maxSubdivisionLevel, oGeomX, oGeomY, oGeomZ);
    }

    protected void processObject(
        Object  obj)
    {
        if (obj instanceof XGeometry)
        {
            oGeomX = ((XGeometry)obj).geometry();
        }
        else if (obj instanceof YGeometry)
        {
            oGeomY = ((YGeometry)obj).geometry();
        }
        else if (obj instanceof ZGeometry)
        {
            oGeomZ = ((ZGeometry)obj).geometry();
        }
    }

    protected void init(
        String      basisName,
        int         maxSubdivisionLevel,
        Matrix4f    geomX,
        Matrix4f    geomY,
        Matrix4f    geomZ)
    {
        if (basisName.equals(BEZIER))
        {
            basis = bezier;
            basisTranspose = bezierTranspose;
        }
        else if (basisName.equals(CARDINAL))
        {
            basis = cardinal;
            basisTranspose = cardinalTranspose;
        }
        else if (basisName.equals(B_SPLINE))
        {
            basis = bspline;
            basisTranspose = bsplineTranspose;
        }
    
        this.oGeomX = geomX;
        this.oGeomY = geomY;
        this.oGeomZ = geomZ;

        this.rGeomX.set(oGeomX);
        this.rGeomY.set(oGeomY);
        this.rGeomZ.set(oGeomZ);

        //
        // set up the root node for the patch subdivision tree and
        // calculate the bounding boxes for a few subdivisions of
        // the patch.
        //
    
        LeafNode leaf = new LeafNode();
    
        leaf.setBBox(geomX, geomY, geomZ);
        leaf.minu = 0;
        leaf.maxu = 1;
        leaf.minv = 0;
        leaf.maxv = 1;
    
        //
        // subdivide and get the bounding box...
        //
        tree = subdivide(leaf, geomX, geomY, geomZ,
                            0, 1, 0, 1, maxSubdivisionLevel, 0);

        this.setBBox(tree.bb);
    }
}
