package org.bouncycastle.graphics.rt.objects;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * A point light source. By default it affects all objects in the scene,
 * unless the choice is constrained by the using the addAffected method.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;PointLight location="0, 0, 0" direction="0, 0, -1" color="0.8, 0.7, 1"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float    "CDATA"&gt;
 *     &lt;!ENTITY %Color3f  "CDATA"&gt;
 *     &lt;!ENTITY %Point3f  "CDATA"&gt;
 *
 *     &lt;!ELEMENT PointLight EMPTY&gt;
 *     &lt;!ATTLIST PointLight
 *         color        %Color3f;  "1, 1, 1"
 *         location     %Point3f;  "0, 0, 0"&gt;
 * </pre>
 * </code>
 */
public class PointLight
    implements Light
{
    Point3f    loc;
    Color3f    col;

    boolean    shadows = true;        // whether or not we produce shadows

    HashMap    affected = null;    // table of objects we affect

    public PointLight(
        Point3f    loc,
        Color3f    col)
    {
        this.loc = new Point3f(loc);
        this.col = new Color3f(col);
    }

    private PointLight()
    {
    }

    public static Light getInstance(
        RMLReader       in)
        throws IOException
    {
        ParametersMap   params = in.parameters();

        PointLight  l = new PointLight();

        l.init(
            params.getPoint("location", "0, 0, 0"),
            params.getColor("color", "1, 1, 1"));

        if (in.token() != RMLReader.EMPTY_TAG)
        {
            in.skipToEndTag(in.toString());
        }

        return l;
    }

    private void init(
        Point3f    loc,
        Color3f    col)
    {
        this.loc = new Point3f(loc);
        this.col = new Color3f(col);
    }

    /**
     * set whether or not this light will cast shadows (if not objects
     * in the world are treated as though invisible).
     */
    public void setShadows(
        boolean    shadows)
    {
        this.shadows = shadows;
    }

    /**
     * transform the location of the light.
     */
    public void transform(
        Transform   trans)
    {
        //
    }

    /**
     * add an object as being one of those affected by this light.
     *
     */
    public void addAffected(
        Renderable    obj)
    {
        if (affected == null)
        {
            affected = new HashMap();
        }

        affected.put(obj, obj);
    }

    public LightData getLightData(
        Hit           hit,
        Illuminated   world)
    {
        if ((affected != null) && (affected.get(hit.obj) == null))
        {
            return null;
        }

        //
        // regular processing
        //
        Vector3f        n = hit.N();
        Ray             lr = Ray.getInstance();
        float           dist;
        float           nDotL;

        lr.dir.sub(loc, hit.P());

        dist = (float)Math.sqrt(lr.dir.dot(lr.dir));

        lr.dir.scale(1 / dist);

        if ((nDotL = n.dot(lr.dir)) < 0)
        {
            return null;
        }

        lr.org.set(hit.P());
        lr.orgObj = hit.obj;
        lr.tMax = dist;
        lr.col.set(col);

        if (shadows && world.isShadowing(lr))
        {
            return null;
        }
            
        LightData ld = LightData.getInstance(nDotL, lr.dir, lr.col);

        lr.dispose();

        return ld;
    }
}
