package org.bouncycastle.graphics.rt.objects;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * A polygon. 
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;Polygon color="1,0,0"&gt;
 *        &lt;Vertices&gt;
 *            0.000, 1.000, 0.000,
 *            1.000, 1.000, 0.000,
 *            1.000, 0.000, 0.000,
 *            0.000, 0.000, 0.000
 *        &lt;/Vertices&gt;
 *    &lt;/Polygon&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT Polygon (Vertices)&gt;
 *     &lt;!ATTLIST Polygon
 *         color  %Color3f; "1, 1, 1"&gt;
 * </pre>
 * </code>
 */
public class Polygon
    extends Bounded
    implements NormalCalculator, TextureMapper
{
    static final float TOLERANCE = (float)0.0001;

    static final byte X = 0;
    static final byte Y = 1;
    static final byte Z = 2;

    byte        axis;
    int        nSides;
    Vector3f    n;
    float        constantTerm;

    float[]        xs, ys, zs;
    float[]        uCoords, vCoords;

    float        t;

    float        tolerance;

    public Polygon(
        int         nSides,
        Point3f[]   verts)
    {
        init(nSides, verts);
    }

    private Polygon()
    {
    }

    public static Polygon getInstance(
        RMLReader       in)
        throws IOException
    {
        Polygon p = new Polygon();

        p.readObject(in);

        return p;
    }

    protected void processObject(
        Object          obj)
    {
        if (obj instanceof Vertices)
        {
            Point3f[]   verts = ((Vertices)obj).vertices();

            init(verts.length, verts);
        }
        else if (obj instanceof Normals)
        {
            // setVertexNormals(((Normals)obj).normals(), this.vertexIndexes());
        }
        else if (obj instanceof Colors)
        {
            // setVertexColors(((Colors)obj).colors(), this.vertexIndexes());
        }
        else
        {
            super.processObject(obj);
        }
    }

    private void init(
        int         nSides,
        Point3f[]   verts)
    {
        int        ccount, ncount, dcount;
        float        vx, vz, radius, area;

        if (nSides < 3)
        {
            throw new IllegalArgumentException(
                    "polygon with < 3 sides!");
        }

        this.nSides = nSides;

        xs = new float[nSides];
        ys = new float[nSides];
        zs = new float[nSides];

        for (int i = 0; i != nSides; i++)
        {
            xs[i] = verts[i].x;
            ys[i] = verts[i].y;
            zs[i] = verts[i].z;
        }

        //
        // calculate the surface normal.
        //
        Vector3f    v1 = new Vector3f();
        Vector3f    v2 = new Vector3f();

        n = new Vector3f();

        v1.sub(verts[1], verts[0]);
        v2.sub(verts[2], verts[1]);

        n.cross(v1, v2);

        if (Math.abs(n.x) > Math.abs(n.y))
        {
            if (Math.abs(n.x) > Math.abs(n.z))
            {
                axis = X;
            }
            else
            {
                axis = Z;
            }
        }
        else
        {
            if (Math.abs(n.y) > Math.abs(n.z))
            {
                axis = Y;
            }
            else
            {
                axis = Z;
            }
        }

        if (n.x == 0 && n.y == 0 && n.z == 0)
        {
            throw new IllegalArgumentException(
                "zero length surface normal for polygon");
        }

        if (nSides >= 3)
        {
            switch (axis)
            {
            case X:
                uCoords = ys;
                vCoords = zs;
                break;
            case Y:
                uCoords = xs;
                vCoords = zs;
                break;
            case Z:
                uCoords = xs;
                vCoords = ys;
                break;
            default:
                throw new IllegalStateException(
                        "rend: bad axis in polygon.\n");
            }
        }
        else
        {
            //
            // as this.n isn't normalised the magnitude of the
            // normal in the part of 3 Space we ignore represents
            // the area of the projected triangle. Dividing through
            // by the component takes care of sign trouble we could
            // have in the intersection test so we don't worry about
            // taking the magnitude...
            //
            Matrix3f    m = new Matrix3f();

            switch (axis)
            {
            case X:
                area = n.x;
                m.m00 = (verts[0].y - verts[2].y) / area;
                m.m10 = (verts[1].y - verts[0].y) / area;
                m.m20 = (verts[2].y - verts[1].y) / area;

                m.m01 = (verts[0].z - verts[2].z) / area;
                m.m11 = (verts[1].z - verts[0].z) / area;
                m.m21 = (verts[2].z - verts[1].z) / area;

                m.m02 = verts[2].y * m.m01
                            - verts[2].z * m.m00;
                m.m12 = verts[0].y * m.m11
                            - verts[0].z * m.m10;
                m.m22 = verts[1].y * m.m21
                            - verts[1].z * m.m20;
                break;
            case Y:
                area = n.y;
                m.m00 = (verts[2].x - verts[0].x) / area;
                m.m10 = (verts[0].x - verts[1].x) / area;
                m.m20 = (verts[1].x - verts[2].x) / area;

                m.m01 = (verts[2].z - verts[0].z) / area;
                m.m11 = (verts[0].z - verts[1].z) / area;
                m.m21 = (verts[1].z - verts[2].z) / area;

                m.m02 = verts[2].x * m.m01
                            - verts[2].z * m.m00;
                m.m12 = verts[0].x * m.m11
                            - verts[0].z * m.m10;
                m.m22 = verts[1].x * m.m21
                            - verts[1].z * m.m20;
                break;
            case Z:
                area = n.z;
                m.m00 = (verts[0].x - verts[2].x) / area;
                m.m10 = (verts[1].x - verts[0].x) / area;
                m.m20 = (verts[2].x - verts[1].x) / area;

                m.m01 = (verts[0].y - verts[2].y) / area;
                m.m11 = (verts[1].y - verts[0].y) / area;
                m.m21 = (verts[2].y - verts[1].y) / area;

                m.m02 = verts[2].x * m.m01
                            - verts[2].y * m.m00;
                m.m12 = verts[0].x * m.m11
                            - verts[0].y * m.m10;
                m.m22 = verts[1].x * m.m21
                            - verts[1].y * m.m20;
                break;
            default:
                throw new IllegalStateException(
                        "rend: bad axis in polygon.\n");
            }
        }

        //
        // set up final plane constants
        //
        n.normalize();

        constantTerm = -(n.x * verts[0].x
                    + n.y * verts[0].y + n.z * verts[0].z);

        //
        // calculate the bounding box
        //
        BBox    wBBox = new BBox(verts[0].x, verts[0].y, verts[0].z,
                verts[1].x, verts[1].y, verts[1].z);

        for (int i = 2; i != nSides; i++)
        {
            wBBox.adjustFor(verts[i]);
        }

        tolerance = (float)((TOLERANCE / 2) * Math.sqrt(
               (wBBox.maxX - wBBox.minX) * (wBBox.maxX - wBBox.minX)
             + (wBBox.maxY - wBBox.minY) * (wBBox.maxY - wBBox.minY)
             + (wBBox.maxZ - wBBox.minZ) * (wBBox.maxZ - wBBox.minZ)));

        if ((wBBox.maxX - wBBox.minX) < TOLERANCE)
        {
            wBBox.minX -= tolerance;
            wBBox.maxX += tolerance;
        }

        if ((wBBox.maxY - wBBox.minY) < TOLERANCE)
        {
            wBBox.minY -= tolerance;
            wBBox.maxY += tolerance;
        }

        if ((wBBox.maxZ - wBBox.minZ) < TOLERANCE)
        {
            wBBox.minZ -= tolerance;
            wBBox.maxZ += tolerance;
        }

        //
        // if the polygon is textured we still have to transform our
        // coordinates into some sort of canonical space, so that the
        // texture will be stable as the object moves.
        //
        Vector3f    v = new Vector3f();
        Transform    trans = new Transform();

                        // rotate so n is z axis
        trans.alignZAxis(n.x, n.y, n.z);
                        // put first vertex at origin
        trans.translate(xs[0], ys[0], zs[0]);

        //
        // we now transform our world bounding box to get the values to
        // scale the polygon by (we scale according to the shortest
        // edge) and use that, plus the other transforms to create
        // a general transform to object space.
        //
        BBox    objBox = wBBox.getTransformedBBox(trans);

        trans.setIdentity();

        if ((objBox.maxX - objBox.minX) < (objBox.maxY - objBox.minY))
        {
            trans.scale(objBox.maxX - objBox.minX,
                objBox.maxX - objBox.minX, 1.0);
        }
        else
        {
            trans.scale(objBox.maxY - objBox.minY,
                objBox.maxY - objBox.minY, 1.0);
        }
                        // rotate so n is z axis
        trans.alignZAxis(n.x, n.y, n.z);
                        // put first vertex at origin
        trans.translate(xs[0], ys[0], zs[0]);

        this.transform(trans);

        this.setBBox(wBBox);
    }

    private boolean isPolygonIntersected(
        Ray    r)
    {
        int        u, v, crosses;
        float        pu, pv;

        t = n.dot(r.dir);

        if (t == 0.0)
            return false;

        t = -((n.x * r.org.x + n.y * r.org.y + n.z * r.org.z)
                            + constantTerm) / t;

        if (t < TOLERANCE)
        {
            return false;
        }

        switch (axis)
        {
        case X:
            pu = r.org.y + t * r.dir.y;
            if (pu < bBox.minY || pu > bBox.maxY)
            {
                return false;
            }
            pv = r.org.z + t * r.dir.z;
            if (pv < bBox.minZ || pv > bBox.maxZ)
            {
                return false;
            }
            break;
        case Y:
            pu = r.org.x + t * r.dir.x;
            if (pu < bBox.minX || pu > bBox.maxX)
            {
                return false;
            }
            pv = r.org.z + t * r.dir.z;
            if (pv < bBox.minZ || pv > bBox.maxZ)
            {
                return false;
            }
            break;
        case Z:
            pu = r.org.x + t * r.dir.x;
            if (pu < bBox.minX || pu > bBox.maxX)
            {
                return false;
            }
            pv = r.org.y + t * r.dir.y;
            if (pv < bBox.minY || pv > bBox.maxY)
            {
                return false;
            }
            break;
        default:
            throw new IllegalStateException("bad axis in polygon.");
        }

        crosses = 0;

        for (int cp = 0, np = nSides - 1; np > -1; cp = np--)
        {
            //
            // If we are both above, or both below the intersection
            // point, go onto the next one.
            //

            if (vCoords[cp] < pv)
            {
                if (vCoords[np] < pv)
                {
                    continue;
                }
            }
            else
            {
                if (vCoords[np] >= pv)
                {
                    continue;
                }
            }

            //
            // Find out if we have crossed in the negative.
            //

            if (uCoords[cp] < pu)
            {
                if (uCoords[np] < pu)
                {            // crossed on left
                    crosses++;
                    continue;
                }
            }
            else if (uCoords[np] >= pu)
            {            // both to right
                continue;
            }

            // 
            // We calculate the u at the intersection of the
            // half-ray and the edge and test it against pu. 
            //
            float    m;

            m = (uCoords[np] - uCoords[cp])
                    / (vCoords[np] - vCoords[cp]);
            if (((uCoords[np] - pu) - m * (vCoords[np] - pv)) < 0)
            {
                crosses++;
            }
        }

        if ((crosses & 1) == 0)
        {
            return false;
        }

        return true;
    }

    public boolean isShadowing(
        Ray    r)
    {
        if (r.orgObj == this)
        {
            return false;
        }

        if (!isPolygonIntersected(r))
        {
            return false;
        }

/*
        if (Vrt_IsSeeThrough(surface))
        {
            // r->dropOff += (t2 - t1) / somefalloff; 

            return VRT_TRUE;
        }
*/
        return (t < r.tMax);
    }

    public HitList intersections(
        Ray         r)
    {
        if (!isPolygonIntersected(r))
        {
            return null;
        }

        HitList list = HitList.getInstance();

        list.add(r, this, r.hit.t, this, this, this, this.getShader());

        return list;
    }

    public boolean isIntersected(
        Ray    r)
    {
        if (!isPolygonIntersected(r))
        {
            return false;
        }

        if (t > tolerance)
        {
            Ray nr = this.createObjectSpaceRay(r);

            r.setHit(this, t, this, this, this, this.getShader());

            nr.dispose();
            return true;
        }

        return false;
    }

    public void getN(
        Hit         hit,
        Vector3f    n)
    {
        n.set(0, 0, 1);
    }

    public boolean getST(
        Tuple3f l,
        float[] uv)
    {
        uv[S] = l.x;
        uv[T] = l.y;

        return true;
    }
}
