package org.bouncycastle.graphics.rt.objects;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.mappers.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a ring.
 */
public class Ring
    extends Bounded
    implements NormalCalculator
{
    static final float TOL = (float)0.0001;

    static final TextureMapper    canonicalRingMap = new DiskMap();

    float    innerR;
    float    innerRSquared;

    float    t;

    float    tolerance;

    TextureMapper    stMapper = canonicalRingMap;

    static BBox    canonicalBBox = new BBox(-1, -1, -TOL, 1, 1, TOL);

    public Ring(
        double        rRing,
        double        rInner)
    {
        innerR = (float)(rRing / rRing);
        innerRSquared = (float)((rInner / rRing) * (rInner / rRing));

        this.setBBox(canonicalBBox);

        Transform   trans = new Transform();

        trans.scale((float)rRing, (float)rRing, (float)rRing);

        this.transform(trans);

        tolerance = TOL;
    }

    private boolean isRingIntersected(
        Ray    r)
    {
        Point3f     org = r.org;
        Vector3f    dir = r.dir;

        if (dir.z == 0)
        {
            return false;
        }

        t = -org.z / dir.z;

        if (t < TOL)
        {
            return false;
        }

        //
        // check to see if we are within the opaque part of the
        // ring when we hit the (x, y)-plan.
        //
        float    x = t * dir.x + org.x;
        float    y = t * dir.y + org.y;
        float    prod = x * x + y * y;
        
        if (prod > 1.0)
        {
            return false;
        }
    
        if (prod < innerRSquared)
        {
            return false;
        }

        return true;
    }

    public boolean isShadowing(
        Ray    r)
    {
        if (r.orgObj == this)
        {
            return false;
        }

        Ray     nr = this.createObjectSpaceRay(r);

        if (!isRingIntersected(nr))
        {
            nr.dispose();

            return false;
        }

        nr.dispose();

        return (t < r.tMax);
    }

    public HitList intersections(
        Ray         r)
    {
        Ray     nr = this.createObjectSpaceRay(r);

        if (!isRingIntersected(r))
        {
            nr.dispose();

            return null;
        }

        HitList list = HitList.getInstance();

        list.add(r, this, r.hit.t, this, this, stMapper, this.getShader());

        nr.dispose();

        return list;
    }

    public boolean isIntersected(
        Ray    r)
    {
        Ray     nr = this.createObjectSpaceRay(r);

        if (!isRingIntersected(nr))
        {
            nr.dispose();

            return false;
        }

        if (t > tolerance)
        {
            r.setHit(this, t, this, this, stMapper, this.getShader());

            nr.dispose();

            return true;
        }

        nr.dispose();

        return false;
    }

    public void getN(
        Hit         hit,
        Vector3f    n)
    {
        n.set(0, 0, 1);
    }
}
