package org.bouncycastle.graphics.rt.objects;

import java.util.*;
import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.mappers.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a sphere.
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;Sphere color="1, 1, 1" radius="1" center="0, 0, 0"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Float   "CDATA"&gt;
 *     &lt;!ENTITY %Point3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT Sphere EMPTY&gt;
 *     &lt;!ATTLIST Sphere
 *         color  %Color3f; "1, 1, 1"
 *         radius %Float;   "1"
 *         center %Point3f; "0, 0, 0"&gt;
 * </pre>
 * </code>
 */
public class Sphere
    extends SimpleObject
    implements NormalCalculator
{
    static final float      TOL = (float)0.0001;

    static BBox             canonicalBBox = new BBox(-1, -1, -1, 1, 1, 1);

    private TextureMapper   stMapper = SphericalMap.canonicalMap;

    private Sphere()
    {
    }

    public Sphere(
        Point3d       c,
        double        r)
    {
        init(new Point3f(c), r);
    }

    public Sphere(
        Point3f        c,
        double        r)
    {
        init(c, r);
    }

    public static Sphere getInstance(
        RMLReader   in)
        throws IOException
    {
        Sphere  s = new Sphere();

        s.readObject(in);

        return s;
    }

    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        init(params.getPoint("center", "0, 0, 0"),
                params.getFloat("radius", "1"));
    }

    private void init(
        Point3f        c,
        double        r)
    {
        setBBox(canonicalBBox);
        setSelfShadowing(false);

        Transform   trans = new Transform();

        trans.scale(r, r, r);

        trans.translate(c.x, c.y, c.z);

        this.transform(trans);

        tolerance = TOL;
    }

    protected boolean isObjectIntersected(
        Ray    r)
    {
        float       a, b, c, d;
        Point3f     org;
        Vector3f    dir;

        org = r.org;
        dir = r.dir;

        a = dir.x * dir.x + dir.y * dir.y 
            + dir.z * dir.z;

        b = -(dir.x * org.x + dir.y * org.y 
            + dir.z * org.z);

        if (b < 0 && a > 0)
        {
            return false;
        }

        c = (org.x * org.x + org.y * org.y 
            + org.z * org.z) - 1;

        d = b * b - a * c;

        if (d < 0)
        {
            return false;
        }
    
        d = (float)Math.sqrt(d);

        t1 = ((b - d) / a);
        t2 = ((b + d) / a);

        if (t2 < tolerance)
        {
            return false;
        }

        enNorm = exNorm = this;
        enShade = exShade = this.getShader();
        enMap = exMap = stMapper;

        return true;
    }

    public void getN(
        Hit         hit,
        Vector3f    n)
    {
        n.set(hit.oP());
    }
}
