package org.bouncycastle.graphics.rt.objects;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.mappers.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * A superquadric.
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;SuperQuadric color="1, 1, 1" order="1" vertex1="0, 0, 0" vertex2="1, 1, 1"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Point3f "CDATA"&gt;
 *     &lt;!ENTITY %Integer "CDATA"&gt;
 *
 *     &lt;!ELEMENT SuperQuadric EMPTY&gt;
 *     &lt;!ATTLIST SuperQuadric
 *         color   %Color3f; "1, 1, 1"
 *         order   %Integer; "1"
 *         vertex1 %Point3f; "0, 0, 0"
 *         vertex2 %Point3f; "1, 1, 1"&gt;
 * </pre>
 * </code>
 */
public class SuperQuadric
    extends SimpleObject
    implements NormalCalculator
{
    /*
     * max allowed value for diff (in rescaled space)
     */
    static final float MAXDIFF    = (float)2.0;

    static final float TOLERANCE    = (float)0.00002;
    static final float TOL         = (float)0.0001;

    public static final BBox canonicalBBox = new BBox(-1, -1, -1, 1, 1, 1);

    int    order = 1;

    public SuperQuadric(
        int        order,
        Point3d    p1,
        Point3d    p2)
    {
        init(order, new Point3f(p1), new Point3f(p2));
    }

    public SuperQuadric(
        int        order,
        Point3f    p1,
        Point3f    p2)
    {
        init(order, p1, p2);
    }

    private SuperQuadric()
    {
    }

    public static SuperQuadric getInstance(
        RMLReader       in)
        throws IOException
    {
        SuperQuadric    s = new SuperQuadric();

        s.readObject(in);

        return s;
    }

    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        Point3f p1 = params.getPoint("vertex1", "0, 0, 0");
        Point3f p2 = params.getPoint("vertex2", "1, 1, 1");
        int     order = params.getInt("order", "1");

        init(order, p1, p2);
    }

    private void init(
        int        order,
        Point3f    p1,
        Point3f    p2)
    {
        int        first;
        float        xLen, yLen, zLen;

        this.order = order;

        this.setBBox(canonicalBBox);

        Transform   trans = new Transform();

        trans.scale(            // by length
            Math.abs(p1.x - p2.x) / 2,
            Math.abs(p1.y - p2.y) / 2,
            Math.abs(p1.z - p2.z) / 2);

        trans.translate(        // by center
            (p1.x + p2.x) / 2,
            (p1.y + p2.y) / 2,
            (p1.z + p2.z) / 2);

        this.transform(trans);

        tolerance = TOLERANCE;
    }

    /**
     *    evaluate the superquadric equation for a given r and t
     */
    private float eval(
        Point3f        org,
        Vector3f    dir,
        float        t)
    {
        return (float)(Math.pow(Math.abs(org.x + dir.x * t), order)
                 + Math.pow(Math.abs(org.y + dir.y * t), order)
                 + Math.pow(Math.abs(org.z + dir.z * t), order)
                 - 1);
    }

    
    private float getTValue(
        Point3f        org,
        Vector3f    dir,
        float        t,
        float        oldT,
        float        supTolerance,
        float        mDiff)
    {
        float        val, oldVal; 

        val = eval(org, dir, oldT);

        while (val > supTolerance)
        {
            float    diff;

            oldVal = val;
            val = eval(org, dir, t);

            if (val >= oldVal)
            {
                return 0;
            }

            diff = (val * (t - oldT)) / (val - oldVal);

            if (Math.abs(diff) > mDiff )
            {
                return 0;
            }

            oldT = t;

            t -= diff;
        }

        return t;
    }

    protected boolean isObjectIntersected(
        Ray        r)
    {
        float           oldT, mdiff;
        Point3f         org = r.org;
        Vector3f        dir = r.dir;
        float[]         tvals = new float[2];

        if (!canonicalBBox.isIntersecting(org, dir, tvals))
        {
            return false;
        }

        t1 = tvals[0];
        t2 = tvals[1];

        //
        // the division by length gives an estimate of the superquadric
        // scaling, assuming that the original ray direction vector is
        // normalized (this assumption is true).
        //
        mdiff = MAXDIFF / dir.length();

        //
        // compute val only if we are near the surface 
        //
        if (Math.abs(t1) < mdiff)
        {
            float    val;

            val = eval(org, dir, t1);

            if (val <= TOL)
            {            // we are on the surface 
                oldT = t1 + TOL;
            }
            else
            {            // we are heading towards it
                oldT = t1 - TOL;
            }
        }
        else
        {
            oldT = t1 - TOL;    // we are heading towards it
        }

        //
        // now calculate an adjusted tolerance
        //
        float supTolerance;

        supTolerance = t1 * TOLERANCE;

        //
        // make sure it's not too small
        //
        if (supTolerance < TOLERANCE)
        {
            supTolerance = TOLERANCE;
        }

        if (t1 > supTolerance)
        {
            t1 = getTValue(org, dir, t1, oldT, supTolerance, mdiff);
        }

        t2 = getTValue(org, dir, t2, t2 + TOL, supTolerance, mdiff);

        if (t2 > tolerance)
        {
            enNorm = exNorm = this;
            enShade = exShade = this.getShader();
            enMap = exMap = SphericalMap.canonicalMap;
            
            return true;
        }

        return false;
    }

    public void getN(
        Hit         hit,
        Vector3f    n)
    {
        Point3f    loc = hit.oP();

        n.set(
            (float)((loc.x > 0) ? Math.pow(loc.x, order - 1)
                    : -Math.pow(-loc.x, order - 1)),
            (float)((loc.y > 0) ? Math.pow(loc.y, order - 1)
                    : -Math.pow(-loc.y, order - 1)),
            (float)((loc.z > 0) ? Math.pow(loc.z, order - 1)
                    : -Math.pow(-loc.z, order - 1)));
    }
}
