package org.bouncycastle.graphics.rt.objects;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.math.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.mappers.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * A torus. Canonical form lies in the (x, y) plan.
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;orus ringRadius="1" pipeRadius="0.1"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Float "CDATA"&gt;
 *
 *     &lt;!ELEMENT Torus EMPTY&gt;
 *     &lt;!ATTLIST Torus
 *         color          %Color3f; "1, 1, 1"
 *         ringRadius     %Float;   "1"
 *         pipeRadius     %Float;   "0.1"&gt;
 * </pre>
 * </code>
 */
public class Torus
    extends Bounded
    implements NormalCalculator
{
    static final double TOL = 0.0001;

    private static final BBox canonicalBBox = new BBox(-2, -2, -2, 2, 2, 2);

    double[]        x2 = new double[3];
    double[]        y2 = new double[3];
    double[]        z2 = new double[3];
    double[]        coef = new double[5];

    double[]            roots = new double[4];
    float[]             tVals = new float[2];
    float[]             dists = new float[4];

    double              constTerm;

    double              tolerance = TOL;

    QuarticSturmSolver  solver = new QuarticSturmSolver();

    TextureMapper       stMapper;

    public Torus(
        double        rRing,
        double        rPipe)
    {
        init(rRing, rPipe);
    }

    private Torus()
    {
    }

    public static Torus getInstance(
        RMLReader       in)
        throws IOException
    {
        Torus   t = new Torus();

        t.readObject(in);

        return t;
    }

    protected void processParameters(
        ParametersMap   params)
    {
        super.processParameters(params);

        init(params.getFloat("ringRadius", "1"),
                    params.getFloat("pipeRadius", "0.1"));
    }

    private void init(
        double        rRing,
        double        rPipe)
    {
        double      radius;
        BBox        b;

        constTerm = 1.0 - ((rPipe / rRing) * (rPipe / rRing));

        radius = 1.0 + rPipe / rRing;

        this.setBBox(-(float)radius, -(float)radius, -(float)radius,
                (float)radius, (float)radius, (float)radius);

        Transform   trans = new Transform();

        trans.scale(rRing, rRing, rRing);

        this.transform(trans);

        stMapper = new ToroidalMap(rRing, rPipe);
    }

    private void initPoly(
        Point3f         org,
        Vector3f        dir,
        double          baseT)
    {
        double        o;

        //
        // we push the ray to the edge of the bounding box to
        // improve the accuracy of the distance calculation.
        //
        o = org.x + baseT * dir.x;

        x2[0] = o * o;
        x2[1] = 2.0 * o * dir.x;
        x2[2] = dir.x * dir.x;

        o = org.y + baseT * dir.y;

        y2[0] = o * o;
        y2[1] = 2.0 * o * dir.y;
        y2[2] = dir.y * dir.y;

        o = org.z + baseT * dir.z;

        z2[0] = o * o;
        z2[1] = 2.0 * o * dir.z;
        z2[2] = dir.z * dir.z;

        coef[4] = 0.0;
        coef[3] = 0.0;
        coef[2] = -4.0 * (x2[2] + y2[2]);
        coef[1] = -4.0 * (x2[1] + y2[1]);
        coef[0] = -4.0 * (x2[0] + y2[0]) + constTerm * constTerm;

        for (int i = 0; i <= 2; i++)
        {
            for (int j = 0; j <= 2; j++)
            {
                coef[i + j] += x2[i] * x2[j] + y2[i] * y2[j]
                    + z2[i] * z2[j]
                      + 2 * (x2[i] * y2[j] + y2[i] * z2[j]
                            + x2[i] * z2[j]);
            }

            coef[i] += 2 * constTerm * (x2[i] + y2[i] + z2[i]);
        }
    }
        
    private int torusIntersect(
        Ray         r,
        float[]     dists,
        boolean     findAllRoots)
    {
        int        np, numroots, atmin, atinf;
        double     baseT;
        int        its;
        float      t, min, max;
        Point3f    org;
        Vector3f   dir;

        org = r.org;
        dir = r.dir;

        if (!canonicalBBox.isIntersecting(org, dir, tVals))
        {
            return 0;
        }

        // only if outside... hmmm is this right?

        if (tVals[0] > 2.0 + TOL)
        {
            baseT = tVals[0] - TOL;
        }
        else
        {
            baseT = 0;
        }

        initPoly(org, dir, baseT);

        //
        // if all the coefficients are > 0.0 we can't have any
        // positive roots!
        //
        if (coef[0] > 0.0 && coef[1] > 0.0
            && coef[2] > 0.0 && coef[3] > 0.0 && coef[4] > 0.0)
        {
            return 0;
        }

        solver.init(coef);

        if (!findAllRoots)
        {
            if (!solver.findRoot(tolerance, tVals[1], roots))
            {
                return 0;
            }

            dists[0] = (float)(baseT + roots[0]);

            return 1;
        }
        else
        {
            int     nRoots = solver.findRoots(tolerance, tVals[1], roots);

            if (nRoots == 0)
            {
                return 0;
            }

            for (int i = 0; i != nRoots; i++)
            {
                dists[i] = (float)(baseT + roots[i]);
            }

            return nRoots;
        }
    }

    public boolean isShadowing(
        Ray     r)
    {
        Ray     nr = this.createObjectSpaceRay(r);

        if (torusIntersect(nr, dists, false) == 0)
        {
            nr.dispose();

            return false;
        }

        nr.dispose();

        return true;
    }

    public HitList intersections(
        Ray     r)
    {
        Ray     nr = this.createObjectSpaceRay(r);

        int nRoots = torusIntersect(nr, dists, true);

        if (nRoots == 0)
        {
            nr.dispose();

            return null;
        }

        HitList list = null;

        for (int i = 0; i != nRoots; i++)
        {
            if (dists[i] > tolerance)
            {
                if (list == null)
                {
                    list = HitList.getInstance();
                }

                list.add(r, this, dists[i], this, this, stMapper, this.getShader());
            }
        }

        nr.dispose();

        return list;
    }

    public boolean isIntersected(
        Ray     r)
    {
        Ray     nr = this.createObjectSpaceRay(r);

        if (torusIntersect(nr, dists, false) == 0)
        {
            nr.dispose();

            return false;
        }

        r.setHit(this, dists[0], this, this, stMapper, this.getShader());

        nr.dispose();

        return true;
    }

    public void getN(
        Hit         hit,
        Vector3f    n)
    {
        float       fact;
        double      constTerm = this.constTerm;
        Point3f     loc = hit.oP();

        fact = (float)(loc.x * loc.x + loc.y * loc.y
                    + loc.z * loc.z + constTerm);

        n.set(loc.x * (fact - 2), loc.y * (fact - 2), loc.z * fact);
    }
}
