package org.bouncycastle.graphics.rt.objects;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a triangle strip.
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;TriangleStrip color="1.000,0.000,0.000"&gt;
 *        &lt;Vertices&gt;
 *            0.000, 1.000, 0.000,
 *            1.000, 1.000, 0.000,
 *            0.000, 0.000, 0.000,
 *            1.000, 0.000, 0.000
 *        &lt;/Vertices&gt;
 *    &lt;/TriangleStrip&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Boolean "(true|false)"&gt;
 *
 *     &lt;!ELEMENT TriangleStrip (Vertices, Normals?, Colors?)&gt;
 *     &lt;!ATTLIST TriangleStrip
 *         color  %Color3f; "1, 1, 1"
 *         smooth %Float;   "false"&gt;
 * </pre>
 * </code>
 */
public class TriangleStrip
    extends IndexedMesh
{
    public TriangleStrip(
        Point3f[]   vertices,
        boolean     smooth)
    {
        init(vertices, smooth);
    }

    public TriangleStrip(
        Point3f[]   vertices,
        Vector3f[]  normals,
        boolean     smooth)
    {
        init(vertices, smooth);

        if (normals != null)
        {
            setVertexNormals(normals, this.vertexIndexes());
        }
    }

    public TriangleStrip(
        Point3f[]   vertices,
        Vector3f[]  normals,
        Color3f[]   colors,
        boolean     smooth)
    {
        init(vertices, smooth);

        if (normals != null)
        {
            setVertexNormals(normals, this.vertexIndexes());
        }

        if (colors != null)
        {
            setVertexColors(colors, this.vertexIndexes());
        }
    }

    private TriangleStrip()
    {
    }

    public static TriangleStrip getInstance(
        RMLReader       in)
        throws IOException
    {
        TriangleStrip strip = new TriangleStrip();

        strip.readObject(in);

        return strip;
    }

    protected void processObject(
        Object          obj)
    {
        if (obj instanceof Vertices)
        {
            init(((Vertices)obj).vertices(), smooth);
        }
        else if (obj instanceof Normals)
        {
            setVertexNormals(((Normals)obj).normals(), this.vertexIndexes());
        }
        else if (obj instanceof Colors)
        {
            setVertexColors(((Colors)obj).colors(), this.vertexIndexes());
        }
    }

    protected final void init(
        Point3f[]   vertices,
        boolean     smooth)
    {
        int     connections = vertices.length - 2;
        short   indexes[][] = new short[connections][];

        for (int i = 0; i != connections; i++)
        {
            indexes[i] = new short[3];

            if ((i & 1) != 0)
            {
                indexes[i][2] = (short)i;
                indexes[i][1] = (short)(i + 1);
                indexes[i][0] = (short)(i + 2);
            }
            else
            {
                indexes[i][0] = (short)i;
                indexes[i][1] = (short)(i + 1);
                indexes[i][2] = (short)(i + 2);
            }
        }

        super.init(vertices, indexes, smooth);
    }
}
