package org.bouncycastle.graphics.rt.renderers;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.objects.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * the standard renderer class.
 */
public class StdRenderer
    extends Renderer
{
    private boolean useGaussian = false;

    public StdRenderer()
    {
    }

    public StdRenderer(
        RMLReader   in)
    {
        ParametersMap   params = in.parameters();

        setImageSize(params.getInt("width", "256"),
                            params.getInt("height", "256"));

        if (params.get("supersampling") != null)
        {
            String   sampling = (String)params.get("supersampling");

            setSupersampling(Integer.parseInt(sampling.substring(0, sampling.indexOf('x'))));
        }

        setJittering(params.getBoolean("jittering", "true"));
        setGaussian(params.getBoolean("gaussian", "false"));
    }

    /**
     * set whether, or not, to clamp (clamping is the process
     * of truncating overilluminated pixels back to 1).
     *
     * @param clamp true if we should, false otherwise.
     */
    public void setClamp(
        boolean    clamp)
    {
        this.clamp = clamp;
    }

    /**
     * set the super sampling rate - basically this is a number which is then
     * used to fetch a sampling grid that is sampling by sampling in size.
     *
     * @param sampling the edge length for the grid.
     */
    public void setSupersampling(
        int    sampling)
    {
        this.samplesPerPixel = sampling * sampling;

        if (samplesPerPixel != 1)
        {
            this.sampleMask = SampleMask.getSquareSampleMask(samplesPerPixel);
        }
    }

    /**
     * set wether or not to apply a gaussian filter over the scanlines
     * before writing them out to the ouput.
     */
    public void setGaussian(
        boolean    useGaussian)
    {
        this.useGaussian = useGaussian;
    }

    public void setJittering(
        boolean    jitter)
    {
        this.jitter = jitter;
    }

    public float getRandNum()
    {
        return rand.getRandNum();
    }

    public int getImageWidth()
    {
        return endX - startX;
    }

    public int getImageHeight()
    {
        return endY - startY;
    }

    public int getImageDepth()
    {
        return 24;
    }

    /**
     * applies a gaussian kernel:
     * <pre>
     *      1 2 1
     *      2 5 2
     *      1 2 1
     * </pre>
     * over the three input scan lines, returning the result in the
     * result scan line.
     */
    private void applyGaussian(
        byte[]  result,
        int     pixSize,
        byte[]  scan1,
        byte[]  scan2,
        byte[]  scan3)
    {
        if (scan1 == null)
        {
            for (int i = 0; i != pixSize; i++)
            {
                result[i] = (byte)((5 * (scan2[i] & 0xff) + 2 * (scan2[i + pixSize] & 0xff)
                    + 2 * (scan3[i] & 0xff) + (scan3[i + pixSize] & 0xff)) / 13);
            }
            for (int i = pixSize; i != result.length - pixSize; i++)
            {
                result[i] = (byte)(((scan3[i - pixSize] & 0xff) + 2 * (scan3[i] & 0xff) + (scan3[i + pixSize] & 0xff)
                    + 2 * (scan2[i - pixSize] & 0xff) + 5 * (scan2[i] & 0xff) + 2 * (scan2[i + pixSize] & 0xff)) / 13);
            }
            for (int i = result.length - pixSize; i != result.length; i++)
            {
                result[i] = (byte)((5 * (scan2[i] & 0xff) + 2 * (scan2[i - pixSize] & 0xff)
                    + 2 * (scan3[i] & 0xff) + (scan3[i - pixSize] & 0xff)) / 13);
            }
        }
        else if (scan3 == null)
        {
            for (int i = 0; i != pixSize; i++)
            {
                result[i] = (byte)((2 * (scan1[i] & 0xff) + (scan1[i + pixSize] & 0xff)
                    + 5 * (scan2[i] & 0xff) + 2 * (scan2[i + pixSize] & 0xff)) / 13);
            }
            for (int i = pixSize; i != result.length - pixSize; i++)
            {
                result[i] = (byte)(((scan1[i - pixSize] & 0xff) + 2 * (scan1[i] & 0xff) + (scan1[i + pixSize] & 0xff)
                    + 2 * (scan2[i - pixSize] & 0xff) + 5 * (scan2[i] & 0xff) + 2 * (scan2[i + pixSize] & 0xff)) / 13);
            }
            for (int i = result.length - pixSize; i != result.length; i++)
            {
                result[i] = (byte)((2 * (scan1[i] & 0xff) + (scan1[i - pixSize] & 0xff)
                    + 5 * (scan2[i] & 0xff) + 2 * (scan2[i - pixSize] & 0xff)) / 13);
            }
        }
        else
        {
            for (int i = 0; i != pixSize; i++)
            {
                result[i] = (byte)((2 * (scan1[i] & 0xff) + (scan1[i + pixSize] & 0xff)
                    + 5 * (scan2[i] & 0xff) + 2 * (scan2[i + pixSize] & 0xff)
                    + 2 * (scan3[i] & 0xff) + (scan3[i + pixSize] & 0xff)) / 13);
            }
            for (int i = pixSize; i != result.length - pixSize; i++)
            {
                result[i] = (byte)(((scan1[i - pixSize] & 0xff) + 2 * (scan1[i] & 0xff) + (scan1[i + pixSize] & 0xff)
                    + 2 * (scan2[i - pixSize] & 0xff) + 5 * (scan2[i] & 0xff) + 2 * (scan2[i + pixSize] & 0xff)
                    + (scan3[i - pixSize] & 0xff) + 2 * (scan3[i] & 0xff) + (scan3[i + pixSize] & 0xff)) / 17);
            }
            for (int i = result.length - pixSize; i != result.length; i++)
            {
                result[i] = (byte)((2 * (scan1[i] & 0xff) + (scan1[i - pixSize] & 0xff)
                    + 5 * (scan2[i] & 0xff) + 2 * (scan2[i - pixSize] & 0xff)
                    + 2 * (scan3[i] & 0xff) + (scan3[i - pixSize] & 0xff)) / 13);
            }
        }
    }

    public void doRender(
        int             format,
        OutputStream    out)
        throws IOException
    {
        if (useGaussian)
        {
            int     size;
            byte[]  scan1;
            byte[]  scan2;
            byte[]  scan3;
            byte[]  scanLine;

            if (format == RGB)
            {
                size = 3;
                scan1 = new byte[(endX - startX) * 3];
                scan2 = new byte[(endX - startX) * 3];
                scan3 = new byte[(endX - startX) * 3];
                scanLine = new byte[(endX - startX) * 3];
            }
            else
            {
                size = 4;
                scan1 = new byte[(endX - startX) * 4];
                scan2 = new byte[(endX - startX) * 4];
                scan3 = new byte[(endX - startX) * 4];
                scanLine = new byte[(endX - startX) * 4];
            }

            renderLineByLine(format, scan1);
            renderLineByLine(format, scan2);

            applyGaussian(scanLine, size, null, scan1, scan2);
            out.write(scanLine, 0, scanLine.length);

            while (renderLineByLine(format, scan3))
            {
                applyGaussian(scanLine, size, scan1, scan2, scan3);
                out.write(scanLine, 0, scanLine.length);

                byte[]  tmp = scan1;

                scan1 = scan2;
                scan2 = scan3;
                scan3 = tmp;
            }

            applyGaussian(scanLine, size, scan1, scan2, null);
            out.write(scanLine, 0, scanLine.length);
        }
        else
        {
            byte[]  scanLine;

            if (format == RGB)
            {
                scanLine = new byte[(endX - startX) * 3];
            }
            else
            {
                scanLine = new byte[(endX - startX) * 4];
            }

            while (renderLineByLine(format, scanLine))
            {
                out.write(scanLine, 0, scanLine.length);
            }
        }
    }

    private float clamp(
        float    val)
    {
        if (val > 1)
        {
            return 1;
        }

        return val;
    }

    public boolean renderLineByLine(
        int             format,
        byte[]          out)
        throws IOException
    {
        Pixel            pixel = new Pixel();

        if (currentY == startY)
        {
            currentY = endY;
            return false;
        }

        if (currentY == endY)
        {
            camera.configureLenseMask(samplesPerPixel, jitter);
        }

        for (int j = startX; j != endX; j++)
        {
            Ray    r;
            float    fRed, fGreen, fBlue, fAlpha;

            fRed = fBlue = fGreen = fAlpha = 0;

            for (int sampleNo = 0;
                sampleNo != samplesPerPixel; sampleNo++)
            {
                float    k1, k2;

                currentTimeSlice =
                    (int)((((float)sampleNo
                            + rand.getRandNum())
                        * maxTimeSamples) /
                        samplesPerPixel);

                k1 = j;
                k2 = currentY;

                if (sampleMask != null)
                {
                    k1 = k1 + sampleMask.points[sampleNo].x;
                    k2 = k2 + sampleMask.points[sampleNo].y;
                    if (jitter)
                    {
                        k1 += sampleMask.magnitude
                          * (1 - rand.getRandNum() * 2);
                        k2 += sampleMask.magnitude
                          * (1 - rand.getRandNum() * 2);
                    }
                }

                r = camera.generateRay(
                        sampleNo,
                        currentTimeSlice,
                        k1 * pixelGap, k2 * pixelGap);

                r.trace(world, background, pixel);

                fRed += pixel.r;
                fGreen += pixel.g;
                fBlue += pixel.b;
                fAlpha += pixel.a;

                r.dispose();
            }

            fRed /= samplesPerPixel;
            fGreen /= samplesPerPixel;
            fBlue /= samplesPerPixel;
            fAlpha /= samplesPerPixel;

            if (clamp)
            {
                fRed = clamp(fRed);
                fGreen = clamp(fGreen);
                fBlue = clamp(fBlue);
                fAlpha = clamp(fAlpha);
            }

            if (format == RGB)
            {
                out[(j - startX) * 3] = ((byte)(fRed * FACTOR));
                out[(j - startX) * 3 + 1] = ((byte)(fGreen * FACTOR));
                out[(j - startX) * 3 + 2] = ((byte)(fBlue * FACTOR));
            }
            else
            {
                out[(j - startX) * 4] = ((byte)(fRed * FACTOR));
                out[(j - startX) * 4 + 1] = ((byte)(fGreen * FACTOR));
                out[(j - startX) * 4 + 2] = ((byte)(fBlue * FACTOR));
                out[(j - startX) * 4 + 3] = ((byte)(fAlpha * FACTOR));
            }
        }

        currentY--;
        
        return true;
    }
}
