package org.bouncycastle.graphics.rt.shaders;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a general purpose Cook-Torrance shader. Contains all the standard attributes.
 */
public abstract class CookTorranceShader
    extends Shader
{
    Pixel                 pix = new Pixel();

    protected Color3f     ambient = new Color3f();

    public CookTorranceShader(
        World    world)
    {
        super(world);
    }

    /**
     * apply the default parameters for a shader,
     * from the passed in parameters table.
     */
    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        this.setAmbient(params.getColor("ambient", "0, 0, 0"));
    }

    /**
     * return the ambient contribution of the given world to this
     * surface.
     */
    protected Color3f ambientContribution(
        World    world)
    {
        return ambient;
    }

    /**
     * return the diffuse contribution of the given world to this
     * hit.
     */
    protected Color3f diffuseContribution(
        LightDataList     contribs,
        Hit               hit)
    {
        Color3f contrib = new Color3f();

        contrib.set(0, 0, 0);

        LightDataIterator it = contribs.iterator();

        while (it.hasNext())
        {
            LightData ld = it.next();

            contrib.scaleAdd(ld.NDotL(), ld.color(), contrib);
        }

        it.dispose();

        return contrib;
    }

    /**
     * return the specular contribution of the given world to this
     * hit, calculated according to the given roughness.
     */
    protected Color3f specularContribution(
        LightDataList contribs,
        Hit           hit,
        float         roughness)
    {
        Color3f contrib = new Color3f();
        
        contrib.set(0, 0, 0);

        LightDataIterator it = contribs.iterator();

        while (it.hasNext())
        {
            LightData   ld = it.next();
            float       fact;

            fact = ld.NDotL() - hit.NDotI();
            if (fact > 0.0)
            {
                fact /= (float)Math.sqrt(2 - 2 * ld.L().dot(hit.I()));
                fact = (float)Math.pow(fact, 1.0 / roughness);

                contrib.scaleAdd(fact, ld.color(), contrib);
            }
        }

        it.dispose();

        return contrib;
    }

    /**
     * set the ambient color for this object (usually multiplied with
     * the base color).
     */
    public void setAmbient(
        double    r,
        double    g,
        double    b)
    {
        ambient.set((float)r, (float)g, (float)b);
    }

    /**
     * set the ambient color for this object (usually multiplied with
     * the base color).
     */
    public void setAmbient(
        Color3f c)
    {
        ambient.set(c.x, c.y, c.z);
    }

    /**
     * the shade routine. 
     */
    public abstract void shade(Hit hit, Pixel p);
}
