package org.bouncycastle.graphics.rt.shaders.ct;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.tools.io.*;
import org.bouncycastle.graphics.rt.util.ParametersMap;
import org.bouncycastle.graphics.rt.shaders.CookTorranceShader;

/**
 * A matte surface shader.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;Matte diffuse="0.9"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float    "CDATA"&gt;
 *     &lt;!ENTITY %Color3f  "CDATA"&gt;
 *
 *     &lt;!ELEMENT Matte&gt;
 *     &lt;!ATTLIST Matte
 *         ambient              (%Color3f; | %Float;) "0, 0, 0"
 *         diffuse              (%Color3f; | %Float;) "1, 1, 1"&gt;
 * </pre>
 * </code>
 */
public class Matte
    extends CookTorranceShader
{
    protected Color3f       diffuse = new Color3f();

    public Matte(
        World world)
    {
        super(world);
    }

    public static Shader getInstance(
        World         world,
        RMLReader       in)
        throws IOException
    {
        Matte   m = new Matte(world);

        m.readObject(in);

        return m;
    }

    /**
     * handle the parameters we want, passing the extras
     * to our super class for processing.
     */
    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        diffuse = params.getColor("diffuse", "1");
    }

    /**
     * set the diffuse components for this object
     */
    public void setDiffuse(
        double    kD)
    {
        diffuse.set((float)kD, (float)kD, (float)kD);
    }

    /**
     * set the diffuse components for this object
     */
    public void setDiffuse(
        double    r,
        double    g,
        double    b)
    {
        diffuse.set((float)r, (float)g, (float)b);
    }

    public void shade(
        Hit      hit,
        Pixel    p)
    {
        p.set(0, 0, 0, 1);

        p.scaledAdd(hit.color(), ambientContribution(world));

        p.scaledAdd(hit.color(), diffuse, diffuseContribution(hit.visibleLights(world), hit));

        p.scale(alpha);
    }
}
