package org.bouncycastle.graphics.rt.shaders.ct;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;
import org.bouncycastle.graphics.rt.shaders.CookTorranceShader;

/**
 * a metal surface shader.
 */
public class Metal
    extends CookTorranceShader
{
    protected int       maxLevel = 6;

    protected float     roughness = 0;
    protected float     reflectionRoughness = 0;

    protected boolean   reflectingSurface = false;
    protected Color3f   reflectance = new Color3f();

    protected Color3f   specular = new Color3f();

    public Metal(
        World world)
    {
        super(world);
    }

    public static Shader getInstance(
        World           world,
        RMLReader       in)
        throws IOException
    {
        Metal m = new Metal(world);

        m.readObject(in);

        return m;
    }

    /**
     * process our parameter set, handing the extras back to the super class.
     */
    public void processParameters(
        ParametersMap   params)
    {
        super.processParameters(params);

        specular = params.getColor("specular", "1, 1, 1");
        roughness = params.getFloat("roughness", "0");
        reflectance = params.getColor("reflectance", "0, 0, 0");
        reflectionRoughness = params.getFloat("reflectionRoughness", "0");

        if ((reflectance.x != 0)
            || (reflectance.y != 0) || (reflectance.z != 0))
        {
            reflectingSurface = true;
        }
    }

    /**
     * set the reflective component for this surface.
     */
    public void setReflectance(
        double    r,
        double    g,
        double    b)
    {
        reflectance.set((float)r, (float)g, (float)b);
        if ((r != 0) || (g != 0) || (b != 0))
        {
            reflectingSurface = true;
        }
    }

    /**
     * return whether, or not, the surface is reflecting.
     */
    public boolean isReflecting()
    {
        return reflectingSurface;
    }

    /**
     * set a roughness value to perturb the reflectance ray by. The value
     * must be between 0 and 1, with a value of 0 giving the sharpest
     * reflections.
     */
    public void setRoughFinishFactor(
        double    reflectionRoughness)
    {
        this.reflectionRoughness = (float)reflectionRoughness;
    }

    /**
     * the shade routine. If you want more than the below, this is
     * the method to override.
     */
    public void shade(
        Hit        hit,
        Pixel      p)
    {
        p.set(0, 0, 0, 1);

        p.scaledAdd(hit.color(), ambientContribution(world));

        p.scaledAdd(hit.color(), specular,
            specularContribution(hit.visibleLights(world), hit, roughness));

        //
        // with secondary rays there is always the chance we
        // can recurse into infinity, so we have to set a hard limit
        // to stop at.
        //
        if (hit.level() >= maxLevel)
        {
            p.scale(alpha);

            return;
        }

        //
        // fire off a reflection ray (if necessary).
        //
        if (isReflecting())
        {
            if (reflectionRoughness != 0)
            {
                hit.perturbN(reflectionRoughness);
            }

            Ray       ir = hit.reflectedRay();
            Pixel     pix = new Pixel();

            ir.trace(world, pix);

            p.scaledAdd(hit.color(), reflectance, pix);

            ir.dispose();
        }

        p.scale(alpha);
    }
}
