package org.bouncycastle.graphics.rt.shaders.ct;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;
import org.bouncycastle.graphics.rt.shaders.CookTorranceShader;

/**
 * A plastic surface shader.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;Plastic diffuse="0.2" specular="0.2" roughness="0.14" reflectance="0.8"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float    "CDATA"&gt;
 *     &lt;!ENTITY %Color3f  "CDATA"&gt;
 *
 *     &lt;!ELEMENT Plastic&gt;
 *     &lt;!ATTLIST Plastic
 *         ambient              (%Color3f; | %Float;) "0, 0, 0"
 *         diffuse              (%Color3f; | %Float;) "1, 1, 1"
 *         specular             (%Color3f; | %Float;) "0, 0, 0"
 *         roughness            %Float                "0"
 *         reflectance          (%Color3f; | %Float;) "0, 0, 0"
 *         reflectionRoughness  %Float                "0"&gt;
 * </pre>
 * </code>
 */
public class Plastic
    extends CookTorranceShader
{
    protected Color3f       diffuse = new Color3f();
    protected Color3f       specular = new Color3f();
    protected Color3f       reflectance = new Color3f();

    protected float         roughness = 0;
    protected float         reflectionRoughness = 0;

    protected boolean       reflectingSurface = false;

    protected int           maxLevel = 6;

    public Plastic(
        World world)
    {
        super(world);
    }

    public static Shader getInstance(
        World           world,
        RMLReader       in)
        throws IOException
    {
        Plastic p = new Plastic(world);

        p.readObject(in);

        return p;
    }

    /**
     * apply the default parameters for a shader,
     * from the passed in parameters table.
     */
    protected void processParameters(
        ParametersMap    params)
    {
        this.setDiffuse(params.getColor("diffuse", "1, 1, 1"));
        this.setSpecular(params.getColor("specular", "0, 0, 0"));
        this.setRoughness(params.getFloat("roughness", "0"));
        this.setReflectance(params.getColor("reflectance", "0, 0, 0"));
        this.setReflectionRoughness(params.getFloat("reflectionRoughness", "0"));

        super.processParameters(params);
    }

    /**
     * set the diffuse components for this object
     */
    public void setDiffuse(
        double    kD)
    {
        diffuse.set((float)kD, (float)kD, (float)kD);
    }

    /**
     * set the diffuse components for this object
     */
    public void setDiffuse(
        double    r,
        double    g,
        double    b)
    {
        diffuse.set((float)r, (float)g, (float)b);
    }

    /**
     * set the diffuse components for this object
     */
    public void setDiffuse(
        Color3f c)
    {
        diffuse.set(c.x, c.y, c.z);
    }

    /**
     * set the specular components for this object
     */
    public void setSpecular(
        Color3f c)
    {
        specular.set(c.x, c.y, c.z);
    }

    /**
     * set the roughness of the surface, the smaller we are the more
     * refined the specular highlights.
     */
    public void setRoughness(
        double    roughness)
    {
        this.roughness = (float)roughness;
    }

    /**
     * return whether, or not, the surface is specular.
     */
    public boolean isSpecular()
    {
        return (roughness != 0);
    }

    /**
     * set the reflective component for this surface.
     */
    public void setReflectance(
        double    r,
        double    g,
        double    b)
    {
        reflectance.set((float)r, (float)g, (float)b);
        if ((r != 0) || (g != 0) || (b != 0))
        {
            reflectingSurface = true;
        }
    }

    /**
     * set the reflective component for this surface.
     */
    public void setReflectance(
        Color3f c)
    {
        setReflectance(c.x, c.y, c.z);
    }

    /**
     * return whether, or not, the surface is reflecting.
     */
    public boolean isReflecting()
    {
        return reflectingSurface;
    }

    public void setReflectionRoughness(
        double    reflectionRoughness)
    {
        this.reflectionRoughness = (float)reflectionRoughness;
    }

    /**
     * the shade routine. If you want more than the below, this is
     * the method to override.
     */
    public void shade(
        Hit        hit,
        Pixel      p)
    {
        p.set(0, 0, 0, 1);

        p.scaledAdd(hit.color(), ambientContribution(world));

        LightDataList   contribs = hit.visibleLights(world);

        p.scaledAdd(hit.color(), diffuse, diffuseContribution(contribs, hit));

        if (isSpecular())
        {
            p.add(specularContribution(contribs, hit, roughness));
        }

        //
        // with secondary rays there is always the chance we
        // can recurse into infinity, so we have to set a hard limit
        // to stop at.
        //
        if (hit.level() >= maxLevel)
        {
            return;
        }

        if (reflectionRoughness != 0)
        {
            hit.perturbN(reflectionRoughness);
        }

        //
        // fire off a reflection ray (if necessary).
        //
        if (isReflecting())
        {
            Ray      ir = hit.reflectedRay();
            Pixel    pix = new Pixel();

            ir.trace(world, pix);

            p.add(pix);

            ir.dispose();
        }

        p.scale(alpha);
    }
}
