package org.bouncycastle.graphics.rt.shaders.ct;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.util.*;
import org.bouncycastle.graphics.rt.tools.io.*;
import org.bouncycastle.graphics.rt.shaders.CookTorranceShader;

/**
 * a general shader with turbulence support
 */
public abstract class TurbulenceShader
    extends Plastic
{
    protected Color3f           base = new Color3f(0, 0, 0);
    protected Color3f           blendColor = new Color3f(0, 0, 0);
    protected ColorMap          colorMap = null;

    protected float             blend = 1;
    protected float             squeeze = 0; 
    protected float             percentageTurbulence = 1;
    protected int               octaves = 6;

    protected NoiseGenerator    generator = NoiseGenerator.getInstance();

    public TurbulenceShader(
        World world)
    {
        super(world);
    }

    /**
     * process the parameters we want before handing them
     * to our super class.
     */
    protected void processParameters(
        ParametersMap    params)
    {
        this.setBlendColor(params.getColor("blendColor", "0, 0, 0"));
        this.setPercentageTurbulence(params.getFloat("turbulence", "1"));
        this.setSqueeze(params.getFloat("squeeze", "0"));

        super.processParameters(params);
    }

    /**
     * process the object passed in if we can use it, or pass
     * it up to our super class.
     */
    protected void processObject(
        Object          obj)
    {
        if (obj instanceof ColorMap)
        {
            this.colorMap = (ColorMap)obj;
        }
        else
        {
            super.processObject(obj);
        }
    }

    /**
     * set the percentage of turbulence to incorporate
     */
    public void setPercentageTurbulence(
        float   percentage)
    {
        this.percentageTurbulence = percentage;
    }

    /**
     * set the squeeze on the texture lines
     */
    public void setSqueeze(
        float   squeeze)
    {
        this.squeeze = squeeze;
    }

    /**
     * set the percentage of turbulence to incorporate
     */
    public void setBlendColor(
        Color3f   blendColor)
    {
        this.blendColor.set(blendColor);
    }

    /**
     * Accumulate a 3D noise function over octaves octaves, scaling
     * each by 1 / f
     */
    protected float turbulence(
        Tuple3f     pos,
        int         octaves)
    {
        double      t = 0.0,
                    f = 1.0;
        Point3f     p = new Point3f(pos);
            
        for (int i = 0; i < octaves; i++)
        {
            t += Math.abs(generator.getNoise(p)) / f;
            p.scale((float)(1 / 0.527849474673));
            f /= 0.527849474673;
        }

        return (float)t;
    }

    public abstract void shade(Hit hit, Pixel p);
}
