package org.bouncycastle.graphics.rt.tools.io;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.objects.*;
import org.bouncycastle.graphics.rt.util.ParametersMap;

/**
 * a buffered reader that parsers JML files.
 */
public class RMLReader
    extends BufferedReader
{
    public static int       EOF = -1;

    public static int       PROCESSING_INSTRUCTION = 0;
    public static int       DOC_TYPE = 1;
    public static int       CDATA_SECTION = 2;
    public static int       COMMENT = 3;

    public static int       BEGIN_TAG = 4;
    public static int       END_TAG = 5;
    public static int       EMPTY_TAG = 6;
    public static int       PCDATA = 7;

    private StringBuffer    buf = new StringBuffer(4096);
    private ParametersMap   parameters = new ParametersMap();
    private StringBuffer    specialBuf = new StringBuffer();
    private int             lineCount = 1;
    private int             token = EOF;

    private boolean         haveLessThan = false;

    private static HashMap  special = new HashMap();

    private Renderer        renderer = null;
    private Shader          currentShader = null;
    private Transform       currentTransform = new Transform();

    private String[]        objectPath =
        {
            "",
            "org.bouncycastle.graphics.rt.objects", 
            "org.bouncycastle.graphics.rt.models", 
            "org.bouncycastle.graphics.rt.shaders.ct",
            "org.bouncycastle.graphics.rt.util",
            "org.bouncycastle.graphics.rt",
            "org.bouncycastle.graphics.rt.renderers"
        };

    private Class[]         standardParameters =
        {
            RMLReader.class
        };

    private Class[]         shaderParameters =
        {
            World.class,
            RMLReader.class
        };

    private Object[]    standardArgs = new Object[1];
    private Object[]    shaderArgs = new Object[2];

    private HashMap     initialisers = new HashMap();

    private HashMap     cameras = new HashMap();
    private HashMap     backgrounds = new HashMap();

    String              cameraName = null;
    String              backgroundName = null;

    public RMLReader(
        Reader in)
    {
        super(in);
    }

    /**
     * read in a begin/end/empty tag
     */
    private int readTag()
        throws IOException
    {
        int c;
        int tagType = BEGIN_TAG;
    
        if ((c = read()) == '/')
        {
            tagType = END_TAG;
        }
        else if (c == '?')          // processing instruction
        {
            tagType = PROCESSING_INSTRUCTION;
        }
        else
        {
            buf.append((char)c);
        }
    
        if (c < 0)
        {
            return EOF;
        }

        while (((c = read()) != '>') && (c > 0))
        {
            buf.append((char)c);
        }
        
        if (c < 0)
        {
            return EOF;
        }

        if (buf.charAt(buf.length() - 1) == '/')
        {
            buf.setLength(buf.length() - 1);
    
            tagType = EMPTY_TAG;
        }
        else if (tagType == PROCESSING_INSTRUCTION)
        {
            buf.setLength(buf.length() - 1);
        }
        else if (buf.charAt(0) == '!')
        {
            if (buf.charAt(1) == 'D')
            {
                tagType = DOC_TYPE;
            }
            else if (buf.charAt(1) == '[')
            {
                tagType = CDATA_SECTION;
            }
            else
            {
                tagType = COMMENT;
            }
        }
    
        haveLessThan = false;

        //
        // if our tag has parameters we need to populate the parameters table
        //
        if ((tagType == BEGIN_TAG) || (tagType == EMPTY_TAG))
        {
            int length = buf.length();
            int idx = 0, endTagName;

            parameters.clear();

            while ((idx < length) && !Character.isWhitespace(buf.charAt(idx)))
            {
                idx++;
            }

            endTagName = idx;

            while (idx < length)
            {
                String  name, value;
                int     next = idx;

                while ((next < length) && (buf.charAt(next) != '='))
                {
                    next++;
                }

                if (next >= length)
                {
                    break;
                }

                name = buf.substring(idx, next).trim();

                //
                // find the next value string.
                //
                idx = next;

                while ((next < length) && (buf.charAt(next) != '"'))
                {
                    next++;
                }

                idx = ++next;

                while ((next < length) && (buf.charAt(next) != '"'))
                {
                    next++;
                }

                if (next >= length)
                {
                    break;
                }

                value = buf.substring(idx, next);

                parameters.put(name, value);

                idx = next + 1;
            }

            buf.setLength(endTagName);
        }

        return tagType;
    }

    /**
     * read in the next token, returning its type value.
     */
    public int readNext()
        throws IOException
    {
        int c = 0;

        buf.setLength(0);

        if (haveLessThan)                   // reading a tag.
        {
            token = readTag();
            return token;
        }
        else                                // reading PCDATA
        {
            while (((c = read()) != '<') && (c > 0))
            {
                if (c == '\n')
                {
                    lineCount++;
                    continue;
                }
                else if (c == '\r')     // just ignore
                {
                    continue;
                }
                else if (c == '&')      // special character
                {
                    specialBuf.setLength(0);
                    while (((c = read()) != ';') && (c > 0))
                    {
                        specialBuf.append((char)c);
                    }
                    buf.append(special.get(specialBuf.toString()));
                }

                buf.append((char)c);
            }

            if (c < 0)
            {
                return EOF;
            }
        }

        //
        // something before the '<' - PCDATA
        //
        if (buf.length() != 0)
        {
            haveLessThan = true;
            token = PCDATA;
            return PCDATA;
        }
       
        token = readTag();

        return token;
    }

    /**
     * return the parameters associated with the last BEGIN_TAG or EMPTY_TAG
     * read.
     */
    public ParametersMap parameters()
    {
        return parameters;
    }

    /**
     * return the string value of what the current token represents
     */
    public String toString()
    {
        return buf.toString();
    }

    /**
     * return the current token type value
     */
    public int token()
    {
        return token;
    }

    /**
     * return the current line number for the input stream
     */
    public int lineCount()
    {
        return lineCount;
    }

    /**
     * skip to the next end tag representing the close of this tag
     */
    public void skipToEndTag(
        String  tagName)
        throws IOException
    {
        int depth = 1;
        int tok;

        while ((tok = this.readNext()) != EOF)
        {
            if (tok == END_TAG)
            {
                if (tagName.equals(this.toString()))
                {
                    if (--depth == 0)
                    {
                        return;
                    }
                }
            }
            else if (tok == BEGIN_TAG)
            {
                if (tagName.equals(this.toString()))
                {
                    depth++;
                }
            }
        }

        if (tok == EOF)
        {
            throw new IOException("unexpected EOF skipping tag "
                                + tagName + " from line " + lineCount);
        }
    }

    private Renderer processSceneTag()
        throws IOException
    {
        ParametersMap   params = this.parameters();
        Renderer        renderer = null;
        Class           parent = null;
        String          className = params.getString("renderer", "StdRenderer");

        cameraName = params.getString("camera", "default");
        backgroundName = params.getString("background", "default");

        cameras.put(cameraName, new Camera());
        backgrounds.put(backgroundName, new Background());

        for (int i = 0; parent == null && i != objectPath.length; i++)
        {
            try
            {
                if (objectPath[i].equals(""))
                {
                    parent = Class.forName(className);
                }
                else
                {
                    parent = Class.forName(objectPath[i] + "." + className);
                }
            }
            catch (ClassNotFoundException e)
            {
                parent = null;
            }
        }

        if (parent == null)
        {
            throw new IOException("Can't find requested renderer " + className);
        }
        else
        {
            Constructor construct = null;

            try
            {
                construct = parent.getConstructor(standardParameters);
            }
            catch (NoSuchMethodException e)
            {
                throw new IOException("Requested renderer object not compatible");
            }

            standardArgs[0] = this;

            try
            {
                renderer = (Renderer)construct.newInstance(standardArgs);
            }
            catch (Exception e)
            {
                e.printStackTrace(System.err);
                throw new IOException("Exception constructing renderer: " + e);
            }
        }

        return renderer;
    }

    private Object processToken()
        throws IOException
    {
        String      value = this.toString();
        Method      initialiser = (Method)initialisers.get(value);

        if (initialiser == null)
        {
            Class       cls = null;

            for (int i = 0; cls == null && i != objectPath.length; i++)
            {
                try
                {
                    if (objectPath[i].equals(""))
                    {
                        cls = Class.forName(value);
                    }
                    else
                    {
                        cls = Class.forName(objectPath[i] + "." + value);
                    }
                }
                catch (ClassNotFoundException e)
                {
                    cls = null;
                }
            }

            if (cls == null)
            {
                throw new IOException("Unknown object "
                            + value + " found on line " + this.lineCount());
            }
            else
            {
                try
                {
                    if (Shader.class.isAssignableFrom(cls))
                    {
                        initialiser = cls.getMethod("getInstance", shaderParameters);
                    }
                    else
                    {
                        initialiser = cls.getMethod("getInstance", standardParameters);
                    }
                }
                catch (NoSuchMethodException e)
                {
                    throw new IOException("Incompatible object "
                            + value + " found on line " + this.lineCount());
                }
            }

            initialisers.put(value, initialiser);
        }

        Class   parent = initialiser.getDeclaringClass();

        if (Shader.class.isAssignableFrom(parent))
        {
            shaderArgs[0] = renderer.getWorld();
            shaderArgs[1] = this;

            try
            {
                return initialiser.invoke(null, shaderArgs);
            }
            catch (Exception e)
            {
                e.printStackTrace(System.err);
                throw new IOException("line " + this.lineCount()
                                        + ", exception on input: " + e);
            }
        }
        else
        {
            standardArgs[0] = this;

            try
            {
                return initialiser.invoke(null, standardArgs);
            }
            catch (Exception e)
            {
                e.printStackTrace(System.err);
                throw new IOException("line " + this.lineCount()
                                        + ", exception on input: " + e);
            }
        }
    }

    /**
     * return the documents idea of the current transform.
     */
    public Transform getTransform()
    {
        return currentTransform;
    }

    /**
     * set the current shader for the objects we are about to parse.
     */
    public void setTransform(
        Transform  t)
    {
        currentTransform = t;
    }

    /**
     * read an object from the RML stream.
     */
    public Object readObject()
        throws IOException
    {
        int tok = this.readNext();

        //
        // skip the header
        //
        while ((tok != RMLReader.EOF)
            && (tok == RMLReader.PROCESSING_INSTRUCTION || tok == PCDATA
            || tok == RMLReader.DOC_TYPE || tok == RMLReader.COMMENT))
                tok = this.readNext();

        if (tok == BEGIN_TAG)
        {
            if (this.toString().equals("Scene"))
            {
                Aggregate  list = new Aggregate();

                renderer = processSceneTag();

                while ((tok = this.readNext()) != RMLReader.EOF)
                {
                    String value = this.toString();

                    if (tok == RMLReader.END_TAG && value.equals("Scene"))
                    {
                        renderer.setObjects(new BvTree(list));
                        renderer.setCamera((Camera)cameras.get(cameraName));
                        renderer.setBackground((Background)backgrounds.get(backgroundName));

                        return renderer;      // we are finished
                    }

                    if ((tok == RMLReader.BEGIN_TAG)
                        || (tok == RMLReader.EMPTY_TAG))
                    {
                        Object  obj = processToken();

                        if (obj instanceof Shader)
                        {
                            list.add((Shader)obj);
                        }
                        else if (obj instanceof Aggregate)
                        {
                            Aggregate   agr = (Aggregate)obj;

                            list.add(agr);
                        }
                        else if (obj instanceof Renderable)
                        {
                            Renderable    o = (Renderable)obj;

                            list.add(o);
                        }
                        else if (obj instanceof Light)
                        {
                            Light   l = (Light)obj;

                            list.add(l);
                        }
                        else if (obj instanceof Camera)
                        {
                            Camera  c = (Camera)obj;

                            cameras.put(c.getName(), c);
                        }
                        else
                        {
                            Background  b = (Background)obj;

                            backgrounds.put(b.getName(), b);
                        }
                    }
                }
            }
            else
            {
                return processToken();
            }
        }
        else if (tok == EMPTY_TAG)
        {
            return processToken();
        }
        else if (tok == END_TAG || tok == EOF)       
        {
            // we interpret a lone end tag as the end of the initiating object
            //
            return null;
        }

        return this.toString();
    }

    static
    {
        //
        // interpreted characters in PCDATA
        //
		special.put("amp", String.valueOf((char)38));
		special.put("gt", String.valueOf((char)62));
		special.put("lt", String.valueOf((char)60));
		special.put("quot", String.valueOf((char)34));
    }
}
