package org.bouncycastle.graphics.rt.util;

import javax.vecmath.*;

/*
 * This is it!. Ken Perlin's noise function... straight from him
 * (Well, SIGGRAPH '92 tutorial 23  course notes)
 */
public class NoiseGenerator
{
    static final int B    = 512;
    static final int NUMPTS    = B;

    static double EPSILON = 0.0001;

    static int[]         p;
    static Vector3f[]    g;

    /*
     * Note we need consistent random numbers and init_noise may be called at
     * any time. Eg say the same scene but with more samples in the
     * lights etc etc - this would cause textures to be different.
     */
    private static RandomTable      rand = new RandomTable();

    private static NoiseGenerator   generator = new NoiseGenerator();

    /**
     * Initialises the psuedo random gradient table G(.x, .y. z) and a
     * psuedo random value at the lattice points (G.r). Also initialises
     * the index table P.
     */
    private NoiseGenerator()
    {
        if (g != null)
        {            // we've already set up.
            return;
        }

        g = new Vector3f[NUMPTS + NUMPTS + 2];
        p = new int[NUMPTS + NUMPTS + 2];

        for (int n = 0; n < NUMPTS; n++)
        {
            double      dot;
            Vector3f    v;

            v = new Vector3f();

            do
            {
                v.set(
                    2 * rand.getRandNum() - 1,
                    2 * rand.getRandNum() - 1,
                    2 * rand.getRandNum() - 1);
                dot = v.dot(v);
            }
            while ((dot > 1) || (dot < EPSILON));

            v.normalize();
            
            g[n] = v;
        }


        //
        // initialize the index table.
        //
        for (int n = 0; n < NUMPTS; n++)
        {
            p[n] = n;
        }

        //
        // Set a random permutation of the first NUMPTS integers
        //
        for (int n = 0; n < NUMPTS; n++)
        {
            int    i, tmp;

            i = (int)(rand.getRandNum() * (NUMPTS - 1));
            tmp = p[n];
            p[n] = p[i];
            p[i] = tmp;
        }

        //
        // Extend g and p arrays to allow faster indexing
        //
        for (int i = 0; i < B + 2; i++)
        {
            p[B + i] = p[i];
            g[B + i] = g[i];
        }
    }

    private float s_curve(
        float    t)
    {
        return (t * t * (3 - 2 * t));
    }

    private float lerp(
        float    t,
        float    a,
        float    b)
    {
        return (a + t * (b - a));
    }

    private float at(
        Vector3f    q,
        float        rx,
        float        ry,
        float        rz)
    {
        return (rx * q.x + ry * q.y + rz * q.z);
    }

    static public NoiseGenerator getInstance()
    {
        return generator;
    }

    public float getNoise(
        Tuple3f    loc)
    {
        int     bx0, bx1, by0, by1, bz0, bz1, b00, b10, b01, b11;
        float   rx0, rx1, ry0, ry1, rz0, rz1, sx, sy, sz;
        float   a, b, c, d, t, u, v;

        //
        // calculate indexes for the x position
        //
        t = loc.x + 10000;

        if (t < 0.0)        // couldn't we use an abs here?
        {
            t += 10000;
        }

        bx0 = (int)t & (B - 1);
        bx1 = (bx0 + 1) & (B - 1);
        rx0 = t - (int)t;
        rx1 = rx0 - 1;

        //
        // calculate indexes for the y position
        //
        t = loc.y + 10000;

        if (t < 0.0)        // couldn't we use an abs here?
        {
            t += 10000;
        }

        by0 = (int)t & (B - 1);
        by1 = (by0 + 1) & (B - 1);
        ry0 = t - (int)t;
        ry1 = ry0 - 1;

        //
        // calculate indexes for the z position
        //
        t = loc.z + 10000;

        if (t < 0.0)        // couldn't we use an abs here?
        {
            t += 10000;
        }

        bz0 = (int)t & (B - 1);
        bz1 = (bz0 + 1) & (B - 1);
        rz0 = t - (int)t;
        rz1 = rz0 - 1;

        //
        // interpolation
        //
        int    i, j;

        i = p[bx0];
        j = p[bx1];

        b00 = p[i + by0];
        b10 = p[j + by0];
        b01 = p[i + by1];
        b11 = p[j + by1];

        sx = s_curve(rx0);
        sy = s_curve(ry0);
        sz = s_curve(rz0);

        u = at(g[b00 + bz0], rx0, ry0, rz0);
        v = at(g[b10 + bz0], rx1, ry0, rz0);
        a = lerp(sx, u, v);

        u = at(g[b01 + bz0], rx0, ry1, rz0);
        v = at(g[b11 + bz0], rx1, ry1, rz0);
        b = lerp(sx, u, v);

        c = lerp(sy, a, b);        // interpolate in y at lo x

        u = at(g[b00 + bz1], rx0, ry0, rz1);
        v = at(g[b10 + bz1], rx1, ry0, rz1);
        a = lerp(sx, u, v);

        u = at(g[b01 + bz1], rx0, ry1, rz1);
        v = at(g[b11 + bz1], rx1, ry1, rz1);
        b = lerp(sx, u, v);

        d = lerp(sy, a, b);        // interpolate in y at hi x 

        return((float)1.5 * lerp(sz, c, d));    // interpolate in z 
    }

    static float OFFSET1 = (float)1000.1234;
    static float OFFSET2 = (float)2000.1234;
    static float OFFSET3 = (float)4000.1234;

    /*
     * A vector valued noise function.
     * (Again, from the '89 paper pp 259)
     */
    public Vector3f getNoiseV(
        Tuple3f        p)
    {
        Vector3f s = new Vector3f();

        s.set(
            p.x - OFFSET1,
            p.y - OFFSET1,
            p.z - OFFSET1);

        float   x = getNoise(s);

        s.set(
            p.x + OFFSET1,
            p.y - OFFSET2,
            p.z + OFFSET3);

        float   y = getNoise(s);

        s.set(
            p.x + OFFSET2,
            p.y + OFFSET2,
            p.z + OFFSET2);

        float   z = getNoise(s);

        s.set(x, y, z);

        return s;
    }
}
