package org.bouncycastle.graphics.rt.util;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * a generic wave source.
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;WaveSource center="0, 0, 0" amplitude="0.5" wavelength="0.2"/&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Color3f "CDATA"&gt;
 *     &lt;!ENTITY %Float   "CDATA"&gt;
 *     &lt;!ENTITY %Point3f "CDATA"&gt;
 *
 *     &lt;!ELEMENT WaveSource EMPTY&gt;
 *     &lt;!ATTLIST WaveSource
 *         color      %Color3f; "1, 1, 1"
 *         wavelength %Float;   "6.2"
 *         amplitude  %Float;   "1"
 *         phase      %Float;   "0"
 *         damping    %Float;   "0"
 *         center     %Point3f; "0, 0, 0"&gt;
 * </pre>
 * </code>
 */
public class WaveSource
{
    Point3f     center;
    Vector3f    tmp = new Vector3f();

    float       wavelength = (float)(2 * Math.PI);
    float       amplitude = 1;
    float       phase = 0;
    float       damping = 0;

    WaveSource  next;

    public WaveSource(
        Point3f     center,
        double      wavelength,
        double      amplitude,
        double      phase,
        double      damping)
    {
        init(center, wavelength, amplitude, phase, damping);
    }

    public static WaveSource getInstance(
        RMLReader       in)
        throws IOException
    {
        ParametersMap   params = in.parameters();

        WaveSource      source = new WaveSource(
                params.getPoint("center", "0, 0, 0"),
                params.getFloat("wavelength", "6.2"),
                params.getFloat("amplitude", "1"),
                params.getFloat("phase", "0"),
                params.getFloat("damping", "0"));

        if (in.token() != RMLReader.EMPTY_TAG)
        {
            in.skipToEndTag(in.toString());
        }

        return source;
    }

    private void init(
        Point3f     center,
        double      wavelength,
        double      amplitude,
        double      phase,
        double      damping)
    {
        this.center = new Point3f(center);

        if (wavelength != 0.0)
        {
            this.wavelength = (float)(2 * Math.PI / wavelength);
        }
        else
        {
            this.wavelength = (float)(2 * Math.PI);
        }
        this.amplitude = (float)amplitude;
        this.phase = (float)(2 * Math.PI * phase);
        this.damping = (float)damping;
    }
        
    public Vector3f perturbationAt(
        Point3f loc)
    {
        float        mag, damp, val;
    
        tmp.sub(loc, this.center);
    
        mag = tmp.dot(tmp);
        if (mag == 0)
        {
            mag = 1;
        }
    
        mag = (float)Math.sqrt(mag);
    
                /* Really 2 * PI / w.wavelength */
        val = mag * this.wavelength;
    
        if (this.damping != 0)
        {
            damp = (float)Math.exp(-this.damping * mag);
        }
        else
        {
            damp = 1;
        }
        
        val += this.phase;    /* Really 2 * M_PI * w->phase */
    
        val = (float)(this.amplitude * damp * Math.sin(val) / mag);
    
        tmp.scale(val);

        return tmp;
    }
}
