package org.bouncycastle.graphics.rt.util;

import java.io.*;

import javax.vecmath.*;

import org.bouncycastle.graphics.rt.*;
import org.bouncycastle.graphics.rt.tools.io.*;

/**
 * The holding class for one or more wave sources.
 * <p>
 * Example:
 * <code>
 * <pre>
 *    &lt;WaveSourceList&gt;
 *       &lt;WaveSource center="0, 0, 0" amplitude="0.5" wavelength="0.2"/&gt;
 *       &lt;WaveSource center="0, 0, 0.5" amplitude="0.5" wavelength="0.2"/&gt;
 *       &lt;WaveSource center="0, 0.5, 0.5" amplitude="0.5" wavelength="0.2"/&gt;
 *    &lt;/WaveSourceList&gt;
 * </pre>
 * </code>
 * <p>
 * Document type definition:
 * <code>
 * <pre>
 *      &lt;!ELEMENT WaveSourceList (WaveSource)+ &gt;
 * </pre>
 * </code>
 */
public class WaveSourceList
{
    private static class WaveSourceIteratorImpl
        implements WaveSourceIterator
    {
        WaveSource              head;
        boolean                 disposed;
        WaveSourceIteratorImpl  next;

        private static WaveSourceIteratorImpl   itList;

        private WaveSourceIteratorImpl()
        {
        }

        static WaveSourceIteratorImpl getInstance(
            WaveSource   head)
        {
            WaveSourceIteratorImpl  it;

            if (itList == null)
            {
                it = new WaveSourceIteratorImpl();
            }
            else
            {
                it = itList;
                itList = itList.next;
            }
 
            it.head = head;
            it.disposed = false;

            return it;
        }

        public boolean hasNext()
        {
            return (head != null);
        }

        public WaveSource next()
        {
            WaveSource  w = head;

            head = head.next;

            return w;
        }

        public void dispose()
        {
            if (disposed)
            {
                throw new IllegalStateException("attempt to dispose of renderable iterator twice");
            }

            disposed = true;

            this.next = itList;
            itList = this;
        }
    }

    WaveSource    sources = null;

    public WaveSourceList()
    {
    }

    public static WaveSourceList getInstance(
        RMLReader   in)
        throws IOException
    {
        Object              obj;
        WaveSourceList      list = new WaveSourceList();

        while ((obj = in.readObject()) != null)
        {
            if (obj instanceof WaveSource)
            {
                list.add((WaveSource)obj);
            }
        }

        return list;
    }

    public void add(
        WaveSource    source)
    {
        if (sources == null)
        {
            sources = source;
        }
        else
        {
            source.next = sources;
            sources = source;
        }
    }

    public WaveSourceIterator iterator()
    {
        return WaveSourceList.WaveSourceIteratorImpl.getInstance(sources);
    }
}
