package org.bouncycastle.math;

/*
 * NOTE: the routines in this file don't take into account the fact that
 * the remainders may vanish unexpectedley. To date this hasn't happened...
 */
public class QuarticSturmSolver
    extends SturmSolver
{
    int    order0, order1, order2, order3;

    double coef0[] = new double[5];
    double coef1[] = new double[4];
    double coef2[] = new double[3];
    double coef3[] = new double[2];
    double coef4[] = new double[1];

    /**
     * calculate the sign changes in the quartic sturm sequence pointed
     * to by sp for x = 0
     */
    private int changesAtZero()
    {
        int    at0 = 0;

        if (((coef0[0] * coef1[0]) < 0) || (coef0[0] == 0.0))
        {
            at0++;
        }

        if (((coef1[0] * coef2[0]) < 0) || (coef1[0] == 0.0))
        {
            at0++;
        }

        if (order2 == 0)
        {
            return at0;
        }

        if (((coef2[0] * coef3[0]) < 0) || (coef2[0] == 0.0))
        {
            at0++;
        }

        if (order3 == 0)
        {
            return at0;
        }

        if (((coef3[0] * coef4[0]) < 0) || (coef3[0] == 0.0))
        {
            at0++;
        }

        return(at0);
    }

    private int changesAtInfinity()
    {
        int    atInf = 0;

        if ((coef0[order0] * coef1[order1]) < 0)
        {
            atInf++;
        }

        if ((coef1[order1] * coef2[order2]) < 0)
        {
            atInf++;
        }

        if (order2 == 0)
        {
            return atInf;
        }

        if ((coef2[order2] * coef3[order3]) < 0)
        {
            atInf++;
        }

        if (order3 == 0)
        {
            return atInf;
        }

        if ((coef3[order3] * coef4[0]) < 0)
        {
            atInf++;
        }

        return atInf;
    }

    public int changesAtX(
        double    x)
    {
        if (x == 0.0)
        {
            return changesAtZero();
        }
        else if ((x == Double.POSITIVE_INFINITY) || (x == Double.NEGATIVE_INFINITY))
        {
            return changesAtInfinity();
        }

        //
        // we do it the long way...
        //
        double      f, lf;
        int         atX = 0;

        lf = coef0[0] + x * (coef0[1] +
            x * (coef0[2] + x * (coef0[3] + x * coef0[4])));

        //
        // x**3 polynomial
        //
        f = coef1[0] + x * (coef1[1] + x * (coef1[2] + x * coef1[3]));

        if (lf * f < 0 || lf == 0.0)
        {
            atX++;
        }

        lf = f;

        //
        // x**2 polynomial
        //
        f = coef2[0] + x * (coef2[1] + x * coef2[2]);

        if (lf * f < 0 || lf == 0.0)
        {
            atX++;
        }

        if (order2 == 0)
        {
            return atX;
        }

        lf = f;

        //
        // x**1 polynomial
        //
        f = coef3[0] + x * coef3[1];
        if (lf * f <  0 || lf == 0.0)
        {
            atX++;
        }

        if (order3 == 0)
        {
            return atX;
        }

        lf = f;

        //
        // x**0 polynomial
        //
        if (lf * coef4[0] < 0 || lf == 0.0)
        {
            atX++;
        }

        return atX;
    }

    /**
     * calculates the modulus of u order 2 / v order 1.
     */
    private int modp3(
        double[]    u,
        double[]    v,
        double[]    r)
    {
        double carry;

        //
        // stage 1.
        //
        carry = v[1] * u[1] - u[1 + 1] * v[1 - 1];

        r[0] = v[1] * u[0];

        //
        // stage 2 - with sign reversal.
        //
        r[0] = -(v[1] * r[0] - carry * v[0 - 0]);

        return 0;
    }

    /**
     * calculates the modulus of u order 3 / v order 2 leaving it in r.
     */
    private int modp2(
        double[]    u,
        double[]    v,
        double[]    r)
    {
        double carry;

        //
        // stage 1.
        //
        carry = v[2] * u[2] - u[2 + 1] * v[2 - 1];
        r[1] = v[2] * u[1] - u[2 + 1] * v[1 - 1];

        r[0] = v[2] * u[0];

        //
        // stage 2 - with sign reversal.
        //
        r[1] = -(v[2] * r[1] - carry * v[1 - 0]);
        r[0] = -(v[2] * r[0] - carry * v[0 - 0]);

        if (r[1] != 0.0)
        {
            double  f = Math.abs(r[1]);

            r[1] /= f;                          // normalise
            r[0] /= f;

            return 1;
        }

        return 0;
    }

    /**
     * called to divide u order 4 / v order 3
     */
    private int modp1(
        double[]    u,
        double[]    v,
        double[]    r)
    {
        double    carry;

        //
        // loop 1
        //
        carry = v[3] * u[3] - u[3 + 1] * v[3 - 1];
        r[2] = v[3] * u[2] - u[3 + 1] * v[2 - 1];
        r[1] = v[3] * u[1] - u[3 + 1] * v[1 - 1];

        r[0] = v[3] * u[0];

        //
        // loop 2 - with sign reversal
        //
        r[2] = -(v[3] * r[2] - carry * v[2 - 0]);
        r[1] = -(v[3] * r[1] - carry * v[1 - 0]);
        r[0] = -(v[3] * r[0] - carry * v[0 - 0]);

        if (r[2] != 0.0)
        {
            double  f = Math.abs(r[2]);

            r[2] /= f;                          // normalise
            r[1] /= f;
            r[0] /= f;

            return 2;
        }
        
        if (r[1] != 0.0)
        {
            double  f = Math.abs(r[1]);

            r[1] /= f;                          // normalise
            r[0] /= f;

            return 1;
        }

        return 0;
    }

    /**
     * build up a sturm sequence for a polynomial, returning
     * the number of polynomials in the sequence
     */
    public void init(
        double[]    coef)
    {
        System.arraycopy(coef, 0, coef0, 0, coef0.length);

        this.repeated = new boolean[4];

        //
        // guard against degenerate cases
        //
        order0 = 4;

        while (order0 > 0 && coef[order0] == 0.0)
        {
            order0--;
        }

        //
        // calculate the derivative
        //
        order1 = order0 - 1;

        for (int i = 1; i <= order0; i++)
        {
            coef1[i - 1] = coef0[i] * i;
        }

        //
        // construct the rest of the Sturm sequence
        //
        order2 = modp1(coef0, coef1, coef2);
        if (order2 != 0)
        {
            order3 = modp2(coef1, coef2, coef3);
            if (order3 != 0)
            {
                modp3(coef2, coef3, coef4);
            }
        }
    }

    public static void main(
        String[]    args)
    {
        int            atMax, atMin, nRoots;
        QuarticSturmSolver    solver = new QuarticSturmSolver();
        double[]        roots = new double[4];
        double[]        coefs1 = { 5, 4, 3, 2, 1 };
        double[]        coefs2 = { 24, -50, 35, -10, 1 };
        
        solver.init(coefs2);

        atMin = solver.changesAtZero();
        atMax = solver.changesAtX(Double.POSITIVE_INFINITY);

        System.out.println("at 0 : " + atMin);

        System.out.println("at +infinity : " + atMax);

        if (solver.findRoot(0, 100000, roots))
        {
            System.out.println("root = " + roots[0]);
            System.out.println("f(x) = " + (coefs2[0] + roots[0]
                * (coefs2[1] + roots[0]
                   * (coefs2[2] + roots[0]
                     * (coefs2[3] + roots[0] * coefs2[4])))));
        }
        else
        {
            System.out.println("no real roots");
        }
        
        nRoots = solver.findRoots(0, 100000, roots);
        for (int i = 0; i != nRoots; i++)
        {
            System.out.println("root " + i + ": " + roots[i]);
        }
    }
}
