package org.bouncycastle.math;

public class SturmSolver
    implements Solver
{
    static final int        MAX_ORDER = 12;
    static final double     ALG_TOLERANCE = 0.001;

    static final double     MAXERROR = 1.0e-11;
    static final double     RELERROR = 1.0e-10;
    static final int        MAXIT = 800;
    static final int        MAXP2 = 32;

    //
    // a coefficient less than MIN_COEF is considered to be zero.
    //
    static final double     MIN_COEF = 1.0e-12;

    private int             order;
    private double[][]      sequence;

    protected boolean[]     repeated;

    /**
     * calculate the sign changes in the sturm sequence pointed
     * to by sp for x = 0
     */
    private int changesAtZero()
    {
        int         count = 0;
        double[]    f = sequence[0];

        for (int s = 1; s < sequence.length; s++)
        {
                double[]    nf = sequence[s];

                if (f[0] * nf[0] < 0 || f[0] == 0.0)
                {
                    count++;
                }

                f = nf;
        }

        return count;
    }

    /**
     * calculate the sign changes in the sturm sequence pointed
     * to by sp for x at infinity.
     */
    private int changesAtInfinity()
    {
        int         count = 0;
        double[]    f = sequence[0];

        for (int s = 1; s < sequence.length; s++)
        {
                double[]    nf = sequence[s];

                if (f[f.length - 1] * nf[nf.length - 1] < 0
                    || f[f.length - 1] == 0.0)
                {
                    count++;
                }

                f = nf;
        }

        return count;
    }

    /**
     * evaluate the polynomial represented in coef at value x.
     */
    private double eval(
        double[]    coef,
        double      x)
    {
        double  f = 0;

        for (int i = coef.length - 1; i >= 0; i--)
        {
            f = f * x + coef[i];
        }

        return f;
    }

    /**
     * calculate the sign changes in the sturm sequence in at value
     * x
     */
    public int changesAtX(
        double x)
    {
        if (x == 0.0)
        {
            return changesAtZero();
        }
        else if ((x == Double.POSITIVE_INFINITY) || (x == Double.NEGATIVE_INFINITY))
        {
            return changesAtInfinity();
        }

        int     count = 0;
        double  f = eval(sequence[0], x);

        for (int s = 1; s < sequence.length; s++)
        {
                double  nf = eval(sequence[s], x);

                if (f * nf < 0 || f == 0.0)
                {
                    count++;
                }

                f = nf;
        }

        return count;
    }

    /**
     * calculates the modulus of u(x) / v(x) leaving it in r, it
     * returns 0 if r(x) is a constant.
     *
     * note: this function assumes the leading coefficient of v is 1 or -1
     */
    private double[] modp(
        double[]    u,
        double[]    v)
    {
        double[]    r = new double[u.length];

        System.arraycopy(u, 0, r, 0, u.length);

        if (v[v.length - 1] < 0.0)
        {
            for (int i = u.length - v.length - 1; i >= 0; i -= 2)
            {
                r[i] = -r[i];
            }

            for (int i = u.length - v.length; i >= 0; i--)
            {
                for (int j = v.length - 1 + i - 1; j >= i; j--)
                {
                    r[j] = -r[j] - r[v.length - 1 + i] * v[j - i];
                }
            }
        }
        else
        {
            for (int i = u.length - v.length; i >= 0; i--)
            {
                for (int j = v.length + i - 2; j >= i; j--)
                {
                    r[j] -= r[v.length - 1 + i] * v[j - i];
                }
            }
        }

        int order = v.length - 1 - 1;

        while (order >= 0 && Math.abs(r[order]) < MIN_COEF)
        {
            r[order] = 0.0;
            order--;
        }

        if (order <= 0)
        {
            double[]    rt = new double[1];

            rt[0] = r[0];

            return rt;
        }

        double[]    rt = new double[order + 1];

        System.arraycopy(r, 0, rt, 0, order + 1);

        return(rt);
    }

    public void init(
        double[] coefs)
    {
        int order = coefs.length - 1;

        //
        // guard against degenerate cases
        //
        while ((order > 0) && (coefs[order] == 0.0))
        {
            order--;
        }

        sequence = new double[order + 1][];
        repeated = new boolean[order];

        sequence[0] = new double[order + 1];
        sequence[1] = new double[order];

        for (int i = 0; i <= order; i++)
        {
            sequence[0][i] = coefs[i];
        }

        //
        // calculate the derivative and normalise the leading
        // coefficient.
        //
        double      x = Math.abs(coefs[order] * order);

        double[]    cf = sequence[0];
        double[]    df = sequence[1];

        for (int i = 1; i <= order; i++)
        {
            df[i - 1] = cf[i] * i / x;
        }

        //
        // construct the rest of the Sturm sequence
        //
        double[]    r = modp(sequence[0], sequence[1]);
        int         seqNo = 1;

        while (r.length > 1)
        {
            //
            // reverse the sign and normalise
            //
            double   f = -Math.abs(r[r.length - 1]);

            for (int i = 0; i != r.length; i++)
            {
                r[i] /= f;
            }

            sequence[seqNo + 1] = r;

            r = modp(sequence[seqNo], sequence[seqNo + 1]);

            seqNo++;
        }

        //
        // reverse the sign
        //
        sequence[seqNo + 1] = r;

        r[0] = -r[0]; 
    }

    private final double findSingleRoot(
        double      min,
        double      max,
        int         atMin,
        int         atMax)
    {
        double mid;
        int    atMid;

        mid = (min + max) / 2;

        for (int i = 0; i < MAXIT; i++)
        {
            atMid = changesAtX(mid);

            if (Math.abs(mid) > RELERROR)
            {
                if (Math.abs((max - min) / mid) < RELERROR)
                {
                    return mid;
                }
            }
            else if (Math.abs(max - min) < RELERROR)
            {
                return mid;
            }

            if ((atMin - atMid) == 0)
            {
                min = mid;
            }
            else
            {
                max = mid;
            }

            mid = (min + max) / 2;
        }

        return mid;
    }

    /*
     * uses a bisection based on the sturm sequence for the polynomial
     * described in to isolate the lowest root of the polynomial
     * in the bounds min and max, if one exists. The return value for
     * the routine is the number of roots in the interval and the value
     * of the root itself is returned in roots[0].
     *
     * @return whether or not a root was found.
     */
    public boolean findRoot(
        double        min,
        double        max,
        double[]      roots)
    {
        double    mid;
        int       n1, its, atMin, atMid, atMax, nRoots;

        atMin = changesAtX(min);
        atMax = changesAtX(max);

        if ((nRoots = atMin - atMax) == 0)
        {
            return false;
        }

        for (its = 0; its < MAXIT; its++)
        {

            mid = (min + max) / 2;

            atMid = changesAtX(mid);

            n1 = atMin - atMid;

            if (n1 == 1)
            {
                /*
                 * first try a less expensive technique.
                 */
                // if (modrf(sseq->coef, min, mid, &roots[0])!

                //
                // if we get here we have to evaluate the root the hard
                // way by using the Sturm sequence. It also means the
                // root is repeated an even number of times
                //
                
                roots[0] = findSingleRoot(min, mid, atMin, atMid);
                repeated[0] = false;    // not true!

                return true;
            }
            else if (n1 == 0)
            {
                min = mid;
            }
            else
            {
                max = mid;
            }
        }

        if (its == MAXIT && (max - min) > ALG_TOLERANCE)
        {
            // sprintf(buf, "sbisect: overflow min %f max %f diff %e\n", min, max, max - min);
            // printf(buf);
        }

        roots[0] = min;

        return true;
    }

    private void bisect(
        double     min,
        double     max,
        int        atMin,
        int        atMax,
        double[]   roots,
        int        rootOffset)
    {
        int    its;

        if ((atMin - atMax) == 1)
        {
            roots[rootOffset] = findSingleRoot(min, max, atMin, atMax);
            return;
        }

        for (its = 0; its < MAXIT; its++)
        {
            double    mid;
            int    atMid;
            int    n1, n2;

            mid = (min + max) / 2;

            atMid = changesAtX(mid);

            n1 = atMin - atMid;
            n2 = atMid - atMax;

            if (n1 != 0 && n2 != 0)
            {
                bisect(min, mid, atMin, atMid,
                            roots, rootOffset);
                bisect(mid, max, atMid, atMax,
                            roots, rootOffset + n1);
                return;
            }
            else if (n1 == 0)
            {
                min = mid;
            }
            else
            {
                max = mid;
            }
        }

        if (its == MAXIT && (max - min) > ALG_TOLERANCE)
        {
            // sprintf(buf, "sbisect: overflow min %f max %f diff %e\n", min, max, max - min);
            // printf(buf);
        }

        roots[rootOffset] = min;
    }

    public int findRoots(
        double      min,
        double      max,
        double[]    roots)
    {
        int       atMin, atMax, nRoots;

        atMin = changesAtX(min);
        atMax = changesAtX(max);

        if ((nRoots = atMin - atMax) == 0)
        {
            return 0;
        }

        bisect(min, max, atMin, atMax, roots, 0);

        return nRoots;
    }

    public static void main(
        String[]    args)
    {
        int            atMax, atMin, nRoots;
        SturmSolver    solver = new SturmSolver();
        double[]       roots = new double[4];
        double[]       coefs1 = { 5, 4, 3, 2, 1 };
        double[]       coefs2 = { 24, -50, 35, -10, 1 };
        
        solver.init(coefs2);

        atMin = solver.changesAtX(0.0);
        atMax = solver.changesAtX(Double.POSITIVE_INFINITY);

        System.out.println("at 0 : " + atMin);

        System.out.println("at +infinity : " + atMax);

        if (solver.findRoot(0, 100000, roots))
        {
            System.out.println("root = " + roots[0]);
            System.out.println("f(x) = " + (coefs2[0] + roots[0]
                * (coefs2[1] + roots[0]
                   * (coefs2[2] + roots[0]
                     * (coefs2[3] + roots[0] * coefs2[4])))));
        }
        else
        {
            System.out.println("no real roots");
        }
        
        nRoots = solver.findRoots(0, 100000, roots);
        for (int i = 0; i != nRoots; i++)
        {
            System.out.println("root " + i + ": " + roots[i]);
        }
    }
}
